package com.vungle.warren;

import android.support.annotation.IntDef;
import android.support.annotation.RestrictTo;

import java.util.Map;

import static com.vungle.warren.AdConfig.AdSize.VUNGLE_DEFAULT;

/**
 * Settings and Configuration Class for playAd Requests.
 *
 *
 *
 * To prevent any issues, generate a new AdConfig object per playAd() request.
 */
public class AdConfig {

    public static final String FLAG_DIRECT_DOWNLOAD = "direct_download";
    /**
     * Bitmask which stores the configuration settings for this advertisement.
     */
    private int settings;
    private int flexviewCloseSec;
    private int ordinal;

    @RestrictTo(RestrictTo.Scope.LIBRARY)
    public int getSettings(){
        return settings;
    }

    @IntDef(flag = true, value = {MUTED, IMMEDIATE_BACK, IMMERSIVE, TRANSITION_ANIMATE, AUTO_ROTATE})
    public @interface Settings{}
    /**
     * Start this advertisement muted.
     */
    public static final int MUTED = 1;
    /**
     * Allow the user to close the ad immediately after starting it. Usually there is a delay before
     * the X or skip is shown. The back button is the same as the X.
     */
    public static final int IMMEDIATE_BACK = 1 << 1;

    /**
     * Whether or not to use immersive mode. Defaults to true.
     */
    public static final int IMMERSIVE = 1 << 2;

    /**
     * Apply a fade-in/fade-out animation to the advertisement activity
     */
    public static final int TRANSITION_ANIMATE = 1 << 3;

    /**
     * When set, the advertisement will auto-rotate to match the device's orientation. If not set,
     * the advertisement will render according to the video dimensions and will not auto-rotate.
     */
    public static final int AUTO_ROTATE = 1 << 4;

    /**
     * When set, the advertisement will support apk direct download feature.
     */
    public static final int APK_DIRECT_DOWNLOAD = 1 << 5;

    /**
     * When set the SDK will check if it has an Ad that can be rendered in the specific size
     */
    private AdSize adSize = VUNGLE_DEFAULT;

    public enum AdSize {
        /** setting this has not effect for {@link Vungle#getNativeAd(String, AdConfig, PlayAdCallback)} **/
        VUNGLE_FULLSCREEN,
        /** should be used only along with {@link Vungle#getNativeAd(String, AdConfig, PlayAdCallback)} for IAB standard MREC size **/
        VUNGLE_MREC,
        /** should be used only along with {@link Vungle#getNativeAd(String, AdConfig, PlayAdCallback)} for generic FlexFeed **/
        VUNGLE_DEFAULT
    }

    /**
     * Sets whether
     * <a href="https://developer.android.com/training/system-ui/immersive.html" target="_blank">immersive mode</a>
     * will be enabled for KitKat (API level 19) and later devices.
     *
     * @param isImmersiveMode <code>true</code> if immersive mode will be
     *                        enabled for KitKat+ devices; otherwise, <code>false</code>.
     */
    public void setImmersiveMode(boolean isImmersiveMode) {
        if(isImmersiveMode) {
            settings = settings | IMMERSIVE;
        } else {
            settings = settings & ~IMMERSIVE;
        }
    }

    /**
     *
     * @return The number of advertisements set by the developer that have played before this ad
     */
    public int getOrdinal() {
        return ordinal;
    }

    /**
     * Inlcuding this advertisement, the number of advertisements that have been played
     * this session. This is not tracked internally because apps can have multiple ad
     * providers.
     * @param ordinal Count in integers
     */
    public void setOrdinal(int ordinal) {
        this.ordinal = ordinal;
    }

    /**
     * <p>Sets whether the Android back button will be immediately enabled during the
     * video ad, or it will be inactive until the on screen close button appears
     * <i>(the default)</i>.</p>
     * <p>
     * <p>Once enabled, the Android back button allows the user to skip the video
     * ad and proceed to the post-roll if one exists; if the ad does not have a
     * post-roll the ad simply ends.</p>
     *
     * @param isBackButtonImmediatelyEnabled <code>true</code> if back button should
     *                                       be enabled before ad close button appears; otherwise, <code>false</code>
     */
    public void setBackButtonImmediatelyEnabled(boolean isBackButtonImmediatelyEnabled) {
        if(isBackButtonImmediatelyEnabled) {
            settings = settings | IMMEDIATE_BACK;
        } else {
            settings = settings & ~IMMEDIATE_BACK;
        }
    }

    /**
     * Sets the orientation of the video ad.
     *
     * @param  autoRotate if the video ad is supposed to auto-rotate or stick to the setting on Ad
     */
    public void setAutoRotate(boolean autoRotate) {
        if(autoRotate) {
            settings = settings | AUTO_ROTATE;
        } else {
            settings = settings & ~AUTO_ROTATE;
        }
    }

    /**
     * <p>Sets whether the video should start with its sound matching the
     * settings of your enclosing application <i>(the default</i> or muted.</p>
     * <p>
     * <p>Note that the user can still manually mute or unmute the sound during
     * playback.</p>
     *
     * @param isMuted <code>true</code> if the video should start with its
     *                       audio settings matching those of your enclosing application;
     *                       <code>false</code> if it should start muted.
     */
    public void setMuted(boolean isMuted) {
        if(isMuted) {
            settings = settings | MUTED;
        } else {
            settings = settings & ~MUTED;
        }
    }

    /**
     * <p>Sets whether the video transition animation should be enabled or disabled.
     * The default is enabled (a fade-in/fade-out animation).</p>
     *
     * @param isTransitionAnimationEnabled <code>true</code> if the
     *                                     video transition animation should be enabled;
     *                                     <code>false</code> if it should be disabled.
     */
    public void setTransitionAnimationEnabled(boolean isTransitionAnimationEnabled) {
        if(isTransitionAnimationEnabled) {
            settings = settings | TRANSITION_ANIMATE;
        } else {
            settings = settings & ~TRANSITION_ANIMATE;
        }
    }

    /**
     * Sets how long a flexview ad is visible for before it closes automatically. Default value of 0 (No time limit).
     * @param seconds Value in seconds
     */
    public void setFlexViewCloseTime(int seconds){
        flexviewCloseSec = seconds;
    }

    /**
     * Gets how long a flexview ad is visible for before it closes automatically. Default value of 0 (No time limit).
     * @return Value in seconds
     */
    public int getFlexViewCloseTime(){
        return flexviewCloseSec;
    }


    /**
     * Sets additional flags for ad.
     * @param extras additional configs of ad. Currently we have direct_download flag.
     */
    public void setExtraSettings(Map<String, Object> extras) {
        if (extras == null || extras.isEmpty()) {
            return;
        }

        if (extras.containsKey(FLAG_DIRECT_DOWNLOAD)) {
            Boolean enabled = (Boolean)extras.get(FLAG_DIRECT_DOWNLOAD);
            if (enabled) {
                settings = settings | APK_DIRECT_DOWNLOAD;
            } else {
                settings = settings & ~APK_DIRECT_DOWNLOAD;
            }
        }

    }

    /**
     * Returns AdSize
     * @return
     */
    public AdSize getAdSize() {
        return adSize;
    }

    /**
     * Sets the AdSize, currently only supported for FlexFeed and MREC ad units
     * should be called immediately before getNativeAd
     * @param adSize
     */
    public void setAdSize(AdSize adSize) {
        this.adSize = adSize;
    }

}
