package com.vungle.warren;

import android.support.annotation.IntDef;
import android.support.annotation.NonNull;
import android.support.annotation.Nullable;
import android.support.annotation.VisibleForTesting;
import android.text.TextUtils;
import android.util.Log;
import android.webkit.URLUtil;

import com.google.gson.JsonArray;
import com.google.gson.JsonObject;
import com.vungle.warren.downloader.AssetDownloadListener;
import com.vungle.warren.downloader.AssetDownloadListener.DownloadError.ErrorReason;
import com.vungle.warren.downloader.AssetDownloader;
import com.vungle.warren.downloader.DownloadRequest;
import com.vungle.warren.error.VungleError;
import com.vungle.warren.error.VungleException;
import com.vungle.warren.model.AdAsset;
import com.vungle.warren.model.AdAsset.FileType;
import com.vungle.warren.model.AdAsset.Status;
import com.vungle.warren.model.Advertisement;
import com.vungle.warren.model.Placement;
import com.vungle.warren.persistence.CacheManager;
import com.vungle.warren.persistence.DatabaseHelper;
import com.vungle.warren.persistence.Repository;
import com.vungle.warren.tasks.DownloadJob;
import com.vungle.warren.tasks.JobRunner;
import com.vungle.warren.ui.HackMraid;
import com.vungle.warren.utility.FileUtility;
import com.vungle.warren.utility.SDKExecutors;
import com.vungle.warren.utility.UnzipUtility;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import retrofit2.Call;
import retrofit2.Callback;
import retrofit2.Response;

import static com.vungle.warren.error.VungleException.AD_FAILED_TO_DOWNLOAD;
import static com.vungle.warren.error.VungleException.ASSET_DOWNLOAD_RECOVERABLE;
import static com.vungle.warren.error.VungleException.DB_ERROR;
import static com.vungle.warren.error.VungleException.DIRECT_DOWNLOAD_VALIDATION_ERROR;
import static com.vungle.warren.error.VungleException.NETWORK_ERROR;
import static com.vungle.warren.error.VungleException.NO_SERVE;
import static com.vungle.warren.error.VungleException.OPERATION_CANCELED;
import static com.vungle.warren.error.VungleException.SERVER_RETRY_ERROR;
import static com.vungle.warren.error.VungleException.SERVER_TEMPORARY_UNAVAILABLE;
import static com.vungle.warren.error.VungleException.UNKNOWN_ERROR;
import static com.vungle.warren.error.VungleException.VUNGLE_NOT_INTIALIZED;
import static com.vungle.warren.model.AdAsset.Status.DOWNLOAD_SUCCESS;
import static com.vungle.warren.model.AdAsset.Status.PROCESSED;
import static com.vungle.warren.model.Advertisement.ERROR;
import static com.vungle.warren.model.Advertisement.NEW;
import static com.vungle.warren.model.Advertisement.READY;
import static com.vungle.warren.model.Advertisement.TYPE_VUNGLE_MRAID;

/**
 * The AdLoader is responsible for loading Ad and it's assets.
 */
public class AdLoader {

    private final static String TAG = AdLoader.class.getCanonicalName();
    public static final long EXPONENTIAL_RATE = 2;
    public static final long RETRY_DELAY = 2 * 1000;
    public static final int RETRY_COUNT = 5;

    /**
     * Simple callback interface for downloading.
     */
    interface DownloadCallback {
        /**
         * Callback handler, triggered when the download has completed.
         */
        void onDownloadCompleted(@NonNull String placementId, @NonNull String advertisementId);

        /**
         * Callback handler, triggered when an error has occurred while downloading advertisement
         * assets.
         *
         * @param throwable Throwable describing the error.
         */
        void onDownloadFailed(@NonNull Throwable throwable, @Nullable String placementId, @Nullable String advertisementId);

        void onReady(@NonNull String id, @NonNull Placement placement, @NonNull Advertisement advertisement);
    }

    @IntDef(value = {ReschedulePolicy.EXPONENTIAL, ReschedulePolicy.EXPONENTIAL_ENDLESS_AD})
    @Retention(RetentionPolicy.SOURCE)
    public @interface ReschedulePolicy {
        int EXPONENTIAL = 0;
        /**
         * Try to download assets {@link AdLoader#RETRY_COUNT} times.
         */
        int EXPONENTIAL_ENDLESS_AD = 1;
    }

    static class Operation {
        final String id;
        final long delay;
        final long retryDelay;
        final int retry;
        final int retryLimit;
        @ReschedulePolicy
        final int policy;
        @NonNull
        final Set<LoadAdCallback> loadAdCallbacks = new CopyOnWriteArraySet<>();
        @Nullable
        final PublisherDirectDownload directDownload;
        @NonNull
        final AtomicBoolean loading;
        boolean cancelable;

        public Operation(String id, long delay, long retryDelay, int retryLimit, @ReschedulePolicy int policy, @Nullable PublisherDirectDownload directDownload, int retry, boolean cancelable, @Nullable LoadAdCallback... loadAdCallbacks) {
            this.id = id;
            this.delay = delay;
            this.retryDelay = retryDelay;
            this.retryLimit = retryLimit;
            this.policy = policy;
            this.directDownload = directDownload;
            this.retry = retry;
            this.loading = new AtomicBoolean();
            this.cancelable = cancelable;
            if (loadAdCallbacks != null) {
                this.loadAdCallbacks.addAll(Arrays.asList(loadAdCallbacks));
            }
        }

        Operation delay(long delay) {
            return new Operation(id, delay, retryDelay, retryLimit, policy, directDownload, retry, cancelable, loadAdCallbacks.toArray(new LoadAdCallback[0]));
        }

        Operation retryDelay(long retryDelay) {
            return new Operation(id, delay, retryDelay, retryLimit, policy, directDownload, retry, cancelable, loadAdCallbacks.toArray(new LoadAdCallback[0]));
        }

        Operation retry(int retry) {
            return new Operation(id, delay, retryDelay, retryLimit, policy, directDownload, retry, cancelable, loadAdCallbacks.toArray(new LoadAdCallback[0]));
        }
    }

    /**
     * A state-tracking field which contains information about ongoing load operations. There can
     * only ever be a single load operation per active placement.
     */
    private final Map<String, Operation> loadOperations = new ConcurrentHashMap<>();
    private final Map<String, Operation> pendingOperations = new ConcurrentHashMap<>();

    //fixed values
    @NonNull
    private final Repository repository;
    @NonNull
    private final SDKExecutors sdkExecutors;
    @NonNull
    private final VungleApiClient vungleApiClient;
    @NonNull
    private final CacheManager cacheManager;
    @NonNull
    private final AssetDownloader downloader;
    @NonNull
    private final RuntimeValues runtimeValues;
    @Nullable
    private JobRunner jobRunner;
    @NonNull
    private final VungleStaticApi vungleApi;

    public void setJobRunner(@NonNull JobRunner jobRunner) {
        this.jobRunner = jobRunner;
    }

    public AdLoader(
            @NonNull SDKExecutors sdkExecutors,
            @NonNull Repository repository,
            @NonNull VungleApiClient vungleApiClient,
            @NonNull CacheManager cacheManager,
            @NonNull AssetDownloader downloader,
            @NonNull RuntimeValues runtimeValues,
            @NonNull VungleStaticApi vungleApi
    ) {
        this.sdkExecutors = sdkExecutors;
        this.repository = repository;
        this.vungleApiClient = vungleApiClient;
        this.cacheManager = cacheManager;
        this.downloader = downloader;
        this.runtimeValues = runtimeValues;
        this.vungleApi = vungleApi;
    }

    private boolean canReDownload(Advertisement advertisement) {

        if (advertisement == null || (advertisement.getState() != NEW && advertisement.getState() != READY)) {
            return false;
        }

        List<AdAsset> adAssets = repository.loadAllAdAssets(advertisement.getId()).get();
        if (adAssets.size() == 0) return false;

        for (AdAsset asset : adAssets) {
            //Discard ad if unzipped asset is missing
            if (asset.fileType == FileType.ZIP_ASSET) {
                File file = new File(asset.localPath);
                if (!fileIsValid(file, asset))
                    return false;

            } else if (TextUtils.isEmpty(asset.serverPath)) {
                return false;
            }
        }

        return true;
    }

    public boolean canPlayAd(Advertisement advertisement) {
        if (advertisement == null) {
            return false;
        }

        return advertisement.getState() == Advertisement.READY && hasAssetsFor(advertisement.getId());
    }

    public synchronized void clear() {
        Set<String> ids = new HashSet<>();
        ids.addAll(loadOperations.keySet());
        ids.addAll(pendingOperations.keySet());
        for (String id : ids) {
            onError(loadOperations.remove(id), VungleException.OPERATION_CANCELED);
            onError(pendingOperations.remove(id), VungleException.OPERATION_CANCELED);
        }
    }

    public boolean isLoading(String id) {
        Operation op = loadOperations.get(id);
        return op != null && op.loading.get();
    }

    private void setLoading(String id, boolean loading) {
        Operation op = loadOperations.get(id);
        if (op != null) {
            op.loading.set(loading);
        }
    }

    /**
     * should be called from {@link DownloadJob} only
     */
    public synchronized void loadPendingInternal(final String id) {
        Operation op = pendingOperations.remove(id);
        if (op == null)
            return;

        load(op.delay(0));
    }

    public synchronized void load(@NonNull Operation op) {
        if (jobRunner == null) {
            onError(op, VUNGLE_NOT_INTIALIZED);
            return;
        }

        @Nullable Operation pending = pendingOperations.remove(op.id);
        if (pending != null) {
            op.loadAdCallbacks.addAll(pending.loadAdCallbacks);
        }

        if (op.delay <= 0) {
            Operation current = loadOperations.get(op.id);
            if (current != null) {
                op.loadAdCallbacks.addAll(current.loadAdCallbacks);
                loadOperations.put(op.id, op);
                setLoading(op.id, true);
            } else {
                loadOperations.put(op.id, op);
                loadAd(op.id, new DownloadCallbackWrapper(sdkExecutors.getVungleExecutor(), new DownloadAdCallback()), op.directDownload);
            }
        } else {
            pendingOperations.put(op.id, op);
            jobRunner.execute(DownloadJob.makeJobInfo(op.id).setDelay(op.delay).setUpdateCurrent(true));
        }
    }

    private void onError(@Nullable Operation op, @VungleException.ExceptionCode int code) {

        if (op != null) {
            for (LoadAdCallback loadAdCallback : op.loadAdCallbacks) {
                if (loadAdCallback != null) {
                    loadAdCallback.onError(op.id, new VungleException(code));
                }
            }
        }
    }

    private Throwable reposeCodeToVungleException(int code) {
        if (recoverableServerCode(code)) {
            return new VungleException(SERVER_TEMPORARY_UNAVAILABLE);
        } else {
            return new VungleException(VungleException.SERVER_ERROR);
        }
    }

    private boolean recoverableServerCode(int code) {
        return code == 408 || (500 <= code && code < 600);
    }

    private Throwable retrofitToVungleException(Throwable throwable) {
        if (throwable instanceof IllegalArgumentException) {
            return new VungleError(VungleError.INVALID_URL);
        } else if (throwable instanceof UnknownHostException) {
            return throwable;
        } else if (throwable instanceof IOException) {
            return new VungleException(NETWORK_ERROR);
        } else {
            return throwable;
        }
    }

    /**
     * Request the Vungle SDK to load the assets for an advertisement by the placement identifier.
     * This will cause us to request an ad from the Ad Server, and if the bid is filled, the assets
     * will be downloaded. It is possible for the bid to get no responses and therefore no assets
     * will be loaded. In this case, the SDK will automatically retryCount at a later time, specified by
     * the Vungle Server. The callback will be notified that the assets are pending download.
     *
     * @param id                 The placement identifier for which assets should be loaded.
     * @param listener           The optional callback which will be notified when the assets are loaded, or
     *                           if they are deferred. Errors will also be sent through this callback.
     * @param pubsDownloadClient The publisher-provided download client interface. This is used to
     *                           verify that the advertisement can be played using the full scope
     */
    private void loadAd(@NonNull final String id, @NonNull final DownloadCallbackWrapper listener, final PublisherDirectDownload pubsDownloadClient) {
        sdkExecutors.getVungleExecutor().execute(new Runnable() {
            @Override
            public void run() {
                if (!vungleApi.isInitialized()) {
                    listener.onDownloadFailed(new VungleException(VUNGLE_NOT_INTIALIZED), id, null);
                    return;
                }

                Placement placement = repository.load(id, Placement.class).get();
                if (placement == null) {
                    listener.onDownloadFailed(
                            new IllegalArgumentException("Placement ID " + id + " is not valid. Please check your configuration on the vungle dashboard."),
                            id,
                            null
                    );
                    return;
                }

                final Advertisement advertisement = repository.findValidAdvertisementForPlacement(placement.getId()).get();

                /// If the assets are already loaded and have not expired, do not download again.
                if (canPlayAd(advertisement)) {
                    listener.onReady(id, placement, advertisement);
                } else if (canReDownload(advertisement)) {
                    Log.d(TAG, "Found valid adv but not ready - downloading content");

                    if (cacheManager.getBytesAvailable() < runtimeValues.settings.getMinimumSpaceForAd()) {
                        if (advertisement.getState() != ERROR) {
                            try {
                                repository.saveAndApplyState(advertisement, id, ERROR);
                            } catch (DatabaseHelper.DBException e) {
                                listener.onDownloadFailed(new VungleException(VungleException.DB_ERROR), id, null);
                                return;
                            }
                        }
                        listener.onDownloadFailed(new VungleException(VungleException.NO_SPACE_TO_DOWNLOAD_ASSETS), id, null);
                        return;
                    }

                    /// Update the instance state
                    setLoading(id, true);

                    if (advertisement.getState() != NEW) {
                        try {
                            repository.saveAndApplyState(advertisement, id, NEW);
                        } catch (DatabaseHelper.DBException e) {
                            listener.onDownloadFailed(new VungleException(VungleException.DB_ERROR), id, null);
                            return;
                        }
                    }

                    downloadAdContent(
                            id,
                            advertisement,
                            pubsDownloadClient,
                            listener
                    );


                } else if (placement.getWakeupTime() > System.currentTimeMillis()) {
                    listener.onDownloadFailed(new VungleException(NO_SERVE), id, null);
                    Log.w(TAG, "Placement " + placement.getId() + " is  snoozed");

                    //rescheduling download to time when placement is active
                    if (placement.isAutoCached()) {
                        Log.d(TAG, "Placement " + placement.getId() + " is sleeping rescheduling it ");
                        loadEndless(placement.getId(), placement.getWakeupTime() - System.currentTimeMillis(), pubsDownloadClient);
                    }

                } else {
                    Log.i(TAG, "didn't find cached adv for " + id + " downloading ");

                    if (advertisement != null) {
                        try {
                            repository.saveAndApplyState(advertisement, id, ERROR);
                        } catch (DatabaseHelper.DBException e) {
                            listener.onDownloadFailed(new VungleException(VungleException.DB_ERROR), id, null);
                            return;
                        }
                    }

                    /// Download the ad
                    if (runtimeValues.settings != null && cacheManager.getBytesAvailable() < runtimeValues.settings.getMinimumSpaceForAd()) {
                        listener.onDownloadFailed(new VungleException((placement == null || !placement.isAutoCached())
                                ? VungleException.NO_SPACE_TO_LOAD_AD : VungleException.NO_SPACE_TO_LOAD_AD_AUTO_CACHED), id, null);
                        return;
                    }

                    //Clear placement data as no valid adv found?
                    Log.d(TAG, "No adv for placement " + placement.getId() + " getting new data ");

                    /// Update the instance state
                    setLoading(id, true);
                    fetchAdMetadata(
                            id,
                            pubsDownloadClient,
                            listener,
                            runtimeValues.headerBiddingCallback
                    );
                }
            }
        });
    }

    /**
     * Multi-step method that downloads the required assets for the given placement identifier. It begins by
     * fetching an advertisement for the given placement, which the ad server provides along with all the
     * metadata and pointers to the downloadable assets. Once this information is ready, we download the
     * individual assets. Once the assets have finished downloading, we inform the callback of the success
     * of the operation.
     *
     * @param id               The placement identifier
     * @param downloadCallback The callback that will notified of the success or failure of the download operation.
     * @param bidTokenCallBack The callback interface to send back header bidding token received in requestAd response.
     */
    private void fetchAdMetadata(final String id, final PublisherDirectDownload pubsDownloadClient,
                                 @NonNull final DownloadCallback downloadCallback,
                                 @Nullable final HeaderBiddingCallback bidTokenCallBack) {
        final boolean isHeaderBiddingEnable = bidTokenCallBack != null;
        vungleApiClient.requestAd(id, isHeaderBiddingEnable).enqueue(new Callback<JsonObject>() {
            @Override
            public void onResponse(@NonNull Call<JsonObject> call, @NonNull final Response<JsonObject> response) {
                //Lets do it in separate thread, since there are few IO calls involved.
                sdkExecutors.getVungleExecutor().execute(new Runnable() {
                    @Override
                    public void run() {
                        /// The Ad Server has given us metadata for the advertisement, we now download the assets
                        if (!response.isSuccessful()) {
                            long retryAfterHeaderValue = vungleApiClient.getRetryAfterHeaderValue(response);
                            Placement tempPlacement = repository.load(id, Placement.class).get();
                            if (retryAfterHeaderValue > 0 && tempPlacement != null && tempPlacement.isAutoCached()) {     //Rescheduling only if Auto cached Ad
                                loadEndless(id, retryAfterHeaderValue);
                                downloadCallback.onDownloadFailed(new VungleException(SERVER_RETRY_ERROR), id, null);
                                return;
                            }

                            /// We can't do anything if the request failed.
                            Log.e(TAG, "Failed to retrieve advertisement information");
                            downloadCallback.onDownloadFailed(reposeCodeToVungleException(response.code()), id, null);
                            return;
                        }

                        /// Load the placement information from disk. If it cannot be found, an error has
                        /// occurred. Most likely this would be due to a cache clear or db dump.
                        Placement placement = repository.load(id, Placement.class).get();
                        if (placement == null) {
                            Log.e(TAG, "Placement metadata not found for requested advertisement.");
                            downloadCallback.onDownloadFailed(new VungleException(UNKNOWN_ERROR), id, null);
                            return;
                        }

                        JsonObject jsonObject = response.body();
                        if (jsonObject != null && jsonObject.has("ads") && !jsonObject.get("ads").isJsonNull()) {

                            JsonArray ads = jsonObject.getAsJsonArray("ads");

                            if (ads == null || ads.size() == 0) {
                                // If there is no ad available, we need to inform the caller about the no
                                // serve response.
                                downloadCallback.onDownloadFailed(new VungleException(NO_SERVE), id, null);
                                return;
                            }
                            JsonObject ad = ads.get(0).getAsJsonObject();
                            try {
                                final Advertisement advertisement = new Advertisement(ad);

                                if (bidTokenCallBack != null) {
                                    bidTokenCallBack.onBidTokenAvailable(id, advertisement.getBidToken());
                                }

                                //clear data if exists
                                repository.deleteAdvertisement(advertisement.getId());

                                Set<Map.Entry<String, String>> entries = advertisement.getDownloadableUrls().entrySet();
                                final File destinationDir = repository.getAdvertisementAssetDirectory(advertisement.getId()).get();

                                for (Map.Entry<String, String> entry : entries) {
                                    if (URLUtil.isHttpsUrl(entry.getValue()) || URLUtil.isHttpUrl(entry.getValue())) {
                                        String path = destinationDir.getPath() + File.separator + entry.getKey();
                                        String url = entry.getValue();

                                        @FileType int type = path.endsWith(Advertisement.KEY_POSTROLL) || path.endsWith(Advertisement.KEY_TEMPLATE)
                                                ? FileType.ZIP
                                                : FileType.ASSET;

                                        AdAsset adAsset = new AdAsset(advertisement.getId(), url, path);
                                        adAsset.status = Status.NEW;
                                        adAsset.fileType = type;
                                        repository.save(adAsset);
                                    } else {
                                        downloadCallback.onDownloadFailed(new VungleError(VungleError.INVALID_URL), id, advertisement.getId());
                                        return;
                                    }
                                }


                                repository.saveAndApplyState(advertisement, id, Advertisement.NEW);
                                downloadAdContent(id, advertisement, pubsDownloadClient, downloadCallback);

                            } catch (IllegalArgumentException badAd) {
                                /// Failed to parse the advertisement, which means there was no ad served.
                                /// Check for a sleep code and schedule a download job.
                                JsonObject admarkup = ad.getAsJsonObject("ad_markup");
                                if (admarkup.has("sleep")) {
                                    int sleep = admarkup.get("sleep").getAsInt();

                                    /// Set this sleep time on the placement. We should not attempt to load
                                    /// more ads for this placement ID until the time has elapsed.
                                    placement.snooze(sleep);
                                    try {
                                        repository.save(placement);
                                    } catch (DatabaseHelper.DBException ignored) {
                                        downloadCallback.onDownloadFailed(new VungleException(VungleException.DB_ERROR), id, null);
                                        return;
                                    }
                                    if (placement.isAutoCached()) {
                                        /// If the placement is auto-cached, schedule a download as soon as
                                        /// it wakes up. Note that since sleep is in seconds, we need to
                                        /// convert to milliseconds.
                                        loadEndless(id, sleep * 1000);
                                    }
                                }

                                downloadCallback.onDownloadFailed(new VungleException(NO_SERVE), id, null);
                            } catch (DatabaseHelper.DBException ignored) {
                                downloadCallback.onDownloadFailed(new VungleException(VungleException.DB_ERROR), id, null);
                            }
                        } else {
                            /// Error
                            downloadCallback.onDownloadFailed(new VungleError(VungleError.NO_AD_AVAILABLE), id, null);
                        }
                    }
                });
            }

            @Override
            public void onFailure(@NonNull Call<JsonObject> call, @NonNull Throwable throwable) {
                downloadCallback.onDownloadFailed(retrofitToVungleException(throwable), id, null);
            }
        });
    }

    private void downloadAdContent(final String placementId, final Advertisement advertisement, PublisherDirectDownload pubsDownloadClient, @NonNull final DownloadCallback downloadCallback) {
        DownloadStrategy downloadStrategy = null;
        //check for DirectDownload feature - This feature requires that the containing
        // application be able to directly download the advertised app. We use the
        // download strategy parameter passed in by these applications to validate
        // the advertisement before downloading assets for it. If no such strategy
        // has been provided, we continue.
        if (pubsDownloadClient != null && !TextUtils.isEmpty(advertisement.getAdMarketId())) {
            downloadStrategy = new DirectDownloadStrategy(pubsDownloadClient);
        }

        if (downloadStrategy != null) {
            downloadStrategy.isApplicationAvailable(advertisement.getAdMarketId(), new DownloadStrategy.VerificationCallback() {
                @Override
                public void onResult(final boolean result) {
                    sdkExecutors.getVungleExecutor().execute(new Runnable() {
                        @Override
                        public void run() {
                            if (result) {
                                Log.d(TAG, "fetchAdMetadata: downloading assets ");
                                downloadAdAssets(advertisement, downloadCallback, placementId);
                            } else {
                                downloadCallback.onDownloadFailed(new VungleException(DIRECT_DOWNLOAD_VALIDATION_ERROR), placementId, null);
                            }
                        }
                    });
                }
            });
        } else {
            downloadAdAssets(advertisement, downloadCallback, placementId);
        }
    }

    /**
     * @param advertisement    Ad
     * @param downloadCallback callback
     * @param placementId      placement id
     */
    private void downloadAdAssets(final Advertisement advertisement, DownloadCallback downloadCallback, final String placementId) {
        //validating URL , if one or more URLs is empty or not valid, SDK cannot determine how important that asset
        // is or what will be its side-effect. Its best to fail and drop the entire Ad-delivery.
        for (Map.Entry<String, String> entry : advertisement.getDownloadableUrls().entrySet()) {
            if (TextUtils.isEmpty(entry.getKey()) || TextUtils.isEmpty(entry.getValue()) ||
                    !URLUtil.isValidUrl(entry.getValue())) {
                downloadCallback.onDownloadFailed(new VungleException(AD_FAILED_TO_DOWNLOAD), placementId, null);
                Log.e(TAG, "Aborting, Failed to download Ad assets for: " + advertisement.getId());
                return;
            }
        }

        final DownloadCallback callback = new DownloadCallbackWrapper(sdkExecutors.getUIExecutor(), downloadCallback);
        final AtomicInteger downloadCount = new AtomicInteger();

        final List<DownloadRequest> requests = new ArrayList<>();

        AssetDownloadListener downloadListener = getAssetDownloadListener(advertisement, placementId,
                callback, downloadCount);

        /// Kick off the downloads in sequence. The downloader implementation will
        /// decide how to handle concurrent downloads, but this operation will
        /// notify the subscriber when all assets have been downloaded.


        List<AdAsset> assets = repository.loadAllAdAssets(advertisement.getId()).get();

        for (AdAsset asset : assets) {

            if (asset.status == Status.DOWNLOAD_SUCCESS) {

                if (fileIsValid(new File(asset.localPath), asset))
                    continue;


                if (asset.fileType == FileType.ZIP_ASSET) {
                    callback.onDownloadFailed(
                            new VungleException(VungleException.ASSET_DOWNLOAD_ERROR), placementId, advertisement.getId()
                    );
                    return;
                }
            }

            if (asset.status == Status.PROCESSED && asset.fileType == FileType.ZIP) {
                continue;
            }

            if (TextUtils.isEmpty(asset.serverPath)) {
                callback.onDownloadFailed(
                        new VungleException(VungleException.ASSET_DOWNLOAD_ERROR), placementId, advertisement.getId()
                );
                return;
            }

            DownloadRequest downloadRequest = new DownloadRequest(asset.serverPath, asset.localPath, asset.identifier);

            if (asset.status == Status.DOWNLOAD_RUNNING) {
                downloader.cancelAndAwait(downloadRequest, 1000);
            }

            Log.d(TAG, "Starting download for " + asset);
            asset.status = Status.DOWNLOAD_RUNNING;
            try {
                repository.save(asset);
            } catch (DatabaseHelper.DBException ignored) {
                callback.onDownloadFailed(new VungleException(DB_ERROR), placementId, advertisement.getId());
                return;
            }
            requests.add(downloadRequest);
        }

        //All assets were downloaded already
        if (requests.size() == 0) {
            onAssetDownloadFinished(placementId, callback, advertisement, Collections.EMPTY_LIST);
            return;
        }

        downloadCount.set(requests.size());

        for (DownloadRequest downloadRequest : requests) {
            downloader.download(downloadRequest, downloadListener);
        }
    }

    @NonNull
    private AssetDownloadListener getAssetDownloadListener(
            final Advertisement advertisement,
            final String placementId,
            final DownloadCallback callback,
            final AtomicInteger downloadCount) {

        return new AssetDownloadListener() {

            List<DownloadError> errors = Collections.synchronizedList(new ArrayList<DownloadError>());

            @Override
            public void onError(@NonNull final DownloadError downloadError, final @Nullable DownloadRequest downloadRequest) {
                sdkExecutors.getVungleExecutor().execute(new Runnable() {
                    @Override
                    public void run() {
                        Log.e(TAG, "Download Failed");
                        if (downloadRequest != null) {
                            String id = downloadRequest.cookieString;
                            AdAsset asset = TextUtils.isEmpty(id) ? null : repository.load(id, AdAsset.class).get();

                            if (asset != null) {
                                errors.add(downloadError);
                                asset.status = Status.DOWNLOAD_FAILED;
                                try {
                                    repository.save(asset);
                                } catch (DatabaseHelper.DBException e) {
                                    errors.add(new DownloadError(
                                            -1,
                                            new VungleException(VungleException.DB_ERROR),
                                            ErrorReason.INTERNAL_ERROR
                                    ));
                                }
                            } else {
                                errors.add(new DownloadError(
                                        -1,
                                        new IOException("Downloaded file not found!"),
                                        ErrorReason.REQUEST_ERROR
                                ));
                            }

                        } else {
                            errors.add(new DownloadError(-1, new RuntimeException("error in request"), ErrorReason.INTERNAL_ERROR));
                        }

                        if (downloadCount.decrementAndGet() <= 0) {
                            onAssetDownloadFinished(placementId, callback, advertisement, errors);
                        }
                    }
                });
            }

            @Override
            public void onProgress(@NonNull Progress progress, @NonNull DownloadRequest downloadRequest) {

            }

            @Override
            public void onSuccess(@NonNull final File downloadedFile, @NonNull final DownloadRequest downloadRequest) {
                sdkExecutors.getVungleExecutor().execute(new Runnable() {
                    @Override
                    public void run() {
                        String id = downloadRequest.cookieString;
                        AdAsset adAsset = id == null ? null : repository.load(id, AdAsset.class).get();

                        if (!downloadedFile.exists()) {
                            onError(new DownloadError(
                                            -1,
                                            new IOException("Downloaded file not found!"),
                                            ErrorReason.FILE_NOT_FOUND_ERROR
                                    ),
                                    downloadRequest);   //AdAsset table will be updated in onError callback
                            return;
                        }

                        if (adAsset == null) {
                            onError(new DownloadError(
                                            -1,
                                            new IOException("Downloaded file not found!"),
                                            ErrorReason.REQUEST_ERROR
                                    ),
                                    downloadRequest);   //AdAsset table will be updated in onError callback
                            return;
                        }

                        adAsset.fileType = isZip(downloadedFile) ? FileType.ZIP : FileType.ASSET;
                        adAsset.fileSize = downloadedFile.length();
                        adAsset.status = Status.DOWNLOAD_SUCCESS;
                        try {
                            repository.save(adAsset);
                        } catch (DatabaseHelper.DBException e) {
                            onError(new DownloadError(-1, new VungleException(VungleException.DB_ERROR), ErrorReason.INTERNAL_ERROR), downloadRequest);
                            return;
                        }

                        if (downloadCount.decrementAndGet() <= 0) {
                            /// In the case of an MRAID ad, we also have to update the
                            /// cacheable_replacements map to point to the local files.N
                            /// Otherwise, the html will load them from the CDN and could
                            /// cause the user to experience lag.

                            onAssetDownloadFinished(placementId, callback, advertisement, errors);
                        }
                    }
                });
            }
        };
    }

    private boolean isZip(File downloadedFile) {
        return downloadedFile.getName().equals(Advertisement.KEY_POSTROLL) || downloadedFile.getName().equals(Advertisement.KEY_TEMPLATE);
    }

    private void onAssetDownloadFinished(@NonNull final String placementId,
                                         @NonNull final DownloadCallback callback,
                                         @NonNull final Advertisement advertisement,
                                         @NonNull List<AssetDownloadListener.DownloadError> errors) {

        if (errors.isEmpty()) {
            List<AdAsset> assets = repository.loadAllAdAssets(advertisement.getId()).get();

            if (assets.size() == 0) {
                callback.onDownloadFailed(new VungleException(VungleException.ASSET_DOWNLOAD_ERROR), placementId, advertisement.getId());
            }

            for (AdAsset asset : assets) {

                if (asset.status == Status.DOWNLOAD_SUCCESS) {
                    File f = new File(asset.localPath);

                    if (!fileIsValid(f, asset)) {
                        callback.onDownloadFailed(new VungleException(VungleException.ASSET_DOWNLOAD_ERROR), placementId, advertisement.getId());
                        return;
                    }

                    if (asset.fileType == FileType.ZIP) {
                        try {
                            unzipFile(advertisement, asset, f, assets);
                        } catch (IOException | DatabaseHelper.DBException e) {
                            callback.onDownloadFailed(new VungleException(VungleException.ASSET_DOWNLOAD_ERROR), placementId, advertisement.getId());
                            return;
                        }
                    }
                } else if (asset.fileType == FileType.ZIP && asset.status != PROCESSED) {
                    callback.onDownloadFailed(new VungleException(VungleException.ASSET_DOWNLOAD_ERROR), placementId, advertisement.getId());
                    return;
                }
            }

            if (advertisement.getAdType() == TYPE_VUNGLE_MRAID) {
                File destinationDir = repository.getAdvertisementAssetDirectory(advertisement.getId()).get();
                Log.d(TAG, "saving MRAID for " + advertisement.getId());
                advertisement.setMraidAssetDir(destinationDir);
                try {
                    repository.save(advertisement);
                } catch (DatabaseHelper.DBException e) {
                    callback.onDownloadFailed(new VungleException(VungleException.DB_ERROR), placementId, advertisement.getId());
                    return;
                }
            }

            callback.onDownloadCompleted(placementId, advertisement.getId());
        } else {

            VungleException endError = null;

            for (AssetDownloadListener.DownloadError downloadError : errors) {
                VungleException error;

                if (VungleException.getExceptionCode(downloadError.cause) == DB_ERROR) {
                    endError = new VungleException(VungleException.DB_ERROR);
                    break;
                }

                if (recoverableServerCode(downloadError.serverCode) && downloadError.reason == ErrorReason.REQUEST_ERROR) {
                    error = new VungleException(VungleException.ASSET_DOWNLOAD_RECOVERABLE);
                } else if (downloadError.reason == ErrorReason.CONNECTION_ERROR) {
                    error = new VungleException(VungleException.ASSET_DOWNLOAD_RECOVERABLE);
                } else {
                    error = new VungleException(VungleException.ASSET_DOWNLOAD_ERROR);
                }

                endError = error;

                if (endError.getExceptionCode() == VungleException.ASSET_DOWNLOAD_ERROR)
                    break;
            }

            final VungleException exception = endError;
            callback.onDownloadFailed(exception, placementId, advertisement.getId());
        }
    }

    private class DownloadAdCallback implements DownloadCallback {

        @Override
        public void onDownloadCompleted(@NonNull final String placementId, @NonNull final String advertisementId) {
            Log.d(TAG, "download completed " + placementId);

            final Advertisement advertisement = TextUtils.isEmpty(advertisementId)
                    ? null
                    : repository.load(advertisementId, Advertisement.class).get();

            final Placement placement = repository.load(placementId, Placement.class).get();
            if (advertisement == null || placement == null) {
                onDownloadFailed(new IllegalStateException("Didn't find adv"), placementId, advertisementId);
                return;
            }
            try {
                repository.saveAndApplyState(advertisement, placementId, READY);
            } catch (DatabaseHelper.DBException e) {
                onDownloadFailed(new VungleException(DB_ERROR), placementId, advertisementId);
                return;
            }
            onReady(placementId, placement, advertisement);
        }


        @Override
        public void onReady(@NonNull final String id, @NonNull final Placement placement, @NonNull final Advertisement advertisement) {
            synchronized (AdLoader.this) {
                setLoading(id, false);

                //for ads that are already cached and SDK receives loadAd it should also fire callback with bid token
                HeaderBiddingCallback headerBiddingCallback = runtimeValues.headerBiddingCallback;
                if (headerBiddingCallback != null) {
                    headerBiddingCallback.adAvailableForBidToken(id, advertisement.getBidToken());
                }

                Log.i(TAG, "found already cached valid adv, calling onAdLoad " + id + " callback ");
                /// We have assets for this placement. Verify that the advertisement link is still valid.
                if (placement.isAutoCached() && runtimeValues.initCallback != null) {
                    runtimeValues.initCallback.onAutoCacheAdAvailable(id);
                }

                Operation operation = loadOperations.remove(id);

                if (operation != null) {
                    for (LoadAdCallback loadAdCallback : operation.loadAdCallbacks) {
                        if (loadAdCallback != null) {
                            loadAdCallback.onAdLoad(id);
                        }
                    }
                }
            }
        }

        @Override
        public void onDownloadFailed(@NonNull Throwable throwable, final String placementId, final String advertisementId) {
            synchronized (AdLoader.this) {
                final Operation op = loadOperations.remove(placementId);
                Placement placement = repository.load(placementId, Placement.class).get();
                Advertisement advertisement = advertisementId == null
                        ? null
                        : repository.load(advertisementId, Advertisement.class).get();

                if (placement == null) {
                    if (advertisement != null) {
                        try {
                            repository.saveAndApplyState(advertisement, placementId, ERROR);
                        } catch (DatabaseHelper.DBException ignored) {
                            throwable = new VungleException(DB_ERROR);
                        }
                    }

                    for (LoadAdCallback loadAdCallback : op.loadAdCallbacks) {
                        loadAdCallback.onError(placementId, throwable);
                    }
                    setLoading(placementId, false);
                    return;
                }


                boolean canRetry = false;
                boolean stopInfinite = false;
                @Advertisement.State int state = ERROR;
                if (throwable instanceof VungleException) {
                    switch (((VungleException) throwable).getExceptionCode()) {
                        case SERVER_TEMPORARY_UNAVAILABLE:
                        case NETWORK_ERROR:
                            canRetry = true;
                            break;
                        case ASSET_DOWNLOAD_RECOVERABLE:
                            if (advertisement != null) {
                                canRetry = true;
                                state = NEW;
                            }
                            break;
                        case OPERATION_CANCELED:
                        case NO_SERVE:
                        case SERVER_RETRY_ERROR:
                            stopInfinite = true;
                            break;
                        default:
                            break;
                    }
                }

                Log.e("Vungle", "Failed to load ad/assets for " + placementId + ". Cause:", throwable);
                setLoading(placementId, false);

                if (op != null) {
                    switch (op.policy) {
                        case ReschedulePolicy.EXPONENTIAL:
                            if (op.retry < op.retryLimit && canRetry) {
                                if (advertisement != null) {
                                    try {
                                        repository.saveAndApplyState(advertisement, placementId, state);
                                    } catch (DatabaseHelper.DBException e) {
                                        throwable = new VungleException(DB_ERROR);
                                        break;
                                    }
                                }
                                load(op.delay(op.retryDelay).retryDelay(op.retryDelay * EXPONENTIAL_RATE).retry(op.retry + 1));
                                return;
                            }
                            break;
                        case ReschedulePolicy.EXPONENTIAL_ENDLESS_AD:
                            if (!stopInfinite) {
                                int retry = op.retry;
                                if (retry < op.retryLimit && canRetry) {
                                    retry++;
                                } else {
                                    retry = 0;//forever request /ads until stop
                                    state = ERROR;
                                }
                                if (advertisement != null) {
                                    try {
                                        repository.saveAndApplyState(advertisement, placementId, state);
                                    } catch (DatabaseHelper.DBException e) {
                                        throwable = new VungleException(DB_ERROR);
                                        break;
                                    }
                                }
                                load(op.delay(op.retryDelay).retryDelay(op.retryDelay * EXPONENTIAL_RATE).retry(retry));
                                return;
                            }
                            break;
                    }
                    if (advertisement != null) {
                        try {
                            repository.saveAndApplyState(advertisement, placementId, ERROR);
                        } catch (DatabaseHelper.DBException e) {
                            throwable = new VungleException(DB_ERROR);
                        }
                    }
                    for (LoadAdCallback loadAdCallback : op.loadAdCallbacks) {
                        if (loadAdCallback != null) {
                            loadAdCallback.onError(placementId, throwable);
                        }
                    }
                }
            }
        }
    }

    //helper functions
    public void load(String id, LoadAdCallback listener, PublisherDirectDownload pubsDownloadClient) {
        load(new Operation(
                id,
                0,
                RETRY_DELAY,
                RETRY_COUNT,
                ReschedulePolicy.EXPONENTIAL,
                pubsDownloadClient,
                0,
                //don't allow second request with same id while it's retrying
                false,
                listener
        ));
    }

    public void loadEndless(final String id, final long delay, final PublisherDirectDownload pubsDownloadClient) {
        load(new Operation(
                id,
                delay,
                RETRY_DELAY,
                RETRY_COUNT,
                ReschedulePolicy.EXPONENTIAL_ENDLESS_AD,
                pubsDownloadClient,
                0,
                true
        ));
    }

    public void loadEndless(final String id, final long delay) {
        loadEndless(id, delay, runtimeValues.publisherDirectDownload);
    }

    private void unzipFile(Advertisement advertisement,
                           AdAsset zipAsset,
                           @NonNull final File downloadedFile,
                           List<AdAsset> allAssets) throws IOException, DatabaseHelper.DBException {
        File destDir = repository.getAdvertisementAssetDirectory(advertisement.getId()).get();

        List<File> extractedFiles;

        final List<String> existingPaths = new ArrayList<>();
        for (AdAsset asset : allAssets) {
            if (asset.fileType == FileType.ASSET) {
                existingPaths.add(asset.localPath);
            }
        }

        extractedFiles = UnzipUtility.unzip(downloadedFile.getPath(), destDir.getPath(), new UnzipUtility.Filter() {
            @Override
            public boolean matches(String extractPath) {

                File toExtract = new File(extractPath);

                for (String existing : existingPaths) {
                    File existingFile = new File(existing);

                    if (existingFile.equals(toExtract))
                        return false;

                    if (toExtract.getPath().startsWith(existingFile.getPath() + File.separator))
                        return false;
                }

                return true;
            }
        });

        if (downloadedFile.getName().equals(Advertisement.KEY_TEMPLATE)) {
            //  Updating mraid.js
            // Find the mraid.js file and append the MRAID code. Why this is not done by the server is
            // confusing to me but I assume there must be some historical context to it.
            File mraidJS = new File(destDir.getPath() + File.separator + "mraid.js");
            if (mraidJS.exists()) {        //mraid.js exits
                PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(mraidJS, true)));
                out.println(HackMraid.VALUE);
                out.close();
            }
        }


        for (File file : extractedFiles) {
            AdAsset extractedAsset = new AdAsset(advertisement.getId(), null, file.getPath());
            extractedAsset.fileSize = file.length();
            extractedAsset.fileType = FileType.ZIP_ASSET;
            extractedAsset.parentId = zipAsset.identifier;
            extractedAsset.status = Status.DOWNLOAD_SUCCESS;
            repository.save(extractedAsset);
        }

        Log.d(TAG, "Uzipped " + destDir);
        FileUtility.printDirectoryTree(destDir);

        zipAsset.status = Status.PROCESSED;
        repository.save(zipAsset, new Repository.SaveCallback() {
            @Override
            public void onSaved() {
                sdkExecutors.getVungleExecutor().execute(new Runnable() {
                    @Override
                    public void run() {
                        try {
                            FileUtility.delete(downloadedFile);
                        } catch (IOException e) {
                            //ignored
                            Log.e(TAG, "Error on deleting zip assets archive", e);
                        }
                    }
                });
            }

            @Override
            public void onError(Exception ignored) {
            }
        });
    }

    public boolean hasAssetsFor(String adId) throws IllegalStateException {
        List<AdAsset> adAssets = repository.loadAllAdAssets(adId).get();

        if (adAssets == null || adAssets.size() == 0) {
            return false;
        }

        boolean isAllAssetAvailable = true;

        for (AdAsset adAsset : adAssets) {

            if (adAsset.fileType == FileType.ZIP) {
                if (adAsset.status == PROCESSED) {
                    continue;
                }

                isAllAssetAvailable = false;
                break;
            }

            if (adAsset.status != DOWNLOAD_SUCCESS) {
                isAllAssetAvailable = false;
                break;
            }

            File file = new File(adAsset.localPath);
            if (!fileIsValid(file, adAsset)) {
                isAllAssetAvailable = false;
                break;
            }

        }

        return isAllAssetAvailable;
    }

    private boolean fileIsValid(File file, AdAsset adAsset) {
        return file.exists() && file.length() == adAsset.fileSize;
    }

    @VisibleForTesting
    Collection<Operation> getPendingOperations() {
        return pendingOperations.values();
    }

    @VisibleForTesting
    Collection<Operation> getRunningOperations() {
        return loadOperations.values();
    }
}
