package com.vungle.warren;

import android.annotation.SuppressLint;
import android.content.Context;
import android.os.AsyncTask;
import android.os.Bundle;
import android.support.annotation.NonNull;
import android.support.annotation.Nullable;
import android.text.TextUtils;
import android.util.Log;
import android.util.Pair;

import com.vungle.warren.analytics.AdAnalytics;
import com.vungle.warren.analytics.JobDelegateAnalytics;
import com.vungle.warren.analytics.MoatTracker;
import com.vungle.warren.model.Advertisement;
import com.vungle.warren.model.Cookie;
import com.vungle.warren.model.Placement;
import com.vungle.warren.persistence.Repository;
import com.vungle.warren.tasks.JobRunner;
import com.vungle.warren.ui.CloseDelegate;
import com.vungle.warren.ui.JavascriptBridge;
import com.vungle.warren.ui.OrientationDelegate;
import com.vungle.warren.ui.contract.WebAdContract;
import com.vungle.warren.ui.presenter.LocalAdPresenter;
import com.vungle.warren.ui.presenter.MRAIDAdPresenter;
import com.vungle.warren.ui.state.OptionsState;
import com.vungle.warren.ui.view.FullAdWidget;
import com.vungle.warren.ui.view.LocalAdView;
import com.vungle.warren.ui.view.MRAIDAdView;
import com.vungle.warren.ui.view.VungleWebClient;
import com.vungle.warren.utility.HandlerScheduler;
import com.vungle.warren.utility.SDKExecutors;

import java.io.File;

import static com.vungle.warren.ui.contract.AdContract.AdView;
import static com.vungle.warren.ui.contract.AdContract.AdvertisementPresenter;


/**
 * Created by zioye on 1/16/2018.
 */

public class AdvertisementPresentationFactory {

    private static final String EXTRA_ADVERTISEMENT = "ADV_FACTORY_ADVERTISEMENT";
    private static final String TAG = AdvertisementPresentationFactory.class.getSimpleName();
    private final OrientationDelegate orientationDelegate;
    private final CloseDelegate closeDelegate;
    private final Bundle savedState;
    private final JobRunner jobRunner;
    private Advertisement advertisement;
    private VungleApiClient apiClient;
    private AsyncTask<Void, Void, PresentationResultHolder> task;
    private final String placementId;
    private Repository repository;

    public AdvertisementPresentationFactory(String placementId,
                                            Bundle savedInstanceState,
                                            OrientationDelegate orientationDelegate,
                                            CloseDelegate closeDelegate) throws InstantiationException {
        this.repository = Vungle._instance.repository;
        this.apiClient = Vungle._instance.vungleApiClient;
        this.jobRunner = Vungle._instance.jobRunner;
        this.orientationDelegate = orientationDelegate;
        this.closeDelegate = closeDelegate;

        if (TextUtils.isEmpty(placementId)) {
            throw new InstantiationException("Missing placementID! Cannot start advertisement.");
        }

        if (repository == null || !Vungle.isInitialized()) {
            throw new InstantiationException("Vungle SDK is not initialized");
        }

        this.placementId = placementId;
        this.savedState = savedInstanceState;

    }

    private Pair<Advertisement, Placement> loadPresentationData(String placementId, Bundle savedInstanceState) throws Exception {
        Placement placement = repository.load(placementId, Placement.class).get();
        if (placement == null) {
            throw new Exception("No placement for ID " + placementId + " found. Please use a valid placement ID");
        }

        if (savedInstanceState == null) {
            advertisement = repository.findValidAdvertisementForPlacement(placementId).get();
        } else {
            // Restore cached adv, it should be deleted from disk already
            String adId = savedInstanceState.getString(EXTRA_ADVERTISEMENT);
            if (!TextUtils.isEmpty(adId)) {
                advertisement = repository.load(adId, Advertisement.class).get();
            }
        }

        if (advertisement == null) throw new Exception("No ad found");

        File assetDir = repository.getAdvertisementAssetDirectory(advertisement.getId()).get();
        if (!assetDir.isDirectory()) {
            throw new Exception("No asset directory for " + placementId + " exists!");
        }


        return new Pair<>(advertisement, placement);

    }

    /**
     * Factory method for creating an {@link AdvertisementPresenter}. Using the placement ID and the
     * persistor and designer objects, it will create the {@link Advertisement} from the data on disk
     * and determine which renderer is most appropriate for this placement.
     *
     * @param context
     * @param fullAdWidget
     * @return An {@link AdvertisementPresenter} appropriate for the type of advertisement being played for this placement.
     * @throws IllegalArgumentException In case the placement identifier is invalid, or no presenter for the advertisement has been designated.
     */
    @SuppressLint("StaticFieldLeak")
    public void getFullScreenPresentation(@NonNull final Context context,
                                          @NonNull final FullAdWidget fullAdWidget,
                                          @Nullable final OptionsState optionsState,
                                          @NonNull final FullScreenCallback fullscreenCallback) {
        /// Extract placement and advertisement information, validate current SDK state.

        if (task != null) {
            task.cancel(true);
        }

        final SDKExecutors sdkExecutors = Vungle.sdkExecutors;

        task = new AsyncTask<Void, Void, PresentationResultHolder>() {


            @Override
            protected PresentationResultHolder doInBackground(Void... voids) {
                Pair<Advertisement, Placement> data;
                try {
                    data = loadPresentationData(placementId, savedState);
                } catch (Exception e) {
                    return new PresentationResultHolder(e);
                }

                advertisement = data.first;
                Placement placement = data.second;

                if (!Vungle.canPlayAd(advertisement)) {
                    Log.e(TAG, "Advertisement is null or assets are missing");
                    return new PresentationResultHolder(new Exception("Advertisement is null or assets are missing"));
                }

                AdAnalytics adAnalytics = new JobDelegateAnalytics(jobRunner);

                String appId = null;
                Cookie appIdCookie = repository.load(Cookie.APP_ID, Cookie.class).get();
                if (appIdCookie != null && !TextUtils.isEmpty(appIdCookie.getString("appId"))) {
                    appId = appIdCookie.getString("appId");
                }

                VungleWebClient webClient = new VungleWebClient(advertisement, placement);
                File assetDir = repository.getAdvertisementAssetDirectory(advertisement.getId()).get();

                switch (advertisement.getAdType()) {
                    case Advertisement.TYPE_VUNGLE_LOCAL:

                        MoatTracker tracker = MoatTracker.connect(fullAdWidget.videoView, apiClient.getMoatEnabled());

                        LocalAdPresenter localPresenter
                                = new LocalAdPresenter(
                                advertisement,
                                placement,
                                repository,
                                new HandlerScheduler(),
                                adAnalytics,
                                tracker,
                                webClient,
                                optionsState,
                                assetDir,
                                sdkExecutors.getIOExecutor(),
                                sdkExecutors.getUIExecutor()
                        );

                        LocalAdView localView = new LocalAdView(context, fullAdWidget, orientationDelegate, closeDelegate);

                        return new PresentationResultHolder(localView, localPresenter, webClient, tracker, appId);
                    case Advertisement.TYPE_VUNGLE_MRAID:
                        MRAIDAdPresenter mraidPresenter
                                = new MRAIDAdPresenter(
                                advertisement,
                                placement,
                                repository,
                                new HandlerScheduler(),
                                adAnalytics,
                                webClient,
                                null,
                                optionsState,
                                assetDir,
                                sdkExecutors.getIOExecutor(),
                                sdkExecutors.getUIExecutor()
                        );

                        MRAIDAdView mraidView = new MRAIDAdView(context, fullAdWidget, orientationDelegate, closeDelegate);

                        return new PresentationResultHolder(mraidView, mraidPresenter, webClient, null, null);

                    default:
                        return new PresentationResultHolder(new Exception("No presenter available for ad type!"));
                }
            }

            @Override
            protected void onPostExecute(PresentationResultHolder result) {
                super.onPostExecute(result);
                if (!isCancelled()) {

                    if (fullscreenCallback == null)
                        return;

                    if (result.ex != null) {
                        Log.e(TAG, "Excpetion on creating presenter", result.ex);
                        fullscreenCallback.onResult(new Pair<AdView, AdvertisementPresenter>(null, null), result.ex);
                        return;
                    }

                    fullAdWidget.linkWebView(result.webClient, new JavascriptBridge(result.advertisementPresenter));

                    if (result.tracker != null) {
                        result.tracker.configure(placementId, advertisement, result.appId);
                    }

                    fullscreenCallback.onResult(new Pair(result.adView, result.advertisementPresenter), result.ex);
                }
            }
        };

        task.execute();
    }

    /**
     * Factory method for creating an {@link AdvertisementPresenter}. Using the placement ID and the
     * persistor and designer objects, it will create the {@link Advertisement} from the data on disk
     * and determine which renderer is most appropriate for this placement.
     *
     * @param downloadAdapter
     * @return An {@link AdvertisementPresenter} appropriate for the type of advertisement being played for this placement.
     * @throws IllegalArgumentException In case the placement identifier is invalid, or no presenter for the advertisement has been designated.
     */
    @SuppressLint("StaticFieldLeak")
    public void getNativeViewPresentation(final DirectDownloadAdapter downloadAdapter,
                                          final ViewCallback viewCallback) {

        final SDKExecutors sdkExecutors = Vungle.sdkExecutors;

        task = new AsyncTask<Void, Void, PresentationResultHolder>() {


            @Override
            protected PresentationResultHolder doInBackground(Void... voids) {


                Pair<Advertisement, Placement> data;
                try {
                    data = loadPresentationData(placementId, savedState);
                } catch (Exception e) {
                    return new PresentationResultHolder(e);
                }

                advertisement = data.first;

                if (advertisement.getAdType() != Advertisement.TYPE_VUNGLE_MRAID) {
                    return new PresentationResultHolder(new IllegalArgumentException("No presenter available for ad type!"));
                }

                Placement placement = data.second;

                if (!Vungle.canPlayAd(advertisement)) {
                    Log.e(TAG, "Advertisement is null or assets are missing");
                    return new PresentationResultHolder(new Exception("Advertisement is null or assets are missing"));
                }


                AdAnalytics adAnalytics = new JobDelegateAnalytics(jobRunner);

                VungleWebClient webClient = new VungleWebClient(advertisement, placement);
                File assetDir = repository.getAdvertisementAssetDirectory(advertisement.getId()).get();


                MRAIDAdPresenter presenter = new MRAIDAdPresenter(
                        advertisement,
                        /*placement*/ placement,
                        repository,
                        new HandlerScheduler(),
                        adAnalytics,
                        webClient,
                        downloadAdapter,
                        null,
                        assetDir,
                        sdkExecutors.getIOExecutor(),
                        sdkExecutors.getUIExecutor());

                return new PresentationResultHolder(null, presenter, webClient, null, null);
            }


            @Override
            protected void onPostExecute(PresentationResultHolder result) {
                super.onPostExecute(result);
                if (!isCancelled()) {

                    if (viewCallback != null) {
                        viewCallback.onResult(new Pair<>((WebAdContract.WebAdPresenter) result.advertisementPresenter, result.webClient), result.ex);
                    }
                }
            }
        };

        task.execute();

    }

    public void saveState(Bundle bundle) {
        bundle.putString(EXTRA_ADVERTISEMENT, advertisement == null ? null : advertisement.getId());
    }


    public void destroy() {
        if (task != null) {
            task.cancel(true);
        }
    }

    public interface FullScreenCallback {
        void onResult(@Nullable Pair<AdView, AdvertisementPresenter> result, @Nullable Exception error);
    }

    public interface ViewCallback {
        void onResult(@Nullable Pair<WebAdContract.WebAdPresenter, VungleWebClient> result, @Nullable Exception error);
    }

    private static class PresentationResultHolder {
        private String appId;
        private AdView adView;
        private AdvertisementPresenter advertisementPresenter;
        private Exception ex;
        private VungleWebClient webClient;
        private MoatTracker tracker;

        PresentationResultHolder(Exception ex) {
            this.ex = ex;
        }

        PresentationResultHolder(AdView adView,
                                 AdvertisementPresenter advertisementPresenter,
                                 VungleWebClient webClient,
                                 MoatTracker tracker,
                                 String appId) {
            this.adView = adView;
            this.advertisementPresenter = advertisementPresenter;
            this.webClient = webClient;
            this.tracker = tracker;
            this.appId = appId;
        }
    }
}
