package com.vungle.warren;

import android.app.Application;
import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.support.annotation.NonNull;
import android.support.annotation.Nullable;
import android.support.v4.content.LocalBroadcastManager;
import android.text.TextUtils;
import android.util.Log;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.vungle.warren.analytics.VungleAnalytics;
import com.vungle.warren.download.APKDirectDownloadManager;
import com.vungle.warren.downloader.AssetDownloader;
import com.vungle.warren.downloader.DownloadRequest;
import com.vungle.warren.error.VungleError;
import com.vungle.warren.error.VungleException;
import com.vungle.warren.model.Advertisement;
import com.vungle.warren.model.Cookie;
import com.vungle.warren.model.JsonUtil;
import com.vungle.warren.model.Placement;
import com.vungle.warren.model.Report;
import com.vungle.warren.persistence.CacheManager;
import com.vungle.warren.persistence.DatabaseHelper;
import com.vungle.warren.persistence.GraphicDesigner;
import com.vungle.warren.persistence.Repository;
import com.vungle.warren.tasks.CleanupJob;
import com.vungle.warren.tasks.JobCreator;
import com.vungle.warren.tasks.JobRunner;
import com.vungle.warren.tasks.ReconfigJob;
import com.vungle.warren.tasks.SendReportsJob;
import com.vungle.warren.tasks.VungleJobCreator;
import com.vungle.warren.tasks.utility.JobRunnerThreadPriorityHelper;
import com.vungle.warren.ui.VungleActivity;
import com.vungle.warren.ui.VungleFlexViewActivity;
import com.vungle.warren.ui.contract.AdContract;
import com.vungle.warren.ui.contract.AdContract.AdvertisementBus;
import com.vungle.warren.ui.contract.AdContract.AdvertisementPresenter;
import com.vungle.warren.ui.view.VungleNativeView;
import com.vungle.warren.utility.NetworkProvider;
import com.vungle.warren.utility.SDKExecutors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;

import retrofit2.Call;
import retrofit2.Callback;
import retrofit2.HttpException;
import retrofit2.Response;

import static com.vungle.warren.error.VungleException.AD_UNABLE_TO_PLAY;
import static com.vungle.warren.error.VungleException.ALREADY_PLAYING_ANOTHER_AD;
import static com.vungle.warren.error.VungleException.APPLICATION_CONTEXT_REQUIRED;
import static com.vungle.warren.error.VungleException.CONFIGURATION_ERROR;
import static com.vungle.warren.error.VungleException.DB_ERROR;
import static com.vungle.warren.error.VungleException.MISSING_REQUIRED_ARGUMENTS_FOR_INIT;
import static com.vungle.warren.error.VungleException.NO_SERVE;
import static com.vungle.warren.error.VungleException.NO_SPACE_TO_INIT;
import static com.vungle.warren.error.VungleException.OPERATION_CANCELED;
import static com.vungle.warren.error.VungleException.OPERATION_ONGOING;
import static com.vungle.warren.error.VungleException.PLACEMENT_NOT_FOUND;
import static com.vungle.warren.error.VungleException.UNKNOWN_ERROR;
import static com.vungle.warren.error.VungleException.VUNGLE_NOT_INTIALIZED;
import static com.vungle.warren.error.VungleException.ZERO_PLACEMENTS;
import static com.vungle.warren.model.Advertisement.DONE;
import static com.vungle.warren.model.Advertisement.ERROR;
import static com.vungle.warren.model.Advertisement.NEW;
import static com.vungle.warren.model.Advertisement.READY;
import static com.vungle.warren.model.Advertisement.VIEWING;
import static java.lang.Boolean.TRUE;

/**
 * Main interface with the Vungle ad network.
 */
public class Vungle {

    private final static String TAG = Vungle.class.getCanonicalName();

    private static final String COM_VUNGLE_SDK = "com.vungle.sdk";

    final static Vungle _instance = new Vungle();

    /**
     * Consent to be maintained if publisher set the consent before calling init.
     * Also to avoid blocking call for {@link Vungle#updateConsentStatus(Consent, String)}
     */
    private volatile Consent consent;

    /**
     * Consent version to be maintained
     */
    private volatile String consentVersion;

    private AdLoader adLoader;

    /**
     * State tracking field which contains true if a current placement is playing an ad.
     */
    private Map<String, Boolean> playOperations = new ConcurrentHashMap<>();

    /**
     * The current application ID. We need to store this in order to reconfigure the SDK as necessary.
     */
    private volatile String appID;

    /**
     * A weak reference to the application context. Once this goes away, we can consider the SDK to be
     * dead, since the host application has lost application context. Holding a strong reference to
     * the context would be a memory leak.
     */
    private Context context;

    /**
     * Downloader instance that we will use to download creative assets.
     */
    private volatile AssetDownloader downloader;

    //Repository
    volatile Repository repository;

    volatile VungleApiClient vungleApiClient;

    static SDKExecutors sdkExecutors = new SDKExecutors();

    /**
     * File cache location manager
     */
    private volatile CacheManager cacheManager;

    private static volatile boolean isInitialized = false;
    private static AtomicBoolean isInitializing = new AtomicBoolean(false);
    private static AtomicBoolean isDepInit = new AtomicBoolean(false);
    volatile JobRunner jobRunner;

    private volatile boolean shouldTransmitIMEI;

    //temporary repository for imei until vungle is initialized
    private volatile String userIMEI;

    private volatile RuntimeValues runtimeValues = new RuntimeValues();
    private static ReconfigJob.ReconfigCall reconfigCall = new ReconfigJob.ReconfigCall() {
        @Override
        public void reConfigVungle() {
            Vungle.reConfigure();
        }
    };

    /**
     * Private no-arg constructor to prevent instance generation of this class.
     */
    private Vungle() {
    }

    /**
     * Initialize the Vungle SDK
     * Please use {@link #init(String, Context, InitCallback)}
     * This method exists for backward compatibility. Placements pass in this method are ignored by the SDK.
     * Placements will be filled in from the dashboard to get a list of Valid Placements, call
     * {@link #getValidPlacements()} after Vungle is initialized successfully
     *
     * @param placements Placements pass here will no longer be passed into SDK. Placements will be passed in from the dashboard.
     * @param appId      The identifier for the publisher application. This can be found in the vungle
     *                   dashboard.
     * @param context    Application context.
     * @param callback   Callback that will be triggered once initialization has completed, or if any errors occur.
     * @throws IllegalArgumentException if InitCallback is null
     */
    @Deprecated
    public static void init(@NonNull final Collection<String> placements,
                            @NonNull String appId,
                            @NonNull Context context,
                            @NonNull InitCallback callback) throws IllegalArgumentException {
        init(appId, context, callback, new VungleSettings.Builder().build());
    }

    /**
     * Initialize the Vungle SDK
     *
     * @param appId    The identifier for the publisher application. This can be found in the vungle
     *                 dashboard.
     * @param context  Application context.
     * @param callback Callback that will be triggered once initialization has completed, or if any errors occur.
     * @throws IllegalArgumentException if InitCallback is null
     */
    public static void init(@NonNull final String appId,
                            @NonNull final Context context,
                            @NonNull final InitCallback callback) throws IllegalArgumentException {
        init(appId, context, callback, new VungleSettings.Builder().build());
    }

    /**
     * Initialize the Vungle SDK
     *
     * @param appId    The identifier for the publisher application. This can be found in the vungle
     *                 dashboard.
     * @param context  Application context.
     * @param callback Callback that will be triggered once initialization has completed, or if any errors occur.
     * @param settings Vungle's settings
     * @throws IllegalArgumentException if InitCallback is null
     */
    public static void init(@NonNull final String appId,
                            @NonNull final Context context,
                            @NonNull final InitCallback callback,
                            @NonNull final VungleSettings settings) throws IllegalArgumentException {

        //String for conforming to Google's versioning logic, DO NOT remove or put this anywhere else
        String VUNGLE_VERSION_STRING = "!SDK-VERSION-STRING!:com.vungle:publisher-sdk-android:" + BuildConfig.VERSION_NAME;

        if (callback == null) {
            throw new IllegalArgumentException("A valid InitCallback required to ensure API calls are being made after initialize is successful");
        }

        _instance.runtimeValues.settings = settings;

        _instance.runtimeValues.initCallback = callback instanceof InitCallbackWrapper
                ? callback
                : new InitCallbackWrapper(sdkExecutors.getUIExecutor(), callback);

        if (context == null || (appId == null || appId.isEmpty())) {
            _instance.runtimeValues.initCallback.onError(new VungleException(MISSING_REQUIRED_ARGUMENTS_FOR_INIT));
            return;
        }

        if (!(context instanceof Application)) {
            _instance.runtimeValues.initCallback.onError(new VungleException(APPLICATION_CONTEXT_REQUIRED));
            return;
        }

        if (isInitialized()) {
            Log.d(TAG, "init already complete");
            _instance.runtimeValues.initCallback.onSuccess();
            return;
        }

        if (isInitializing.getAndSet(true)) {
            Log.d(TAG, "init ongoing");
            _instance.runtimeValues.initCallback.onError(new VungleException(OPERATION_ONGOING));
            return;
        }

        sdkExecutors.getVungleExecutor().execute(new Runnable() {
            @Override
            public void run() {
                if (!isDepInit.getAndSet(true)) {

                    if (_instance.downloader == null) {
                        _instance.downloader = new AssetDownloader(
                                4,
                                NetworkProvider.getInstance(context),
                                sdkExecutors.getUIExecutor()
                        );
                    }
                    if (_instance.cacheManager == null)
                        _instance.cacheManager = new CacheManager(context);
                    if (_instance.runtimeValues.settings != null
                            && _instance.cacheManager.getBytesAvailable() < _instance.runtimeValues.settings.getMinimumSpaceForInit()) {
                        onError(_instance.runtimeValues.initCallback, new VungleException(NO_SPACE_TO_INIT));
                        deInit();
                        return;
                    }
                    _instance.cacheManager.addListener(cacheListener);

                    GraphicDesigner designer = new GraphicDesigner(_instance.cacheManager);

                    if (_instance.repository == null) {
                        _instance.repository = new Repository(context, designer,
                                sdkExecutors.getIOExecutor(), sdkExecutors.getUIExecutor());
                    }

                    /// Save the context reference to track application state.
                    _instance.context = context;
                    _instance.appID = appId;


                    //init first
                    try {
                        _instance.repository.init();
                    } catch (DatabaseHelper.DBException e) {
                        onError(_instance.runtimeValues.initCallback, new VungleException(DB_ERROR));
                        deInit();
                        return;
                    }

                    if (_instance.vungleApiClient == null) {
                        _instance.vungleApiClient = new VungleApiClient(context,
                                appId, _instance.cacheManager, _instance.repository);
                    }

                    if (!TextUtils.isEmpty(_instance.userIMEI)) {
                        _instance.vungleApiClient.updateIMEI(_instance.userIMEI, _instance.shouldTransmitIMEI);
                    }

                    if (_instance.runtimeValues.settings != null) {
                        _instance.vungleApiClient.setDefaultIdFallbackDisabled(
                                _instance.runtimeValues.settings.getAndroidIdOptOut());
                    }

                    if (_instance.adLoader == null) {
                        _instance.adLoader = new AdLoader(sdkExecutors,
                                _instance.repository,
                                _instance.vungleApiClient,
                                _instance.cacheManager,
                                _instance.downloader,
                                _instance.runtimeValues,
                                vungleApi);
                    }

                    if (_instance.jobRunner == null) {
                        JobCreator jobCreator = new VungleJobCreator(
                                _instance.repository,
                                designer,
                                _instance.vungleApiClient,
                                new VungleAnalytics(_instance.vungleApiClient),
                                reconfigCall,
                                _instance.adLoader,
                                vungleApi);

                        /// Initialize the SDK Components
                        _instance.jobRunner = new VungleJobRunner(
                                jobCreator,
                                sdkExecutors.getJobExecutor(),
                                new JobRunnerThreadPriorityHelper()
                        );
                    }

                    _instance.adLoader.setJobRunner(_instance.jobRunner);

                    //Restore the consent if publisher set it before calling init
                    if (_instance.consent != null && !TextUtils.isEmpty(_instance.consentVersion)) {
                        updateConsentStatus(_instance.consent, _instance.consentVersion);
                    } else {
                        //Restore it from storage
                        Cookie gdprConsent = _instance.repository.load(Cookie.CONSENT_COOKIE, Cookie.class).get();
                        _instance.consent = getConsent(gdprConsent);
                        _instance.consentVersion = getConsentMessageVersion(gdprConsent);
                    }

                    Cookie appIdCookie;

                    appIdCookie = _instance.repository.load(Cookie.APP_ID, Cookie.class).get();

                    if (appIdCookie == null) {
                        appIdCookie = new Cookie(Cookie.APP_ID);
                    }

                    appIdCookie.putValue("appId", appId);
                    try {
                        _instance.repository.save(appIdCookie);
                    } catch (DatabaseHelper.DBException e) {
                        onError(_instance.runtimeValues.initCallback, new VungleException(NO_SPACE_TO_INIT));
                        deInit();
                        return;
                    }
                }

                /// Configure the instance.
                _instance.configure(_instance.runtimeValues.initCallback);

            }
        });
    }

    private static void onError(InitCallback initCallback, VungleException e) {
        if (initCallback != null) {
            initCallback.onError(e);
        }
    }

    static void reConfigure() {
        if (isInitialized()) {
            sdkExecutors.getVungleExecutor().execute(new Runnable() {
                @Override
                public void run() {
                    _instance.configure(_instance.runtimeValues.initCallback);
                }
            });
        } else {
            init(_instance.appID, _instance.context, _instance.runtimeValues.initCallback);
        }
    }

    /**
     * Request a configuration from the server. This updates the current list of placements and
     * schedules another configuration job in the future. It also has the side-effect of loading the
     * auto_cached ad if it is not downloaded currently.
     *
     * @param callback Callback that will be called when initialization has completed or failed.
     */
    private void configure(@NonNull final InitCallback callback) {
        /// Request a configuration from the server. This happens asynchronously on the network thread.
        try {
            Response<JsonObject> response = vungleApiClient.config();

            if (response == null) {
                callback.onError(new VungleException(UNKNOWN_ERROR));
                isInitializing.set(false);
                return;
            }

            if (!response.isSuccessful()) {
                long retryAfterHeaderValue = vungleApiClient.getRetryAfterHeaderValue(response);
                if (retryAfterHeaderValue > 0) {
                    jobRunner.execute(ReconfigJob.makeJobInfo(_instance.appID).setDelay(retryAfterHeaderValue));
                    callback.onError(new VungleException(VungleException.SERVER_RETRY_ERROR));
                    isInitializing.set(false);
                    return;
                }

                callback.onError(new VungleException(CONFIGURATION_ERROR));
                isInitializing.set(false);
                return;
            }

            /// If we have not reported this install, fire it off to the ad server
            if (!context.getSharedPreferences("com.vungle.sdk", Context.MODE_PRIVATE).getBoolean("reported", false)) {
                vungleApiClient.reportNew().enqueue(new Callback<JsonObject>() {
                    @Override
                    public void onResponse(Call<JsonObject> call, Response<JsonObject> response) {
                        if (response.isSuccessful()) {
                            /// Save the reported state to shared preferences
                            SharedPreferences.Editor editor = context.getSharedPreferences(COM_VUNGLE_SDK, Context.MODE_PRIVATE).edit();
                            editor.putBoolean("reported", true);
                            editor.apply();
                            Log.d(TAG, "Saving reported state to shared preferences");
                        }
                    }

                    @Override
                    public void onFailure(Call<JsonObject> call, Throwable throwable) {
                        /// Do nothing. Install will be reported on the next init.
                        /// TODO: Should this retryCount more?
                    }
                });
            }

            JsonObject jsonObject = response.body();
            /// Parse out the placements
            JsonArray placementsData = jsonObject.getAsJsonArray("placements");

            /// Check for minimum placement requirements
            if (placementsData.size() == 0) {
                /// In the case that there are no useable placements, we need to notify the
                /// publisher app.
                callback.onError(new VungleException(ZERO_PLACEMENTS));
                isInitializing.set(false);
                return;
            }

            playOperations.clear();
            _instance.adLoader.clear();

            List<Placement> newPlacements = new ArrayList<>();
            for (JsonElement jsonElement : placementsData) {
                newPlacements.add(new Placement(jsonElement.getAsJsonObject()));
            }

            _instance.repository.setValidPlacements(newPlacements);

            if (jsonObject.has("gdpr")) {
                // create cooking regardless of whether user is in CountryDataProtected or not
                Cookie gdprConsent;

                gdprConsent = _instance.repository.load(Cookie.CONSENT_COOKIE, Cookie.class).get();

                if (gdprConsent == null) {
                    // Otherwise, create a new one set default values
                    gdprConsent = new Cookie(Cookie.CONSENT_COOKIE);
                    gdprConsent.putValue("consent_status", "unknown");
                    gdprConsent.putValue("consent_source", "no_interaction");
                    gdprConsent.putValue("timestamp", 0L);
                }

                JsonObject gdprJsonObject = jsonObject.getAsJsonObject("gdpr");

                boolean isCountryDataProtected = JsonUtil.hasNonNull(gdprJsonObject, "is_country_data_protected") && gdprJsonObject.get("is_country_data_protected").getAsBoolean();
                String consentTitle = JsonUtil.hasNonNull(gdprJsonObject, "consent_title") ? gdprJsonObject.get("consent_title").getAsString() : "";
                String consentMessage = JsonUtil.hasNonNull(gdprJsonObject, "consent_message") ? gdprJsonObject.get("consent_message").getAsString() : "";
                String consentMessageVersion = JsonUtil.hasNonNull(gdprJsonObject, "consent_message_version") ? gdprJsonObject.get("consent_message_version").getAsString() : "";
                String acceptText = JsonUtil.hasNonNull(gdprJsonObject, "button_accept") ? gdprJsonObject.get("button_accept").getAsString() : "";
                String denyText = JsonUtil.hasNonNull(gdprJsonObject, "button_deny") ? gdprJsonObject.get("button_deny").getAsString() : "";

                gdprConsent.putValue("is_country_data_protected", isCountryDataProtected);

                gdprConsent.putValue("consent_title", TextUtils.isEmpty(consentTitle) ? "Targeted Ads" : consentTitle);

                gdprConsent.putValue("consent_message", TextUtils.isEmpty(consentMessage) ?
                        "To receive more relevant ad content based on your interactions with our ads, click \"I Consent\" below. Either way, you will see the same amount of ads." : consentMessage);

                // what ever SDK get messageVersion from server, SDK saves and pass
                String strGDPRCurrentSource = gdprConsent.getString("consent_source");
                if (!"publisher".equalsIgnoreCase(strGDPRCurrentSource)) {      //Condition for not to override the publisher version with server version
                    gdprConsent.putValue("consent_message_version", TextUtils.isEmpty(consentMessageVersion) ? "" : consentMessageVersion);
                }

                gdprConsent.putValue("button_accept", TextUtils.isEmpty(acceptText) ? "I Consent" : acceptText);

                gdprConsent.putValue("button_deny", TextUtils.isEmpty(denyText) ? "I Do Not Consent" : denyText);

                _instance.repository.save(gdprConsent);
            }

            ///Direct download
            if (jsonObject.has("apk_direct_download") && jsonObject.getAsJsonObject("apk_direct_download").has("enabled")) {
                JsonObject ddObj = jsonObject.getAsJsonObject("apk_direct_download");
                boolean isDirectDownloadEnabled = ddObj.get("enabled").getAsBoolean();

                if (isDirectDownloadEnabled) {
                    APKDirectDownloadManager.init(context, new AssetDownloader(4, NetworkProvider.getInstance(context), sdkExecutors.getUIExecutor()));
                }
                APKDirectDownloadManager.setDirectDownloadStatus(isDirectDownloadEnabled ?
                        APKDirectDownloadManager.DIRECT_DOWNLOAD_FLAG_ENABLED : APKDirectDownloadManager.DIRECT_DOWNLOAD_FLAG_DISABLED);

            } else {
                APKDirectDownloadManager.init(context, new AssetDownloader(4, NetworkProvider.getInstance(context), sdkExecutors.getUIExecutor()));
                APKDirectDownloadManager.setDirectDownloadStatus(APKDirectDownloadManager.DIRECT_DOWNLOAD_FLAG_NOT_SET); // Not set
            }

            if (jsonObject.has("ri")) {
                Cookie configCookie;

                configCookie = _instance.repository.load(Cookie.CONFIG_COOKIE, Cookie.class).get();

                if (configCookie == null) {
                    configCookie = new Cookie(Cookie.CONFIG_COOKIE);
                }
                boolean isReportIncentivizedEnabled = jsonObject.getAsJsonObject("ri").get("enabled").getAsBoolean();
                configCookie.putValue("isReportIncentivizedEnabled", isReportIncentivizedEnabled);
                _instance.repository.save(configCookie);
            }

            if (jsonObject.has("attribution_reporting")) {
                JsonObject attributionReporting = jsonObject.getAsJsonObject("attribution_reporting");
                if (attributionReporting.has("should_transmit_imei")) {
                    shouldTransmitIMEI = attributionReporting.get("should_transmit_imei").getAsBoolean();
                } else {
                    shouldTransmitIMEI = false;
                }
            } else {
                shouldTransmitIMEI = false;
            }

            /// Schedule the reconfig job
            if (jsonObject.has("config")) {
                long sleep = jsonObject.getAsJsonObject("config").get("refresh_time").getAsLong();
                jobRunner.execute(ReconfigJob.makeJobInfo(appID).setDelay(sleep));
            }

            //this is earliest we think that SDK basic init is success
            isInitialized = true;
            /// Inform the publisher that initialization has succeeded.

            callback.onSuccess();


            isInitializing.set(false);

            Collection<Placement> placements = repository.loadValidPlacements().get();

            /// Clean up the asset and metadata caches
            jobRunner.execute(CleanupJob.makeJobInfo());
            /// Download assets for the auto-cached placement immediately. If assets are already
            /// available, this will do nothing except inform the publisher that the auto-cached
            /// placement is ready.
            /// Download assets for the auto-cached placement immediately. If assets are already
            /// available, this will do nothing except inform the publisher that the auto-cached
            /// placement is ready.
            if (placements != null) {
                for (Placement placement : placements) {
                    if (placement.isAutoCached()) {
                        Log.d(TAG, "starting jobs for autocached advs");
                        adLoader.loadEndless(placement.getId(), 0);
                    }
                }
            }
            /// Send any pending ad reports
            jobRunner.execute(SendReportsJob.makeJobInfo(true));

        } catch (final Throwable throwable) {
            isInitialized = false;
            isInitializing.set(false);
            Log.e(TAG, Log.getStackTraceString(throwable));
            if (throwable instanceof HttpException) {
                callback.onError(new VungleException(CONFIGURATION_ERROR));
            } else if (throwable instanceof DatabaseHelper.DBException) {
                callback.onError(new VungleException(DB_ERROR));
            } else {
                callback.onError(new VungleException(UNKNOWN_ERROR));
            }
        }
    }


    /**
     * Checks if Vungle's SDK is already in initialized state
     *
     * @return true if Vungle SDK is initialized, false otherwise.
     */
    public static boolean isInitialized() {
        return isInitialized &&
                _instance.repository != null &&
                _instance.repository.getValidPlacements() != null &&
                _instance.repository.getValidPlacements().get().size() > 0 &&
                _instance.context != null;
    }

    /**
     * Overrides the previously-set incentivized fields which are used when warning the user if they
     * are attempting to exit an incentivized advertisement prematurely. If there are no local
     * values set, the SDK will use the values set in the Vungle Dashboard as the defaults. This method
     * must be called **before** {@link #playAd(String, AdConfig, PlayAdCallback)} is called.
     *
     * @param title        The dialog box title
     * @param body         The dialog box body text. Values longer than the space allows for will be ellipsised
     *                     at the end.
     * @param keepWatching Continue button text
     * @param close        Close button text.
     */
    public static void setIncentivizedFields(@Nullable final String userID,
                                             @Nullable final String title,
                                             final @Nullable String body,
                                             final @Nullable String keepWatching,
                                             final @Nullable String close) {

        sdkExecutors.getVungleExecutor().execute(new Runnable() {
            @Override
            public void run() {
                if (!isInitialized()) {
                    Log.e(TAG, "Vungle is not initialized");
                    return;
                }

                Cookie incentivizedCookie = _instance.repository.load(Cookie.INCENTIVIZED_TEXT_COOKIE, Cookie.class).get();
                if (incentivizedCookie == null) {
                    incentivizedCookie = new Cookie(Cookie.INCENTIVIZED_TEXT_COOKIE);
                }

                boolean changed = false;

                if (!TextUtils.isEmpty(title)) {
                    changed = true;
                    incentivizedCookie.putValue("title", title);
                }

                if (!TextUtils.isEmpty(body)) {
                    changed = true;
                    incentivizedCookie.putValue("body", body);
                }

                if (!TextUtils.isEmpty(keepWatching)) {
                    changed = true;
                    incentivizedCookie.putValue("continue", keepWatching);
                }

                if (!TextUtils.isEmpty(close)) {
                    changed = true;
                    incentivizedCookie.putValue("close", close);
                }

                if (!TextUtils.isEmpty(userID)) {
                    changed = true;
                    incentivizedCookie.putValue("userID", userID);
                }

                if (changed) {
                    try {
                        _instance.repository.save(incentivizedCookie);
                    } catch (DatabaseHelper.DBException e) {
                        e.printStackTrace();
                    }
                }
            }
        });
    }

    /**
     * Check if we can play an advertisement for the given placement. This method checks out file
     * system to see if there are asset files for the given placement and returns true if we have
     * assets stored which have not expired.
     *
     * @param id The placement identifier.
     * @return true if an advertisement can be played immediately, false otherwise.
     */
    public static boolean canPlayAd(@NonNull String id) {
        if (!isInitialized()) {
            Log.e(TAG, "Vungle is not initialized");
            return false;
        }

        Advertisement advertisement = _instance.repository.findValidAdvertisementForPlacement(id).get();

        return canPlayAd(advertisement);
    }

    static boolean canPlayAd(Advertisement advertisement) {
        return _instance.adLoader.canPlayAd(advertisement);
    }

    /**
     * Play an ad for the given placement ID. If this placement ID is valid and an advertisement is
     * ready to be played, this will cause the {@link VungleActivity} to start and the advertisement
     * will be rendered.
     *
     * @param id       The placement identifier.
     * @param settings Optional settings for playing the advertisement. The full list can be found
     *                 at {@link AdConfig} documentation.
     * @param callback Optional, though strongly encouraged, event listener. This object will be
     *                 notified of the advertisement starting, ending, and any errors that occur
     *                 during rendering.
     */
    public static void playAd(@NonNull final String id, final AdConfig settings, @Nullable final PlayAdCallback callback) {
        if (!isInitialized()) {
            if (callback != null) {
                callback.onError(id, new VungleException(VUNGLE_NOT_INTIALIZED));
            }
            return;
        }

        final PlayAdCallback listener = new PlayAdCallbackWrapper(sdkExecutors.getUIExecutor(), callback);

        sdkExecutors.getVungleExecutor().execute(new Runnable() {
            @Override
            public void run() {

                Placement placement = _instance.repository.load(id, Placement.class).get();

                Exception exception = null;

                //don't allow ad to play when either loading or playing
                if (TRUE.equals(_instance.playOperations.get(id)) || _instance.adLoader.isLoading(id)) {
                    //already playing ad
                    exception = new VungleException(OPERATION_ONGOING);
                }

                if (placement == null) {
                    exception = new VungleException(PLACEMENT_NOT_FOUND);
                }

                if (exception != null) {
                    listener.onError(id, exception);
                    return;
                }

                /// Check if the id is valid
                boolean streamingOnly = false;

                /// Check if an advertisement for this placement exists.
                Advertisement advertisement = _instance.repository.findValidAdvertisementForPlacement(id).get();

                try {
                    if (!canPlayAd(advertisement)) {
                        /// Even if we don't have a cached ad prepared, we can still play a streaming ad if
                        /// one is available.
                        streamingOnly = true;

                        if (advertisement != null && advertisement.getState() == READY) {
                            //assets were somehow deleted
                            _instance.repository.saveAndApplyState(advertisement, id, ERROR);
                            if (placement.isAutoCached()) {
                                _instance.adLoader.loadEndless(placement.getId(), 0);
                            }
                        }
                        listener.onError(id, new VungleException(AD_UNABLE_TO_PLAY));//todo no return?
                    } else {
                        /// The advertisement we have cached is valid and can be used, apply the settings and save them
                        advertisement.configure(settings);
                        _instance.repository.save(advertisement);
                    }
                } catch (DatabaseHelper.DBException e) {
                    listener.onError(id, new VungleException(DB_ERROR));
                    return;
                }


                if (_instance.context != null) {
                    /// Inform the ad server that we are about to plan an ad. If there is a better ad available,
                    /// it will be substituted here.
                    final boolean finalStreamingOnly = streamingOnly;
                    final Advertisement finalAdvertisement = advertisement;
                    _instance.vungleApiClient.willPlayAd(placement.getId(), placement.isAutoCached(), streamingOnly ? "" : advertisement.getAdToken()).enqueue(new Callback<JsonObject>() {
                        @Override
                        public void onResponse(Call<JsonObject> call, final Response<JsonObject> response) {
                            //Lets do it in separate thread, since there are few IO calls involved.
                            sdkExecutors.getVungleExecutor().execute(new Runnable() {
                                @Override
                                public void run() {
                                    Advertisement advReplace = null;
                                    if (response.isSuccessful()) {
                                        JsonObject responseBody = response.body();
                                        if (responseBody != null && responseBody.has("ad")) {
                                            try {
                                                JsonObject adJson = responseBody.getAsJsonObject("ad");
                                                advReplace = new Advertisement(adJson);

                                                /// Update the settings for this advertisement
                                                advReplace.configure(settings);

                                                /// If we make it here, it means that there is a replacement streaming ad for
                                                /// this placement. So we update the metadata and save it so the activity can
                                                /// use the fresh values.
                                                _instance.repository.saveAndApplyState(advReplace, id, NEW);
                                            } catch (IllegalArgumentException illegalArg) {
                                                /// The response did not contain an ad. Do nothing.
                                                Log.v("Vungle", "Will Play Ad did not respond with a replacement. Move on.");
                                            } catch (Exception e) {
                                                Log.e("Vungle", "Error using will_play_ad!", e);
                                            } catch (VungleError vungleError) {
                                                /// Sometimes the server will disable will_play_ad functionality, this is expected.
                                                if (vungleError.getErrorCode() != VungleError.WILL_PLAY_AD_DISABLED) {
                                                    Log.e("Vungle", "Error using will_play_ad!", vungleError);
                                                } else {
                                                    Log.e(TAG, "will_play_ad was disabled by the configuration settings. This is expected.");
                                                }
                                            }

                                        }
                                    }

                                    if (finalStreamingOnly) {
                                        if (advReplace == null) {
                                            /// No streaming ad available, pass back an error
                                            listener.onError(id, new VungleException(NO_SERVE));
                                        } else {
                                            renderAd(id, listener, id, advReplace);
                                        }
                                    } else {
                                        renderAd(id, listener, id, finalAdvertisement);
                                    }
                                }
                            });

                        }

                        @Override
                        public void onFailure(Call<JsonObject> call, Throwable throwable) {
                            //Lets do it in separate thread, since there are few IO calls involved.
                            sdkExecutors.getVungleExecutor().execute(new Runnable() {
                                @Override
                                public void run() {
                                    if (finalStreamingOnly) {
                                        listener.onError(id, new VungleException(NO_SERVE));
                                    } else {
                                        renderAd(id, listener, id, finalAdvertisement);
                                    }
                                }
                            });

                        }
                    });
                }
            }
        });

    }

    /**
     * Private helper method to reduce callback hell, this method starts playing the given advertisement
     * in a {@link VungleActivity}. It also creates an event listener in order to control the flow
     * of events once the advertisement ends.
     *
     * @param id            The placement identifier
     * @param listener      The optional listener for playback events
     * @param placementId   The placement id in which the advertisement is being rendered
     * @param advertisement The advertisement metadata for the advertisement that will be played.
     */
    private static synchronized void renderAd(@NonNull final String id, @Nullable final PlayAdCallback listener, final String placementId, final Advertisement advertisement) {
        /// Subscribe to the event bus of the activity before starting the activity, otherwise
        /// the Publisher notices it has no subscribers and does not emit the start value.

        _instance.playOperations.put(id, true);
        VungleActivity.setEventListener(new AdvertisementPresenter.EventListener() {
            boolean succesfulView = false;
            /**
             * The percentage of the ad that was viewed. We initialize to -1 in order to differentiate
             * between the ad never starting.
             */
            int percentViewed = -1;

            @Override
            public void onNext(String event, String value, String id) {
                try {
                    if (event.equals("start")) {
                        _instance.repository.saveAndApplyState(advertisement, id, VIEWING);

                        if (listener != null) {
                            listener.onAdStart(id);
                        }

                        percentViewed = 0;

                        Placement placement = _instance.repository.load(placementId, Placement.class).get();
                        if (placement != null && placement.isAutoCached()) {
                            /// If the placement is auto-cached, download a new advertisement for it soon.
                            _instance.adLoader.loadEndless(placementId, 0);
                        }
                    } else if (event.equals("end")) {
                        Log.d("Vungle", "Cleaning up metadata and assets for placement " + id + " and advertisement " + advertisement.getId());
                        _instance.repository.saveAndApplyState(advertisement, id, DONE);
                        _instance.repository.updateAndSaveReportState(id, advertisement.getAppID(), Report.NEW, Report.READY);
                        VungleActivity.setEventListener(null);
                        _instance.playOperations.put(id, false);

                        /// Schedule a job to send the report soon
                        _instance.jobRunner.execute(SendReportsJob.makeJobInfo(false));

                        /// Inform the listener about the advertisement ending
                        if (listener != null) {
                            listener.onAdEnd(id,
                                    (succesfulView || percentViewed >= 80),
                                    (value != null && value.equals("isCTAClicked")));

                        }

                    } else if (event.equals("successfulView")) {
                        succesfulView = true;
                    } else if (event.startsWith("percentViewed")) {
                        /// Parse out the value passed in as a String.
                        String[] tokens = event.split(":");
                        if (tokens.length == 2) {
                            String percentString = tokens[1];
                            percentViewed = Integer.parseInt(percentString);
                        }
                    }
                } catch (DatabaseHelper.DBException ignored) {
                    onError(new VungleException(DB_ERROR), id);
                }
            }

            @Override

            public void onError(Throwable throwable, String id) {
                if (VungleException.getExceptionCode(throwable) != ALREADY_PLAYING_ANOTHER_AD &&
                        VungleException.getExceptionCode(throwable) != OPERATION_CANCELED) {
                    try {
                        _instance.repository.saveAndApplyState(advertisement, id, Advertisement.ERROR);
                    } catch (DatabaseHelper.DBException e) {
                        throwable = new VungleException(DB_ERROR);
                    }
                }

                VungleActivity.setEventListener(null);

                _instance.playOperations.put(id, false);

                if (listener != null) {
                    listener.onError(id, throwable);
                }
            }
        });

        /// Start the activity, and if there are any extras that have been overriden by the application,
        /// apply them.
        // Vungleflex is same as VungleActivity, this is all about android rotation issue in 8.0 and translucent attribute
        boolean isFlex = advertisement != null && "flexview".equals(advertisement.getTemplateType());
        Intent intent = new Intent(_instance.context, isFlex ? VungleFlexViewActivity.class : VungleActivity.class);
        intent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
        intent.putExtra(VungleActivity.PLACEMENT_EXTRA, id);

        _instance.context.startActivity(intent);

    }

    /**
     * Request the Vungle SDK to load the assets for an advertisement by the placement identifier.
     * This will cause us to request an ad from the Ad Server, and if the bid is filled, the assets
     * will be downloaded. It is possible for the bid to get no responses and therefore no assets
     * will be loaded. In this case, the SDK will automatically retry at a later time, specified by
     * the Vungle Server. The callback will be notified that the assets are pending download.
     *
     * @param id       The placement identifier for which assets should be loaded.
     * @param callback The optional callback which will be notified when the assets are loaded, or
     *                 if they are deferred. Errors will also be sent through this callback.
     */
    public static void loadAd(@NonNull final String id, @Nullable final LoadAdCallback callback) {
        /// Validate SDK State
        if (!isInitialized()) {
            if (callback != null) {
                callback.onError(id, new VungleException(VUNGLE_NOT_INTIALIZED));
            }
            return;
        }

        final LoadAdCallback listener = new LoadAdCallbackWrapper(sdkExecutors.getUIExecutor(), callback);
        _instance.adLoader.load(id, listener, null);
    }

    /**
     * Clear the internal caches and files. This triggers a config call to happen again.
     * This method is private but used in test app using reflection.
     */
    private static void clearCache() {
        if (isInitialized()) {
            sdkExecutors.getVungleExecutor().execute(new Runnable() {
                @Override
                public void run() {
                    //Clear the persisted data
                    _instance.repository.clearAllData();

                    /// Configure again, which will auto-load the auto-cached placement and hydrate our metadata.
                    _instance.configure(_instance.runtimeValues.initCallback);
                }
            });
        } else {
            Log.e(TAG, "Vungle not initialized");
        }
    }

    /**
     * VungleNativeAd can be used inside any Android view container.
     *
     * @param placementId    - placementId for Native ad Placement configured for the right AdSize
     * @param playAdCallback - Callback to receive onAdStart, onAdEnd, onAdError Callbacks
     * @return VungleNativeAd
     */
    @Deprecated
    @Nullable
    public static VungleNativeAd getNativeAd(String placementId, final PlayAdCallback playAdCallback) {
        return getNativeAd(placementId, null, playAdCallback);
    }

    /**
     * VungleNativeAd can be used inside any Android view container.
     *
     * @param placementId    - placementId for Native ad Placement configured for the right AdSize
     * @param adConfig       - Set the  {@link AdConfig.AdSize} for Native Ad view corresponding to the Placement configuration
     * @param playAdCallback - Callback to receive onAdStart, onAdEnd, onAdError Callbacks
     * @return VungleNativeAd
     */
    @Nullable
    public static VungleNativeAd getNativeAd(String placementId, AdConfig adConfig, final PlayAdCallback playAdCallback) {
        if (!isInitialized()) {
            Log.e(TAG, "Vungle is not initialized, returned VungleNativeAd = null");
            if (playAdCallback != null) {
                playAdCallback.onError(placementId, new VungleException(VUNGLE_NOT_INTIALIZED));
            }
            return null;
        }

        final Placement placement = _instance.repository.load(placementId, Placement.class).get();
        if (placement == null) {
            Log.e(TAG, "No Placement for ID");
            if (playAdCallback != null) {
                playAdCallback.onError(placementId, new VungleException(PLACEMENT_NOT_FOUND));
            }
            return null;
        }

        final Advertisement advertisement = _instance.repository.findValidAdvertisementForPlacement(placementId).get();
        if (advertisement == null) {
            /// The placement does not have an advertisement. This case is checked for before arriving
            /// here, but it's possible that this Activity has been restarted by the OS due to a crash
            /// or developer testing with "Do Not Keep Activities", which causes the Activity to be
            /// recreated *without* a savedInstanceState. In this case, the advertisement has already
            /// been at least partially played, so we should not restart it or a view discrepancy will
            /// occur.
            Log.e(TAG, "No Advertisement for ID");
            if (playAdCallback != null) {
                playAdCallback.onError(placementId, new VungleException(VungleException.AD_UNABLE_TO_PLAY));
            }
            return null;
        }

        if (!canPlayAd(advertisement)) {
            if (advertisement != null && advertisement.getState() == READY) {
                try {
                    _instance.repository.saveAndApplyState(advertisement, placementId, ERROR);
                } catch (DatabaseHelper.DBException e) {
                    if (playAdCallback != null) {
                        playAdCallback.onError(placementId, new VungleException(DB_ERROR));
                    }
                }

                if (placement.isAutoCached()) {
                    _instance.adLoader.loadEndless(placement.getId(), 0, null);
                }
            }
            return null;
        }

        if (TRUE.equals(_instance.playOperations.get(placementId)) || _instance.adLoader.isLoading(placementId)) {
            Log.e(TAG, "Playing or Loading operation ongoing");
            if (playAdCallback != null) {
                playAdCallback.onError(placementId, new VungleException(VungleException.OPERATION_ONGOING));
            }
            return null;
        }

        if (advertisement.getAdType() != Advertisement.TYPE_VUNGLE_MRAID) {
            Log.e(TAG, "Invalid Ad Type for Native Ad.");
            if (playAdCallback != null) {
                playAdCallback.onError(placementId, new VungleException(VungleException.AD_UNABLE_TO_PLAY));
            }
            return null;
        }

        AdConfig.AdSize adSize = (adConfig == null) ? AdConfig.AdSize.VUNGLE_DEFAULT : adConfig.getAdSize();
        if ("mrec".equals(advertisement.getTemplateType()) &&
                adSize != AdConfig.AdSize.VUNGLE_MREC ||
                "flexfeed".equals(advertisement.getTemplateType()) &&
                        adSize != AdConfig.AdSize.VUNGLE_DEFAULT) {
            Log.e(TAG, "Corresponding AdConfig#setAdSize must be passed for the type/size of native ad");
            return null;
        }

        _instance.playOperations.put(placementId, true);
        VungleNativeView nativeView = null;
        try {
            nativeView = new VungleNativeView(_instance.context.getApplicationContext(), placementId, null, new AdContract.AdvertisementPresenter.EventListener() {
                boolean succesfulView = false;
                /**
                 * The percentage of the ad that was viewed. We initialize to -1 in order to differentiate
                 * between the ad never starting.
                 */
                int percentViewed = -1;

                @Override
                public void onNext(String s, String value, String placementId) {
                    try {
                        if (s.equals("start")) {

                            _instance.repository.saveAndApplyState(advertisement, placementId, VIEWING);

                            if (playAdCallback != null) {
                                playAdCallback.onAdStart(placementId);
                            }

                            percentViewed = 0;

                            Placement placement = _instance.repository.load(placementId, Placement.class).get();
                            if (placement != null && placement.isAutoCached()) {
                                /// If the placement is auto-cached, download a new advertisement for it soon.
                                _instance.adLoader.loadEndless(placementId, 0, null);
                            }
                        } else if (s.equals("end")) {
                            Log.d("Vungle", "Cleaning up metadata and assets for placement " + placementId + " and advertisement " + advertisement.getId());
                            /// Cleanup the asset folder for the advertisement
                            _instance.repository.saveAndApplyState(advertisement, placementId, Advertisement.DONE);
                            _instance.repository.updateAndSaveReportState(placementId, advertisement.getAppID(), Report.NEW, Report.READY);

                            /// Schedule a job to send the report soon
                            _instance.jobRunner.execute(SendReportsJob.makeJobInfo(false));

                            _instance.playOperations.put(placementId, false);

                            /// Inform the listener about the advertisement ending
                            if (playAdCallback != null) {
                                playAdCallback.onAdEnd(placementId,
                                        (succesfulView || percentViewed >= 80),
                                        (value != null && value.equals("isCTAClicked")));

                            }
                        } else if (s.equals("successfulView")) {
                            succesfulView = true;
                        } else if (s.startsWith("percentViewed")) {
                            /// Parse out the value passed in as a String.
                            String[] tokens = s.split(":");
                            if (tokens.length == 2) {
                                String percentString = tokens[1];
                                percentViewed = Integer.parseInt(percentString);
                            }
                        }
                    } catch (DatabaseHelper.DBException ignored) {
                        onError(new VungleException(DB_ERROR), placementId);
                    }
                }

                @Override
                public void onError(Throwable throwable, String placementId) {
                    _instance.playOperations.put(placementId, false);
                    if (VungleException.getExceptionCode(throwable) != OPERATION_CANCELED) {
                        try {
                            _instance.repository.saveAndApplyState(advertisement, placementId, Advertisement.ERROR);
                        } catch (DatabaseHelper.DBException e) {
                            throwable = new VungleException(DB_ERROR);
                        }
                    }
                    if (playAdCallback != null) {
                        playAdCallback.onError(placementId, throwable);
                    }
                }
            });
        } catch (Exception ex) {
            _instance.playOperations.put(placementId, false);
            if (playAdCallback != null) {
                playAdCallback.onError(placementId, new InstantiationException(ex.getMessage()));
            }

            if (BuildConfig.DEBUG) {
                throw ex;
            }
        }

        return nativeView;
    }

    public static Collection<String> getValidPlacements() {
        if (!isInitialized()) {
            Log.e(TAG, "Vungle is not initialized return empty placemetns list");
            return Collections.emptyList();
        }

        return _instance.repository.getValidPlacements().get();
    }

    /**
     * GDPR Consent Status
     */
    public enum Consent {
        OPTED_IN,
        OPTED_OUT
    }

    /**
     * If handling GDPR Consent dialog with own implementation, use this dialog to update Vungle on
     * user consent status.
     * <p>
     * Updates the data-gathering consent status of the user. This is gathered by the publisher
     * and provided to Vungle. It is then stored on disk and used whenever we are sending information
     * to the ad server.
     *
     * @param status                If true, the user has consented to us gathering data about their device.
     * @param consentMessageVersion Optional version string that can be passed indicating the version
     *                              of your shown consent message users acted on
     */
    public static void updateConsentStatus(@NonNull final Consent status, @NonNull final String consentMessageVersion) {
        //To ensure update state should be maintained.
        _instance.consent = status;
        _instance.consentVersion = consentMessageVersion;

        if (!isDepInit.get() || _instance.repository == null) {
            return;
        }

        /// If there is already a cookie on disk for consent status, re-use it.
        _instance.repository.load(Cookie.CONSENT_COOKIE, Cookie.class, new Repository.LoadCallback<Cookie>() {
            @Override
            public void onLoaded(Cookie gdprConsent) {
                if (gdprConsent == null) {
                    /// Otherwise, create a new one
                    gdprConsent = new Cookie(Cookie.CONSENT_COOKIE);
                }
                gdprConsent.putValue("consent_status", status == Consent.OPTED_IN ? "opted_in" : "opted_out");
                gdprConsent.putValue("timestamp", System.currentTimeMillis() / 1000); /// Server requires seconds.
                gdprConsent.putValue("consent_source", "publisher");
                gdprConsent.putValue("consent_message_version", consentMessageVersion == null ? "" : consentMessageVersion);
                _instance.repository.save(gdprConsent, null);
            }
        });

    }


    /**
     * @return Whether a user for Vungle has Accepted GDPR Consent. Returns null if user has had the opportunity yet to make opt in or out.
     */
    public static Consent getConsentStatus() {
        if (!isInitialized()) {
            Log.e(TAG, "Vungle is not initialized, consent is null");
            return null;
        }
        return _instance.consent;
    }

    public static String getConsentMessageVersion() {
        if (!isInitialized()) {
            Log.e(TAG, "Vungle is not initialized, please wait initialize or wait until Vungle is intialized to get Consent Message Version");
            return null;
        }
        return _instance.consentVersion;
    }

    private static Consent getConsent(Cookie gdprConsent) {
        if (gdprConsent == null) {
            return null;
        }
        return "opted_in".equals(gdprConsent.getString("consent_status")) ? Consent.OPTED_IN : Consent.OPTED_OUT;

    }

    private static String getConsentMessageVersion(Cookie gdprConsent) {
        if (gdprConsent == null) {
            return null;
        } else {
            return gdprConsent.getString("consent_message_version");
        }
    }


    /**
     * Closes currently playing Ad (only for Native FlexView , not supported for the Interstitial Ad formats)
     * Currently ability to close already Playing ad is supported only for Placements
     * that show FlexView ads.
     *
     * @param placementReferenceId placement id.
     * @return True if Vungle will attempt to close this ad. False if placement is invalid or app has state issue.
     * For other errors you will receive an error message, check debug logs for additional detail.
     */
    public static boolean closeFlexViewAd(@NonNull final String placementReferenceId) {
        if (!isInitialized()) {
            Log.e(TAG, "Vungle is not initialized, can't close flex view ad");
            return false;
        }
        /// Use a broadcast receiver to ferry messages to the Activity without needing to reference
        /// it directly.
        Intent broadcast = new Intent(AdvertisementBus.ACTION);
        broadcast.putExtra(AdvertisementBus.PLACEMENT, placementReferenceId);
        broadcast.putExtra(AdvertisementBus.COMMAND, AdvertisementBus.CLOSE_FLEX);
        LocalBroadcastManager.getInstance(_instance.context).sendBroadcast(broadcast);
        return true;
    }

    /**
     * Set the user's legacy ID to pass through to Third Party Trackers for Attribution
     * This would be effective only if this feature is enabled.
     * We do not store this legacy ID, so this has to be set in every session before or after calling
     * {@link #init(String, Context, InitCallback)}.
     *
     * @param legacyId any legacyID that can use used to uniquely identify user for attribution
     */
    public static void setUserLegacyID(String legacyId) {
        // if Vungle is not initialized or initializing... store it
        // covers cases in which user calls setUserImei immediately before init or after
        if (isInitialized() || isInitializing.get()) {
            _instance.vungleApiClient.updateIMEI(legacyId, _instance.shouldTransmitIMEI);
        } else {
            _instance.userIMEI = legacyId;
        }
    }

    /**
     * Method to register Header bidding callback. It also enable SDK to participate in Header
     * bidding.
     *
     * @param headerBiddingCallback {@link HeaderBiddingCallback} instance to get notified about
     *                              bidding token and ad availability.
     */
    public static void setHeaderBiddingCallback(HeaderBiddingCallback headerBiddingCallback) {
        _instance.runtimeValues.headerBiddingCallback = new HeaderBiddingCallbackWrapper(sdkExecutors.getUIExecutor(), headerBiddingCallback);
    }

    protected static void mockDependencies(Context context, String appId, AssetDownloader downloader, Repository repository, VungleApiClient vungleApiClient, JobRunner jobRunner, SDKExecutors executors, CacheManager cacheManager, AdLoader adLoader) {
        sdkExecutors = executors;

        _instance.context = context;
        _instance.appID = appId;
        _instance.downloader = downloader;
        _instance.repository = repository;
        _instance.vungleApiClient = vungleApiClient;
        _instance.jobRunner = jobRunner;
        _instance.cacheManager = cacheManager;
        _instance.adLoader = adLoader;
    }


    /**
     * Lite version of de-initialization, can be used in Unit Tests and {@link Vungle#init(String, Context, InitCallback)}
     */
    protected static void deInit() {
        if (_instance.cacheManager != null) {
            _instance.cacheManager.removeListener(cacheListener);
        }

        _instance.context = null;
        _instance.appID = null;
        _instance.downloader = null;
        _instance.repository = null;
        _instance.vungleApiClient = null;
        _instance.jobRunner = null;
        _instance.cacheManager = null;
        _instance.adLoader = null;
        _instance.runtimeValues.initCallback = null;

        isInitialized = false;
        isDepInit.set(false);
        isInitializing.set(false);
    }

    private static CacheManager.Listener cacheListener = new CacheManager.Listener() {
        @Override
        public void onCacheChanged() {
            stopPlaying();

            if (_instance != null && _instance.cacheManager.getCache() != null) {
                List<DownloadRequest> requests = _instance.downloader.getAllRequests();
                String newPath = _instance.cacheManager.getCache().getPath();
                for (DownloadRequest request : requests) {

                    if (!request.path.startsWith(newPath)) {
                        _instance.downloader.cancel(request);
                    }

                }
            }
        }
    };

    private static void stopPlaying() {
        if (_instance.context == null)
            return;

        Intent broadcast = new Intent(AdvertisementBus.ACTION);
        broadcast.putExtra(AdvertisementBus.COMMAND, AdvertisementBus.STOP_ALL);
        LocalBroadcastManager.getInstance(_instance.context).sendBroadcast(broadcast);
    }

    private final static VungleStaticApi vungleApi = new VungleStaticApi() {
        @Override
        public boolean isInitialized() {
            return Vungle.isInitialized();
        }

        @Override
        public Collection<String> getValidPlacements() {
            return Vungle.getValidPlacements();
        }
    };
}