package com.vungle.warren;

import android.content.Context;

/**
 * Vungle's setting for
 * {@link Vungle#init(String, Context, InitCallback, VungleSettings)},
 */
public final class VungleSettings {

    private final static long MEGABYTE = 1024 * 1024;
    private volatile long minimumSpaceForInit;
    private volatile long minimumSpaceForAd;
    private volatile boolean androidIdOptedOut;

    /**
     * Builder for {@link VungleSettings}
     */
    public static class Builder {
        private volatile long minimumSpaceForInit = 51 * MEGABYTE;
        private volatile long minimumSpaceForAd = 50 * MEGABYTE;
        private volatile boolean androidIdOptedOut = false;

        public VungleSettings build() {
            return new VungleSettings(this);
        }

        /**
         * Sets minimum amount of free storage required for initialization.
         * Default value is 50 megabytes.
         *
         * @param minimumSpaceForInit minimum storage in bytes.
         * @return {@link VungleSettings.Builder}
         */
        public Builder setMinimumSpaceForInit(long minimumSpaceForInit) {
            this.minimumSpaceForInit = minimumSpaceForInit;
            return this;
        }

        /**
         * Sets minimum amount of free storage required for ad loading.
         * Default value is 51 megabytes.
         *
         * @param minimumSpaceForAd minimum storage in bytes.
         * @return {@link VungleSettings.Builder}
         */
        public Builder setMinimumSpaceForAd(long minimumSpaceForAd) {
            this.minimumSpaceForAd = minimumSpaceForAd;
            return this;
        }

        /**
         * Enables/disables automatic fallback collection of Android ID in case ifa is not available.
         * Default value is false.
         *
         * @param disabled if true disables the collection.
         * @return {@link VungleSettings.Builder}
         */
        public Builder setAndroidIdOptOut(boolean disabled) {
            this.androidIdOptedOut = disabled;
            return this;
        }
    }

    private VungleSettings(Builder builder) {
        this.minimumSpaceForAd = builder.minimumSpaceForAd;
        this.minimumSpaceForInit = builder.minimumSpaceForInit;
        this.androidIdOptedOut = builder.androidIdOptedOut;
    }

    public long getMinimumSpaceForAd() {
        return minimumSpaceForAd;
    }

    public long getMinimumSpaceForInit() {
        return minimumSpaceForInit;
    }

    public boolean getAndroidIdOptOut() {
        return androidIdOptedOut;
    }
}
