package com.vungle.warren.analytics;

import android.support.annotation.NonNull;
import android.text.TextUtils;
import android.util.Log;
import android.util.Pair;
import android.widget.VideoView;

import com.moat.analytics.mobile.vng.MoatAdEvent;
import com.moat.analytics.mobile.vng.MoatAdEventType;
import com.moat.analytics.mobile.vng.MoatFactory;
import com.moat.analytics.mobile.vng.MoatOptions;
import com.moat.analytics.mobile.vng.ReactiveVideoTracker;
import com.moat.analytics.mobile.vng.ReactiveVideoTrackerPlugin;
import com.vungle.warren.BuildConfig;
import com.vungle.warren.model.Advertisement;

import org.json.JSONException;
import org.json.JSONObject;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Queue;

public class MoatTracker implements AnalyticsVideoTracker {

    private static final String TAG = MoatTracker.class.getSimpleName();
    private static final String UNKNOWN = "unknown";
    public static final String VUNGLE_ID = "vunglenativevideo893259554489";
    private VideoView videoView;
    private boolean isMoatEnabled;
    private boolean configured = true;
    private ReactiveVideoTracker tracker;
    private HashMap<String, String> adIds = new HashMap<>();
    /**
     * Queue to store View Percentages MoatTracking
     * Integer represents Percentage 0-100
     */
    private Queue<Pair<Integer, MoatAdEventType>> moatQuartileTrackers;

    private MoatTracker(@NonNull VideoView videoView, boolean isMoatEnabled) {
        this.isMoatEnabled = isMoatEnabled;
        this.videoView = videoView;

    }

    public void configure(@NonNull String placementId, @NonNull Advertisement advertisement, @NonNull String appId) {
        this.tracker = MoatFactory.create().createCustomTracker(new ReactiveVideoTrackerPlugin(VUNGLE_ID));
        this.isMoatEnabled = isMoatEnabled && !TextUtils.isEmpty(placementId) && advertisement != null && advertisement.getMoatEnabled();
        //Initialize moat for this advertisement if config.moat is true and ad.moat is true
        if (isMoatEnabled) {
            MoatOptions options = new MoatOptions();
            options.disableAdIdCollection = true;
            options.disableLocationServices = true;
            options.loggingEnabled = BuildConfig.DEBUG;

            //Create Tuple Queue to store MoatAdEvents <Percent(0-100), MoatAdEventType>
            //Works only once per creation of activity. If we want to play multiple ads on our
            //videoview, this will need to change for quartile tracking

            moatQuartileTrackers = new LinkedList<>();

            moatQuartileTrackers.add(new Pair<>(0, MoatAdEventType.AD_EVT_START));
            moatQuartileTrackers.add(new Pair<>(25, MoatAdEventType.AD_EVT_FIRST_QUARTILE));
            moatQuartileTrackers.add(new Pair<>(50, MoatAdEventType.AD_EVT_MID_POINT));
            moatQuartileTrackers.add(new Pair<>(75, MoatAdEventType.AD_EVT_THIRD_QUARTILE));

            if (!advertisement.getMoatVastExtra().isEmpty()) {
                adIds.put("zMoatVASTIDs", advertisement.getMoatVastExtra());
            }

            String advertiserAppId = advertisement.getAppID();
            final String rawAppIdJson = advertisement.getAppID();
            if (rawAppIdJson != null && rawAppIdJson.length() > 3) {
                try {
                    JSONObject appIdJson = new JSONObject(rawAppIdJson.substring(3));
                    advertiserAppId = (appIdJson.isNull("app_id") ? null : appIdJson.optString("app_id", null));
                } catch (JSONException e) {
                    Log.e(TAG, "JsonException : ", e);
                }
            }
            String campaignId = null;
            String creativeId = null;
            final String campaign = advertisement.getCampaign();
            if (!TextUtils.isEmpty(campaign)) {
                final int firstBar = campaign.indexOf('|');
                if (firstBar != -1) {
                    campaignId = campaign.substring(0, firstBar);

                    final int secondBar = campaign.indexOf('|', firstBar + 1);
                    if (secondBar != -1) {
                        creativeId = campaign.substring(firstBar + 1, secondBar);
                    } else {
                        creativeId = campaign.substring(firstBar + 1);
                    }
                } else {
                    campaignId = campaign;
                }
            }

            adIds.put("level1", TextUtils.isEmpty(advertiserAppId) ? UNKNOWN : advertiserAppId);
            adIds.put("level2", TextUtils.isEmpty(campaignId) ? UNKNOWN : campaignId);
            adIds.put("level3", TextUtils.isEmpty(creativeId) ? UNKNOWN : creativeId);
            adIds.put("level4", TextUtils.isEmpty(placementId) ? UNKNOWN : placementId);

            if (!TextUtils.isEmpty(appId)) {
                adIds.put("slicer1", appId);
            }

            configured = true;
        }

        configured = configured && isMoatEnabled;
    }

    public static MoatTracker connect(@NonNull VideoView videoView, boolean isEnabled) {
        return new MoatTracker(videoView, isEnabled);
    }

    @Override
    public void onProgress(int progress) {
        if (configured) {

            if (progress >= 100) {
                //have to send current video position in millis to moat or else view is not counted
                tracker.dispatchEvent(new MoatAdEvent(MoatAdEventType.AD_EVT_COMPLETE, videoView.getCurrentPosition()));
                tracker.stopTracking();
            } else {
                //Moat Quartile Tracking. Queue Polls for every completed quartile
                if (!moatQuartileTrackers.isEmpty() && progress >= moatQuartileTrackers.peek().first) {
                    MoatAdEvent event = new MoatAdEvent(moatQuartileTrackers.poll().second, progress);
                    tracker.dispatchEvent(event);
                }

            }
        }

    }

    @Override
    public void start(int videoDuration) {
        if (configured) {
            Log.d(TAG, "start");
            tracker.trackVideoAd(adIds, videoDuration, videoView);
        }
    }

    @Override
    public void stop() {
        if (configured) {
            /// If user did not finish video, send moat analytics.
            /// This must be done on the same thread as the WebView was created in.
            int currentVideoPosition = (videoView != null) ? videoView.getCurrentPosition() : 0;
            Log.d(TAG, "stopViewabilityTracker: " + currentVideoPosition);
            tracker.dispatchEvent(new MoatAdEvent(MoatAdEventType.AD_EVT_STOPPED, currentVideoPosition));
            tracker.stopTracking();
            Log.d(TAG, "stopViewabilityTracker: Success !!");
        }
    }

}
