package com.vungle.warren.download;

import android.Manifest;
import android.app.NotificationChannel;
import android.app.NotificationManager;
import android.content.Context;
import android.content.Intent;
import android.net.Uri;
import android.os.Build;
import android.os.Environment;
import android.support.annotation.Nullable;
import android.support.v4.app.NotificationCompat;
import android.support.v4.content.FileProvider;
import android.support.v4.content.PermissionChecker;
import android.text.TextUtils;
import android.util.Log;
import android.webkit.MimeTypeMap;
import android.widget.Toast;

import com.vungle.warren.downloader.AssetDownloadListener;
import com.vungle.warren.downloader.AssetDownloadListener.Progress;
import com.vungle.warren.downloader.AssetDownloadListener.Progress.ProgressStatus;
import com.vungle.warren.downloader.AssetDownloader;
import com.vungle.warren.downloader.DownloadRequest;
import com.vungle.warren.locale.LocaleString;
import com.vungle.warren.ui.VungleWebViewActivity;

import java.io.File;
import java.io.IOException;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.List;

import static com.vungle.warren.downloader.AssetDownloader.NetworkType.WIFI;

/**
 * Utility class for apk direct download feature.
 */
public class APKDirectDownloadManager {

    public static final int DIRECT_DOWNLOAD_FLAG_NOT_SET = -1;
    public static final int DIRECT_DOWNLOAD_FLAG_DISABLED = 0;
    public static final int DIRECT_DOWNLOAD_FLAG_ENABLED = 1;
    private final static String TAG = "DirectDownloadManager";

    private static final String APK_POSTFIX = "apk";
    private static final String FOLDER_NAME = "apks_vungle";
    private static final String FOLDER_APK = "apk";

    private static final String NOTIFICATION_CHANNEL_ID = "1";
    private static final String NOTIFICATION_CHANNEL_NAME = "ApkDirectDownload";

    private static APKDirectDownloadManager _instance = new APKDirectDownloadManager();

    private List<Integer> notificationList = new ArrayList<>();
    private int apkDirectDownloadStatus = DIRECT_DOWNLOAD_FLAG_NOT_SET;
    /**
     * Downloader instance that we will use to download apk.
     */

    private AssetDownloader downloader;

    private WeakReference<Context> context;
    private NotificationManager notifyManager;
    private NotificationCompat.Builder mBuilder;


    public static void init(Context context, AssetDownloader assetDownloader) {
        _instance.context = new WeakReference<>(context);
        _instance.clearDownloadApkCache();
        if (_instance.downloader == null) {
            _instance.downloader = assetDownloader;
        }
    }

    /**
     * Check apk direct download feature is enabled or not (V2.0)
     * Follow server side config. If apk_direct_download config and requires_sideloading
     * config are both true, the feature is enabled.
     * if apk_direct_download config is not set, use client config as a fallback check.
     */
    public static boolean isDirectDownloadEnabled(boolean clientFlag, boolean requiresSideLoading) {

        /**
         * First check external storage status.
         */
        if (!_instance.checkExternalStorageAvailable()) {
            return false;
        }

        /** Check "apk_direct_download" status.
         * 1. NOT_SET:
         * --- Check client settings.
         *     if set disabled:
         *        return false;
         *     else return the "requires_sideloading"
         *
         * 2. ENABLED;
         *  ---Check the "requires_sideloading"
         *
         * 3. DISABLED:
         * return false;
         *
         */

        switch (_instance.apkDirectDownloadStatus) {
            case DIRECT_DOWNLOAD_FLAG_NOT_SET:
                if (clientFlag) {
                    return requiresSideLoading;
                } else {
                    return false;
                }

            case DIRECT_DOWNLOAD_FLAG_ENABLED:
                return requiresSideLoading;

            case DIRECT_DOWNLOAD_FLAG_DISABLED:
            default:
                return false;

        }
    }

    /**
     * Set apk direct download status from server side.
     */
    public static void setDirectDownloadStatus(int status) {
        _instance.apkDirectDownloadStatus = status;
    }

    public static void download(String url) {

        if (_instance == null)
            return;

        // 1. check the download link is apk url or landing page url;
        if (TextUtils.isEmpty(url) || getContext() == null) {
            return;
        }

        if (_instance.isApkUrl(url)) {
            String identifier = String.valueOf(System.currentTimeMillis());

            final int notificationId = identifier.hashCode();
            _instance.notificationList.add(notificationId);

            final File destFile = _instance.getApkDirectory(identifier);
            if (destFile == null) {
                Log.e(TAG, "apk file is missing!");
                _instance.openUrl(url);
                return;
            }

            Toast.makeText(getContext(), LocaleString.getLocaleText(LocaleString.ID_START_DOWNLOAD), Toast.LENGTH_LONG).show();

            DownloadRequest downloadRequest = new DownloadRequest(WIFI, DownloadRequest.Priority.DEFAULT, url, destFile.getAbsolutePath(), true, null);

            _instance.downloader.download(downloadRequest, new AssetDownloadListener() {

                @Override
                public void onError(DownloadError error, DownloadRequest downloadRequest) {
                    Log.d(TAG, "download onError :"
                            + (error.cause == null ? "null" : error.cause.getMessage())
                            + " id:" + downloadRequest.url);
                    _instance.dismissNotification(notificationId);
                    _instance.notificationList.remove((Integer) notificationId);
                }

                @Override
                public void onProgress(Progress progress, DownloadRequest downloadRequest) {
                    Log.d(TAG, "download progress :"
                            + progress.progressPercent + "status:" + downloadRequest.getStatus());

                        if (progress.status == ProgressStatus.CANCELLED) {
                            _instance.dismissNotification(notificationId);
                        } else {
                            _instance.notifyProgress(notificationId, progress);
                        }
                }

                @Override
                public void onSuccess(File file, DownloadRequest downloadRequest) {
                    if (file == null) {
                        onError(new DownloadError(-1, new IOException("Downloaded file not found!"), -1), downloadRequest);
                        return;
                    }

                    Log.d(TAG, "download complete :" + file.getAbsolutePath());
                    _instance.installApk(destFile);

                    _instance.notificationList.remove((Integer) notificationId);
                }

            });

        } else if (getContext() != null) {
            Intent intent = new Intent();
            intent.setClass(getContext(), VungleWebViewActivity.class);
            intent.putExtra(VungleWebViewActivity.INTENT_URL, url);
            intent.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
            getContext().startActivity(intent);
        }

    }

    public static boolean isApkUrl(String url) {

        String ext = MimeTypeMap.getFileExtensionFromUrl(url);
        return ext.toLowerCase().endsWith(APK_POSTFIX);
    }

    private void clearDownloadApkCache() {
        new Thread(new Runnable() {
            @Override
            public void run() {

                File apkCacheFile = getCacheDirectory();
                if (apkCacheFile == null) {
                    return;
                }
                File[] files = apkCacheFile.listFiles();
                if (files == null || files.length == 0) {
                    return;
                }
                for (int i = 0; i < files.length; i++) {
                    if (files[i].isFile() && files[i].exists() && files[i].isFile()) {
                        Log.d(TAG, "Clear cache:" + files[i].getName());
                        files[i].delete();
                    }

                }
            }
        }).start();
    }

    private void openUrl(String url) {
        // Open the Destination URL
        try {
            Intent i = Intent.parseUri(url, 0);
            i.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
            Context ctx = getContext();
            if (ctx != null) {
                ctx.startActivity(i);
            }
        } catch (Exception e) {
            Log.e(TAG, "Unable to start activity " + e.getLocalizedMessage());
        }
    }

    private File getApkDirectory(String identifier) {
        File cacheDir = getCacheDirectory();
        if (cacheDir == null) {
            return null;
        }
        File folderDir = new File(cacheDir.getPath() + File.separator);
        if (!folderDir.exists()) {
            folderDir.mkdir();
        }

        File apkFile = new File(cacheDir + File.separator + identifier + "." + APK_POSTFIX);
        return apkFile;
    }

    private boolean checkExternalStorageAvailable() {
        try {
            String state = Environment.getExternalStorageState();
            if (Environment.MEDIA_MOUNTED.equals(state)) {
                File cacheDir = getCacheDirectory();
                if (cacheDir != null) {
                    return true;
                }
            }

        } catch (Exception ex) {
            Log.e(TAG, "Storage state error:" + ex.getMessage());
        }
        return false;
    }

    private File getCacheDirectory() {
        // Use external cache path to save apk file to avoid access issue during install process.
        if (getContext() == null || getContext().getExternalCacheDir() == null) {
            return null;
        }

        File dir = new File(getContext().getExternalCacheDir().getPath() + File.separator + FOLDER_NAME);

        if (!dir.exists()) {
            dir.mkdir();
        }
        return dir;
    }

    private void installApk(File apkFile) {
        if (getContext() == null) {
            Log.e(TAG, "context is null");
            return;
        }
        Intent installApkIntent = new Intent();
        installApkIntent.setAction(Intent.ACTION_VIEW);
        installApkIntent.addCategory(Intent.CATEGORY_DEFAULT);
        installApkIntent.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK);

        if (Build.VERSION.SDK_INT > Build.VERSION_CODES.M) {
            installApkIntent.setDataAndType(FileProvider.getUriForFile(getContext(), "com.vungle.download.provider", apkFile), "application/vnd.android.package-archive");
            installApkIntent.addFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION | Intent.FLAG_GRANT_WRITE_URI_PERMISSION);
        } else {
            installApkIntent.setDataAndType(Uri.fromFile(apkFile), "application/vnd.android.package-archive");
        }

        if (getContext().getPackageManager().queryIntentActivities(installApkIntent, 0).size() > 0) {
            getContext().startActivity(installApkIntent);
        }
        String name = apkFile.getName();
        String identifier = name.substring(0, name.length() - 4);
        Log.d(TAG, "identifier is" + identifier);
        dismissNotification(identifier.hashCode());
    }

    private void notifyProgress(int notifyId, Progress progress) {
        Log.d(TAG, "notify id is :" + notifyId + " progress:" + progress);
        if (getContext() == null) {
            Log.e(TAG, "context is null.");
            return;
        }

        if (notifyManager == null) {
            notifyManager = (NotificationManager)
                    getContext().getSystemService(getContext().NOTIFICATION_SERVICE);

            mBuilder = new NotificationCompat.Builder(getContext());
            mBuilder.setSmallIcon(android.R.drawable.stat_sys_download_done);

            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                NotificationChannel channel = new NotificationChannel(NOTIFICATION_CHANNEL_ID,
                        NOTIFICATION_CHANNEL_NAME, NotificationManager.IMPORTANCE_LOW);
                notifyManager.createNotificationChannel(channel);
                mBuilder.setChannelId(NOTIFICATION_CHANNEL_ID);
            }

        }

        if (progress.status == ProgressStatus.DONE) {
            mBuilder.setContentText(LocaleString.getLocaleText(LocaleString.ID_DOWNLOAD_COMPLETE))
                    .setProgress(0, 0, false);
        } else if (progress.status == ProgressStatus.LOST_CONNECTION || progress.status == ProgressStatus.PAUSED) {
            mBuilder.setContentTitle(LocaleString.getLocaleText(LocaleString.ID_DOWNLOADING))
                    .setContentText(LocaleString.getLocaleText(LocaleString.ID_WAITING_FOR_WIFI))
                    .setProgress(0, 0, false);
        } else {
            mBuilder.setContentTitle(LocaleString.getLocaleText(LocaleString.ID_DOWNLOADING))
                    .setContentText(LocaleString.getLocaleText(LocaleString.ID_PROGRESS));
            int progressInt = Math.min(100, Math.max(progress.progressPercent, 0));
            mBuilder.setProgress(100, progressInt, false);
        }

        notifyManager.notify(notifyId, mBuilder.build());

    }

    private void dismissNotification(int identifier) {
        if (notifyManager != null) {
            notifyManager.cancel(identifier);
        }
    }

    @Nullable
    private static Context getContext() {
        return _instance == null || _instance.context == null ? null : _instance.context.get();
    }

}