package com.vungle.warren.downloader;

import android.annotation.SuppressLint;
import android.annotation.TargetApi;
import android.os.Build;
import android.support.annotation.IntDef;
import android.support.annotation.NonNull;
import android.support.annotation.Nullable;
import android.text.TextUtils;
import android.util.Log;

import com.vungle.warren.BuildConfig;
import com.vungle.warren.downloader.AssetDownloadListener.DownloadError.ErrorReason;
import com.vungle.warren.downloader.DownloadRequest.Status;
import com.vungle.warren.utility.FileUtility;
import com.vungle.warren.utility.NetworkProvider;
import com.vungle.warren.utility.PriorityRunnable;

import java.io.File;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import javax.net.ssl.SSLException;

import okhttp3.Call;
import okhttp3.Headers;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.Response;
import okhttp3.ResponseBody;
import okhttp3.internal.http.HttpHeaders;
import okhttp3.internal.http.RealResponseBody;
import okhttp3.logging.HttpLoggingInterceptor;
import okio.BufferedSink;
import okio.BufferedSource;
import okio.GzipSource;
import okio.Okio;

import static android.net.ConnectivityManager.TYPE_BLUETOOTH;
import static android.net.ConnectivityManager.TYPE_ETHERNET;
import static android.net.ConnectivityManager.TYPE_MOBILE;
import static android.net.ConnectivityManager.TYPE_MOBILE_DUN;
import static android.net.ConnectivityManager.TYPE_VPN;
import static android.net.ConnectivityManager.TYPE_WIFI;
import static android.net.ConnectivityManager.TYPE_WIMAX;
import static com.vungle.warren.downloader.AssetDownloadListener.DownloadError;
import static com.vungle.warren.downloader.AssetDownloadListener.Progress;
import static com.vungle.warren.downloader.AssetDownloadListener.Progress.ProgressStatus;
import static java.net.HttpURLConnection.HTTP_OK;
import static java.net.HttpURLConnection.HTTP_PARTIAL;

@SuppressLint("LogNotTimber")
public class AssetDownloader {

    private static final String BYTES = "bytes";
    private static final String RANGE = "Range";
    private static final String LAST_MODIFIED = "Last-Modified";
    private static final String ETAG = "ETag";
    private static final String ACCEPT_RANGES = "Accept-Ranges";
    private static final String CONTENT_ENCODING = "Content-Encoding";
    private static final String CONTENT_RANGE = "Content-Range";
    private static final String CONTENT_TYPE = "Content-Type";
    private static final String ACCEPT_ENCODING = "Accept-Encoding";
    private static final String IF_RANGE = "If-Range";
    private static final String IDENTITY = "identity";
    private static final String GZIP = "gzip";

    private static final String META_POSTFIX_EXT = ".vng_meta";

    private static final int TIMEOUT = 30;
    private static final int PROGRESS_STEP = 5;
    private static final String TAG = AssetDownloader.class.getSimpleName();
    private static final int RETRY_COUNT_ON_CONNECTION_LOST = 5;
    private static final int CONNECTION_RETRY_TIMEOUT = 300;
    private static final int MAX_RECONNECT_ATTEMPTS = 10;
    private static final int RANGE_NOT_SATISFIABLE = 416;
    private static final long MAX_PERCENT = 100;

    //For testing
    int retryCountOnConnectionLost = RETRY_COUNT_ON_CONNECTION_LOST;
    int maxReconnectAttempts = MAX_RECONNECT_ATTEMPTS;
    int reconnectTimeout = CONNECTION_RETRY_TIMEOUT;

    private final NetworkProvider networkProvider;
    private final ExecutorService downloadExecutor;
    private final OkHttpClient okHttpClient;
    private final ExecutorService uiExecutor;

    private static final int DOWNLOAD_CHUNK_SIZE = 2048; //Same as Okio Segment.SIZE

    private final ConcurrentHashMap<String, AssetDownloadListener> listeners = new ConcurrentHashMap<>();
    private final ConcurrentHashMap<String, DownloadRequest> connections = new ConcurrentHashMap<>();

    private volatile int progressStep = PROGRESS_STEP;

    public AssetDownloader(int concurrentDownloads, NetworkProvider networkProvider, ExecutorService uiExecutor) {
        int concurrentNum = Math.max(concurrentDownloads, 1);
        ThreadPoolExecutor executor = new ThreadPoolExecutor(concurrentNum, concurrentNum, 1, TimeUnit.SECONDS, new PriorityBlockingQueue<Runnable>());
        executor.allowCoreThreadTimeOut(true);
        this.downloadExecutor = executor;
        this.networkProvider = networkProvider;
        this.uiExecutor = uiExecutor;

        OkHttpClient.Builder builder = new OkHttpClient.Builder()
                .readTimeout(TIMEOUT, TimeUnit.SECONDS)
                .connectTimeout(TIMEOUT, TimeUnit.SECONDS)
                .cache(null)
                .followRedirects(true)
                .followSslRedirects(true);

        if (BuildConfig.DEBUG) {
            HttpLoggingInterceptor loggingInterceptor = new HttpLoggingInterceptor();
            loggingInterceptor.setLevel(HttpLoggingInterceptor.Level.HEADERS);
            builder.addNetworkInterceptor(loggingInterceptor);
        }

        okHttpClient = builder.build();
    }

    public synchronized void download(final DownloadRequest downloadRequest, final AssetDownloadListener downloadListener) {

        if (downloadRequest == null) {
            if (downloadListener != null) {
                uiExecutor.execute(new Runnable() {
                    @Override
                    public void run() {
                        downloadListener.onError(
                                new DownloadError(
                                        -1,
                                        new IllegalArgumentException("DownloadRequest is null"),
                                        ErrorReason.REQUEST_ERROR),
                                null
                        );
                    }
                });
            }
            return;
        }

        //Prevent duplicate
        final DownloadRequest existing = connections.get(downloadRequest.id);

        if (existing != null) {
            Log.d(TAG, "Request with same url and path already present " + existing.id);
            // Should be discussed, re-connect/return listener to existing download?
            if (downloadListener != null) {
                uiExecutor.execute(new Runnable() {
                    @Override
                    public void run() {
                        downloadListener.onError(
                                new DownloadError(-1,
                                        new IllegalArgumentException("Request with same url and path already present"),
                                        ErrorReason.REQUEST_ERROR),
                                downloadRequest
                        );
                    }
                });
            }

            return;
        }

        for (DownloadRequest connection : connections.values()) {
            if (connection.path.equals(downloadRequest.path)) {
                Log.d(TAG, "Already present request for same path");
                if (downloadListener != null) {
                    uiExecutor.execute(new Runnable() {
                        @Override
                        public void run() {
                            downloadListener.onError(
                                    new DownloadError(-1,
                                            new IllegalArgumentException("Already present request for same path"),
                                            ErrorReason.REQUEST_ERROR),
                                    downloadRequest
                            );
                        }
                    });
                }

                return;
            }
        }

        listeners.put(downloadRequest.id, downloadListener);
        connections.put(downloadRequest.id, downloadRequest);
        networkProvider.addListener(networkListener);

        // add check current connection
        load(downloadRequest);
    }

    public List<DownloadRequest> getAllRequests() {
        return new ArrayList<>(connections.values());
    }

    private synchronized void load(final DownloadRequest downloadRequest) {
        downloadRequest.set(Status.IN_PROGRESS);

        downloadExecutor.execute(new DownloadPriorityRunnable(downloadRequest.priority) {
            @Override
            public void run() {
                int reconnectRetries = 0;
                boolean done = false;
                Progress progress = new Progress();
                progress.timestampDownloadStart = System.currentTimeMillis();
                DownloadError downloadError = null;
                boolean paused = false;
                while (!done) {
                    done = true;
                    String url = downloadRequest.url;
                    String path = downloadRequest.path;

                    Log.d(TAG, "Start load: " + downloadRequest.id + " url: " + url);

                    BufferedSink sink = null;
                    BufferedSource source = null;
                    long totalRead = 0;
                    Call call = null;
                    long downloaded;
                    int code = -1;
                    Response response = null;
                    File file = null;
                    try {

                        if (!downloadRequest.is(Status.IN_PROGRESS)) {
                            Log.w(TAG, "Abort download, wrong state " + debugString(downloadRequest));
                            return;
                        }

                        if (!isConnected(downloadRequest)) {
                            Log.d(TAG, "Request is not connected to reuired network");
                            throw new IOException("Not connected to correct network");
                        } else {
                            downloadRequest.setConnected(true);
                        }

                        file = new File(path);
                        File metaFile = new File(path + META_POSTFIX_EXT);
                        if (file.getParentFile() != null && !file.getParentFile().exists()) {
                            file.getParentFile().mkdirs();
                        }

                        downloaded = file.exists() ? file.length() : 0;

                        Log.d(TAG, "already downloaded : "
                                + downloaded
                                + ", file exists = "
                                + file.exists()
                                + debugString(downloadRequest));

                        Request.Builder requestBuilder = new Request.Builder()
                                .url(url)
                                .addHeader(ACCEPT_ENCODING, IDENTITY);

                        if (file.exists() && metaFile.exists()) {
                            Map<String, String> meta = FileUtility.readMap(metaFile.getPath());
                            // We want to make sure that server accepts bytes and
                            // also we don't yet deal with gzip
                            if (BYTES.equalsIgnoreCase(meta.get(ACCEPT_RANGES)) &&
                                    (IDENTITY.equalsIgnoreCase(meta.get(CONTENT_ENCODING)) || meta.get(CONTENT_ENCODING) == null)) {

                                requestBuilder.addHeader(RANGE, BYTES + "=" + downloaded + "-");

                                String ifRangeValue = meta.get(ETAG);

                                if (ifRangeValue == null)
                                    ifRangeValue = meta.get(LAST_MODIFIED);

                                if (ifRangeValue != null)
                                    requestBuilder.addHeader(IF_RANGE, ifRangeValue);
                            }
                        }

                        call = okHttpClient.newCall(requestBuilder.build());

                        response = call.execute();
                        long contentLength = getContentLength(response);

                        Log.d(TAG, "Response code: " + response.code() + " " + downloadRequest.id);

                        code = response.code();

                        if (fullyDownloadedContent(file, response, downloadRequest)) {
                            downloadRequest.set(Status.DONE);
                            return;
                        }

                        if ((code == HttpURLConnection.HTTP_PARTIAL && !satisfiesPartialDownload(response, downloaded, downloadRequest))
                                || code == RANGE_NOT_SATISFIABLE) {

                            if (file.exists())
                                file.delete();

                            if (metaFile.exists())
                                metaFile.delete();

                            //retry to get ad from the scratch
                            if (reconnectRetries++ < maxReconnectAttempts) {
                                done = false;
                                continue;
                            } else {
                                throw new RequestException("Code: " + code);
                            }
                        }

                        if (!response.isSuccessful()) {
                            throw new RequestException("Code: " + code);
                        }

                        if (code != HttpURLConnection.HTTP_PARTIAL) {
                            downloaded = 0;
                            if (file.exists())
                                file.delete();
                        }

                        metaFile.delete();

                        Headers headers = response.headers();

                        String contentEncoding = headers.get(CONTENT_ENCODING);
                        if (contentEncoding != null &&
                                !GZIP.equalsIgnoreCase(contentEncoding) &&
                                !IDENTITY.equalsIgnoreCase(contentEncoding)) {
                            throw new IOException("Unknown " + CONTENT_ENCODING);
                        }

                        HashMap<String, String> map = new HashMap<>();
                        map.put(ETAG, headers.get(ETAG));
                        map.put(LAST_MODIFIED, headers.get(LAST_MODIFIED));
                        map.put(ACCEPT_RANGES, headers.get(ACCEPT_RANGES));
                        map.put(CONTENT_ENCODING, headers.get(CONTENT_ENCODING));
                        FileUtility.writeMap(metaFile.getPath(), map);

                        ResponseBody body = decodeGzipIfNeeded(response);

                        if (body == null)
                            throw new IOException("Response body is null");

                        source = body.source();

                        Log.d(TAG, "Start download from bytes: " + downloaded
                                + debugString(downloadRequest));

                        long offset = downloaded;
                        contentLength += offset;

                        Log.d(TAG, "final offset = " + offset);

                        sink = Okio.buffer(offset == 0 ? Okio.sink(file) : Okio.appendingSink(file));

                        long read;
                        progress.status = ProgressStatus.STARTED;
                        progress.sizeBytes = body.contentLength();
                        progress.startBytes = offset;

                        deliverProgress(downloadRequest.id, progress);
                        int current = 0;
                        while (downloadRequest.is(Status.IN_PROGRESS)
                                && (read = (source.read(sink.buffer(), DOWNLOAD_CHUNK_SIZE))) != -1) {

                            sink.emit();
                            totalRead += read;

                            downloaded = offset + totalRead;

                            if (contentLength > 0) {
                                current = (int) ((downloaded * MAX_PERCENT) / contentLength);
                            }

                            if (!downloadRequest.isConnected()) {
                                throw new IOException("Request is not connected");
                            }

                            if (current <= progress.progressPercent)
                                continue;

                            progress.progressPercent = current;

                            if (progress.progressPercent % progressStep == 0) {
                                progress.status = ProgressStatus.IN_PROGRESS;
                                deliverProgress(downloadRequest.id, progress);
                            }
                        }

                        sink.flush();

                        if (downloadRequest.is(Status.IN_PROGRESS)) {
                            downloadRequest.set(Status.DONE);
                        } else {
                            progress.status = ProgressStatus.STATE_CHANGED;
                            deliverProgress(downloadRequest.id, progress);
                            Log.d(TAG, "State has changed, cancelling download " + debugString(downloadRequest));
                        }

                    } catch (Throwable throwable) {
                        Log.e(TAG, "Exception on download", throwable);
                        downloadRequest.set(Status.ERROR);

                        if (throwable instanceof IOException) {
                            boolean connected = isConnected(downloadRequest);
                            downloadRequest.setConnected(connected);

                            if (!connected) {
                                progress.status = ProgressStatus.LOST_CONNECTION;
                                deliverProgress(downloadRequest.id, progress);

                                if (!downloadRequest.is(Status.CANCELLED)
                                        && reconnectRetries++ < maxReconnectAttempts) {

                                    for (int i = 0; i < retryCountOnConnectionLost; i++) {
                                        try {
                                            Thread.sleep(Math.max(0, reconnectTimeout));
                                        } catch (InterruptedException ie) {
                                            ie.printStackTrace();
                                        }

                                        if (downloadRequest.is(Status.CANCELLED)) {
                                            break;
                                        }

                                        Log.d(TAG, "Trying to reconnect");

                                        if (isConnected(downloadRequest)) {
                                            Log.d(TAG, "Reconnected, starting download again");
                                            done = false;
                                            downloadRequest.setConnected(true);
                                            downloadRequest.set(Status.IN_PROGRESS);
                                            break;
                                        }

                                        downloadRequest.setConnected(false);
                                    }
                                }
                            }

                            if (done && !downloadRequest.is(Status.CANCELLED)) {
                                if (shouldPause(downloadRequest)) {
                                    paused = true;
                                    downloadRequest.set(Status.PAUSED);
                                    onPaused(progress, downloadRequest);
                                } else {
                                    downloadRequest.set(Status.ERROR);
                                    downloadError = new DownloadError(code, throwable,
                                            mapExceptionToReason(throwable, connected));
                                }
                            }


                        } else if (throwable instanceof RequestException) {
                            downloadError = new DownloadError(code, throwable,
                                    ErrorReason.REQUEST_ERROR);
                        } else {
                            downloadError = new DownloadError(code, throwable,
                                    ErrorReason.INTERNAL_ERROR);
                        }

                    } finally {

                        if (response != null && response.body() != null) {
                            response.body().close();
                        }

                        if (call != null) {
                            call.cancel();
                        }

                        Log.d(TAG, "request is done " + debugString(downloadRequest));


                        synchronized (AssetDownloader.this) {

                            if (done) {
                                switch (downloadRequest.getStatus()) {
                                    case Status.PAUSED:
                                        //no-op
                                        break;
                                    case Status.DONE:
                                        onSuccess(file, downloadRequest);
                                        break;
                                    case Status.ERROR:
                                        onError(downloadError, downloadRequest);
                                        break;
                                    case Status.CANCELLED:
                                        onCancelled(downloadRequest, progress);
                                    default:
                                        if (!paused) {
                                            connections.remove(downloadRequest.id);
                                            listeners.remove(downloadRequest.id);
                                        }
                                        break;
                                }

                                Log.d(TAG, "Removing connections and listener " + debugString(downloadRequest));
                            } else {
                                Log.d(TAG, "Not removing connections and listener " + debugString(downloadRequest));
                            }

                            if (connections.isEmpty()) {
                                networkProvider.removeListener(networkListener);
                            }

                        }

                        FileUtility.closeQuietly(sink);
                        FileUtility.closeQuietly(source);
                    }
                }
            }
        });
    }

    private ResponseBody decodeGzipIfNeeded(Response networkResponse) {
        if ("gzip".equalsIgnoreCase(networkResponse.header(CONTENT_ENCODING))
                && HttpHeaders.hasBody(networkResponse)) {
            GzipSource responseBody = new GzipSource(networkResponse.body().source());
            String contentType = networkResponse.header(CONTENT_TYPE);
            return new RealResponseBody(contentType, -1L, Okio.buffer(responseBody));
        }

        return networkResponse.body();
    }

    private void onCancelled(@NonNull final DownloadRequest downloadRequest, @Nullable final Progress progress) {
        connections.remove(downloadRequest.id);
        final AssetDownloadListener downloadListener = listeners.remove(downloadRequest.id);
        final Progress progressEnd;
        if (progress == null) {
            progressEnd = new Progress();
        } else {
            progressEnd = progress;
        }

        progressEnd.status = ProgressStatus.CANCELLED;

        if (downloadListener != null) {
            uiExecutor.execute(new Runnable() {
                @Override
                public void run() {

                    downloadListener.onProgress(progressEnd, downloadRequest);
                }
            });
        }
        Log.d(TAG, "Cancelled " + debugString(downloadRequest));
    }

    private int mapExceptionToReason(Throwable e, boolean connected) {
        @ErrorReason int reason;
        if (e instanceof RuntimeException) {
            reason = ErrorReason.INTERNAL_ERROR;
        } else if (!connected
                || e instanceof SocketException
                || e instanceof SocketTimeoutException) {
            reason = ErrorReason.CONNECTION_ERROR;
        } else if (e instanceof UnknownHostException
                || e instanceof SSLException) {
            reason = ErrorReason.REQUEST_ERROR;
        } else {
            reason = ErrorReason.DISK_ERROR;
        }
        return reason;
    }

    private long getContentLength(Response response) {
        if (response == null)
            return -1;

        String header = response.headers().get("Content-Length");
        if (TextUtils.isEmpty(header))
            return -1;

        try {
            return Long.parseLong(header);
        } catch (Throwable t) {
            return -1;
        }
    }

    private boolean fullyDownloadedContent(File file, Response response, DownloadRequest downloadRequest) {
        if (!file.exists() || file.length() <= 0
                || GZIP.equalsIgnoreCase(response.headers().get(CONTENT_ENCODING)))
            return false;

        int code = response.code();
        long contentLength = getContentLength(response);

        if (code == HTTP_OK && contentLength == file.length()) {
            Log.d(TAG, "200 code, data size matches file size " + debugString(downloadRequest));
            return responseVersionMatches(file, response);
        }

        if (code == RANGE_NOT_SATISFIABLE) {
            String rangeHeader = response.headers().get(CONTENT_RANGE);
            if (TextUtils.isEmpty(rangeHeader))
                return false;

            RangeResponse rangeResponse = new RangeResponse(rangeHeader);

            boolean result = BYTES.equalsIgnoreCase(rangeResponse.dimension)
                    && rangeResponse.total > 0
                    && rangeResponse.total == file.length()
                    && responseVersionMatches(file, response);

            Log.d(TAG, "416 code, data size matches file size " + debugString(downloadRequest));

            return result;
        }

        return false;
    }

    private synchronized void onSuccess(final File file, @NonNull final DownloadRequest downloadRequest) {
        connections.remove(downloadRequest.id);
        final AssetDownloadListener downloadListener = listeners.remove(downloadRequest.id);

        Log.d(TAG, "OnComplete - Removing connections and listener " + downloadRequest.id);

        if (downloadListener != null) {
            uiExecutor.execute(new Runnable() {
                @Override
                public void run() {
                    downloadListener.onSuccess(file, downloadRequest);
                }
            });
        }
        Log.d(TAG, "Finished " + debugString(downloadRequest));
    }

    private synchronized void onPaused(Progress progress, DownloadRequest downloadRequest) {
        Log.d(TAG, "Pausing download " + debugString(downloadRequest));
        progress.status = ProgressStatus.PAUSED;
        deliverProgress(downloadRequest.id, progress);
    }

    private synchronized void onError(DownloadError downloadError, @Nullable final DownloadRequest downloadRequest) {
        if (downloadRequest == null)
            return;

        final AssetDownloadListener downloadListener = listeners.remove(downloadRequest.id);
        connections.remove(downloadRequest.id);
        downloadRequest.set(Status.ERROR);

        final DownloadError endError;
        if (downloadError == null) {
            endError = new DownloadError(-1, new RuntimeException(), ErrorReason.INTERNAL_ERROR);
        } else {
            endError = downloadError;
        }

        Log.d(TAG, "OnError - Removing connections and listener " + downloadRequest.id);
        if (downloadListener != null) {
            Log.e(TAG, "On download error " + downloadError);
            uiExecutor.execute(new Runnable() {
                @Override
                public void run() {
                    downloadListener.onError(endError, downloadRequest);
                }
            });
        }

    }

    private boolean satisfiesPartialDownload(Response response, long bytesLoaded, DownloadRequest downloadRequest) {
        RangeResponse rangeResponse = new RangeResponse(response.headers().get(CONTENT_RANGE));
        boolean result = response.code() == HTTP_PARTIAL
                && BYTES.equalsIgnoreCase(rangeResponse.dimension)
                && rangeResponse.rangeStart >= 0
                && bytesLoaded == rangeResponse.rangeStart;
        Log.d(TAG, "satisfies partial download: " + result + " " + debugString(downloadRequest));
        return result;
    }

    private String debugString(DownloadRequest request) {
        return " id - " + request.id
                + ", url - " + request.url
                + ", path - " + request.path
                + ", th - " + Thread.currentThread().getName();
    }

    private boolean shouldPause(DownloadRequest request) {

        if (!request.pauseOnConnectionLost)
            return false;

        return !isConnected(request);
    }


    @TargetApi(Build.VERSION_CODES.LOLLIPOP)
    private boolean isConnected(DownloadRequest request) {
        int conType = networkProvider.getCurrentNetworkType();

        if (conType >= 0 && request.networkType == NetworkType.ANY)
            return true;

        int mapped;
        switch (conType) {
            case TYPE_ETHERNET:
            case TYPE_WIFI:
            case TYPE_WIMAX:
                mapped = NetworkType.WIFI;
                break;
            case TYPE_BLUETOOTH:
            case TYPE_VPN:
            case TYPE_MOBILE_DUN:
            case TYPE_MOBILE:
                mapped = NetworkType.CELLULAR;
                break;
            default:
                mapped = -1;
        }

        boolean connected = mapped > 0 && (request.networkType & mapped) == mapped;

        Log.d(TAG, "checking pause for type: " + conType
                + " connected " + connected
                + debugString(request));

        return connected;
    }

    private void deliverProgress(
            String id,
            Progress progress
    ) {

        final Progress copy = Progress.copy(progress);
        Log.d(TAG, "Progress " + progress.progressPercent + " status " + progress.status + " " + id);

        final AssetDownloadListener downloadListener = listeners.get(id);
        final DownloadRequest downloadRequest = connections.get(id);
        if (downloadListener != null) {
            uiExecutor.execute(new Runnable() {
                @Override
                public void run() {
                    downloadListener.onProgress(copy, downloadRequest);
                }
            });
        }
    }

    public synchronized void cancel(@Nullable DownloadRequest request) {
        if (request == null) return;

        int status = request.getAndSetStatus(Status.CANCELLED);
        if (status != Status.IN_PROGRESS && status != Status.CANCELLED) {
            onCancelled(request, null);
            if (connections.isEmpty()) {
                networkProvider.removeListener(networkListener);
            }
        }
    }

    public boolean cancelAndAwait(@Nullable DownloadRequest request, long timeout) {
        if (request == null) return true;

        cancel(request);
        final String id = request.id;
        long until = System.currentTimeMillis() + Math.max(0, timeout);

        while (System.currentTimeMillis() < until) {
            DownloadRequest downloadRequest = connections.get(id);
            if (downloadRequest == null || downloadRequest.getStatus() != Status.CANCELLED) {
                Log.d(TAG, "Request is not present or status changed - finish await "
                        + (downloadRequest == null ? null : downloadRequest.getStatus()));
                return true;
            }
            try {
                Thread.sleep(10);
            } catch (InterruptedException e) {
                e.printStackTrace();
                return false;
            }
        }

        return false;
    }

    public synchronized void cancelAll() {
        for (DownloadRequest request : connections.values()) {
            cancel(request);
        }
    }

    public void setProgressStep(int value) {
        if (value != 0) {
            progressStep = value;
        }
    }

    private synchronized void onNetworkChanged(int connectionType) {
        Log.d(TAG, "Num of connections: " + connections.values().size());
        for (DownloadRequest request : connections.values()) {

            if (request.is(Status.CANCELLED)) {
                Log.d(TAG, "Result cancelled");
                continue;
            }

            boolean connected = isConnected(request);

            Log.d(TAG, "Connected = " + connected + " for " + connectionType);

            request.setConnected(connected);

            if (request.pauseOnConnectionLost) {
                if (connected && request.is(Status.PAUSED)) {
                    load(request);
                    Log.d(TAG, "resumed " + request.id);
                }
            }

        }
    }


    private boolean responseVersionMatches(File file, Response response) {
        File metaFile = new File(file.getPath() + META_POSTFIX_EXT);
        Map<String, String> meta = FileUtility.readMap(metaFile.getPath());

        Headers headers = response.headers();

        String etag = headers.get(ETAG);
        String lastModified = headers.get(LAST_MODIFIED);
        Log.d(TAG, "server etag: " + etag);
        Log.d(TAG, "server lastModified: " + lastModified);

        if (etag != null && !etag.equals(meta.get(ETAG))) {
            Log.d(TAG, "etags miss match current: " + meta.get(ETAG));
            return false;
        }

        if (lastModified != null && !lastModified.equals(meta.get(LAST_MODIFIED))) {
            Log.d(TAG, "lastModified miss match current: " + meta.get(LAST_MODIFIED));
            return false;
        }

        return true;
    }

    @IntDef(flag = true, value = {NetworkType.WIFI, NetworkType.CELLULAR, NetworkType.ANY})
    public @interface NetworkType {
        int CELLULAR = 1;
        int WIFI = 1 << 1;
        int ANY = CELLULAR | WIFI;
    }

    synchronized void shutdown() {
        cancel(null);
        listeners.clear();
        connections.clear();
        uiExecutor.shutdownNow();
        downloadExecutor.shutdownNow();
        try {
            downloadExecutor.awaitTermination(2, TimeUnit.SECONDS);
            uiExecutor.awaitTermination(2, TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    private static abstract class DownloadPriorityRunnable extends PriorityRunnable {
        private final int priority;

        private DownloadPriorityRunnable(int priority) {
            this.priority = priority;
        }

        @Override
        public Integer getPriority() {
            return priority;
        }
    }

    private final NetworkProvider.NetworkListener networkListener = new NetworkProvider.NetworkListener() {
        @Override
        public void onChanged(int type) {
            Log.d(TAG, "Network changed: " + type);
            onNetworkChanged(type);
        }
    };

    private static class RequestException extends Exception {

        RequestException(String message) {
            super(message);
        }

    }

}
