package com.vungle.warren.model;

import android.support.annotation.NonNull;

import java.util.HashMap;
import java.util.Map;

/**
 * Cookies allow our sdk to store state on disk without requiring a global state manager singleton.
 * They are used to store arbitrary data in key-value form for any part of the SDK that requires
 * persisting information.
 */
public class Cookie {

    /**
     * Special cookie identifier for the cookie that stores publisher appID for Moat
     */
    public static final String APP_ID = "appId";

    /**
     * Special cookie identifier for the cookie that stores GDPR consent state.
     */
    public static final String CONSENT_COOKIE = "consentIsImportantToVungle";

    /**
     * Special cookie identifier for the cookie that stores incentivized back button text by pubs.
     */
    public static final String INCENTIVIZED_TEXT_COOKIE = "incentivizedTextSetByPub";

    public static final String CONFIG_COOKIE = "configSettings";

    /**
     * Cookie identifier to store User Agent in cookie.
     */
    public static final String USER_AGENT_ID_COOKIE = "userAgent";

    /**
     * All of the strings stored within this cookie
     */
    Map<String, String> strings = new HashMap<>();

    /**
     * All of the booleans stored within this cookie
     */
    Map<String, Boolean> booleans = new HashMap<>();

    /**
     * All of the integers stored within this cookie
     */
    Map<String, Integer> integers = new HashMap<>();

    /**
     * All of the longs stored within this cookie
     */
    Map<String, Long> longs = new HashMap<>();

    /**
     * The identifier for this cookie. This is set by the creator of the cookie and is used to retrieve
     * it.
     */
    String identifier;

    public Cookie(String identifier) {
        this.identifier = identifier;
    }


    public <T> void putValue(String key, T value) {
        if (value instanceof String) {
            strings.put(key, (String) value);
        } else if (value instanceof Boolean) {
            booleans.put(key, (Boolean) value);
        } else if (value instanceof Integer) {
            integers.put(key, (Integer) value);
        } else if (value instanceof Long) {
            longs.put(key, (Long) value);
        } else {
            throw new IllegalArgumentException("Value type is not supported!");
        }
    }

    public Integer getInt(String key) {
        return integers.get(key);
    }

    public String getString(String key) {
        return strings.get(key);
    }

    public Boolean getBoolean(String key) {
        return booleans.get(key) != null ? booleans.get(key) : false;
    }

    public Long getLong(String key) {
        return longs.get(key) != null ? longs.get(key) : 0L;
    }

    @NonNull
    public String getId() {
        return identifier;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Cookie cookie = (Cookie) o;

        if (strings != null ? !strings.equals(cookie.strings) : cookie.strings != null)
            return false;
        if (booleans != null ? !booleans.equals(cookie.booleans) : cookie.booleans != null)
            return false;
        if (integers != null ? !integers.equals(cookie.integers) : cookie.integers != null)
            return false;
        if (longs != null ? !longs.equals(cookie.longs) : cookie.longs != null) return false;
        return identifier != null ? identifier.equals(cookie.identifier) : cookie.identifier == null;
    }

    @Override
    public int hashCode() {
        int result = strings != null ? strings.hashCode() : 0;
        result = 31 * result + (booleans != null ? booleans.hashCode() : 0);
        result = 31 * result + (integers != null ? integers.hashCode() : 0);
        result = 31 * result + (longs != null ? longs.hashCode() : 0);
        result = 31 * result + (identifier != null ? identifier.hashCode() : 0);
        return result;
    }
}
