package com.vungle.warren.model;

import android.support.annotation.IntDef;
import android.support.annotation.NonNull;
import android.support.annotation.Nullable;
import android.text.TextUtils;
import com.google.gson.JsonArray;
import com.google.gson.JsonObject;

import java.util.ArrayList;

/**
 * A container for an advertisement report, this will be populated as the advertisement plays and
 * once the advertisement is over, will save itself to a report directory, where it can be picked
 * up by the reporting mechanism to send to the server.
 */
public class Report {

	@IntDef({NEW, READY, SENDING, FAILED})
	public @interface Status {}

	public static final int NEW = 0;
	public static final int READY = 1;
	public static final int SENDING = 2;
	public static final int FAILED = 3;

	/**
	 * status = NEW for newly created reports.
	 * status = READY for Ad done playing and Ready to be Send.
	 * status = SENDING, is being send.
	 * status = FAILED if received network error.
	 *
	 * Default = NEW
	 */
	int status = NEW;

	/**
	 * Unique identifier for the placement where the advertisement was played.
	 */
	String placementId;

	/**
	 * Ad Token of the advertisement that was played. The server will use this identifier to link
	 * attribution back to us.
	 */
	String adToken;

	/**
	 * The identifier of the publisher application the ad was played in.
	 */
	String appId;

	/**
	 * Whether or not the advertisement was incentivized.
	 */
	boolean incentivized;

    /**
	 * The time at which the ad started playing, in seconds since epoch.
	 */
    long adStartTime;

	/**
	 * The url where the advertisement was played from.
	 */
	String url;

	/**
	 * The length of the video, in milliseconds. The advertisement metadata does not always
	 * contain this information, sometimes we have to add it to the report based on the media file
	 * information.
	 */
	long videoLength;
	/**
	 * Duration of complete ad experience in milliseconds.
	 */
	long adDuration;

	/**
	 * Legacy field for time to download the ad, in milliseconds.
	 */
	int ttDownload;

	/**
	 * The creative campaign where the advertisement was sourced from.
	 */
	String campaign;

	/**
	 * The identifier of the advertisement which this report is about.
	 */
	String advertisementID;


	/**
	 * The amount of the advertisement that was viewed, in milliseconds. This will always be less than
	 * or equal to {@link #videoLength}.
	 */
	int videoViewed;

	/**
	 * The dynamically growing list of user actions, a new action is added to this collection when
	 * the user interacts with the advertisement or continues watching it.
	 */
	final ArrayList<UserAction> userActions = new ArrayList<>();

	/**
	 *
	 */
	final ArrayList<String> clickedThrough = new ArrayList<>();


	/**
	 * The errors that occurred while playing the ad, if any.
	 */
	final ArrayList<String> errors = new ArrayList<>();

	/**
	 * The vungle ad type for the reported ad. e.g. "vungle_local" or "vungle_mraid"
	 */
	String adType;

	/**
	 * The identifier for the dynamic template used, if any.
	 */
	String templateId;

    /**
	 * There are some cases where we need to include a user identifier with our report, which allows
	 * publishers to verify their own in-app reward systems with the user identifiers of those players
	 * who watched the advertisement.
	 */
	String userID;

	/**
	 * Tracks the number of advertisements that have been played during this application's lifecycle.
	 * It is stored along with the placement and extracted and deleted when creating the advertisement
	 * report. This variable is set by the publisher because it is possible for them to have multiple
	 * advertisement providers.
	 *
	 * It has a sentinel value of -1, indicating it should not be included in the report.
	 */
	int ordinal;

	volatile boolean wasCTAClicked;

	Report() {

	}

	/**
	 * Construct a new report for an ad that is being played for the given {@link Placement}, and
	 * playing the given {@link Advertisement}. This should be created when the advertisement starts
	 * playing and actions can be added to it by the presenter.
	 * @param advertisement The Advertisement unit that is being played.
	 * @param placement The placement where the advertisement is being played.
	 * @param startTime The time (in seconds since epoch) when the advertisement started playing.
	 */
	public Report(@NonNull Advertisement advertisement, @NonNull Placement placement, long startTime){
		this(advertisement, placement, startTime, null);
	}

	/**
	 * Construct a new report for an ad that is being played for the given {@link Placement}, and
	 * playing the given {@link Advertisement}. This should be created when the advertisement starts
	 * playing and actions can be added to it by the presenter.
	 * @param advertisement The Advertisement unit that is being played.
	 * @param placement The placement where the advertisement is being played.
	 * @param startTime The time (in seconds since epoch) when the advertisement started playing.
	 * @param userID The identifier of the user watching the advertisement, provided by the publisher.
	 */
	public Report(@NonNull Advertisement advertisement, @NonNull Placement placement, long startTime, @Nullable String userID){
		placementId = placement.getId();
		adToken = advertisement.getAdToken();
		advertisementID = advertisement.getId();
		appId = advertisement.getAppID();
		incentivized = placement.isIncentivized();
		adStartTime = startTime;
		url = advertisement.getUrl();
		ttDownload = -1; /// TODO: store the time to download in the advertisement class?
		campaign = advertisement.getCampaign();
		switch (advertisement.getAdType()) {
			case Advertisement.TYPE_VUNGLE_LOCAL:
				adType = "vungle_local";
				break;
			case Advertisement.TYPE_VUNGLE_MRAID:
				adType = "vungle_mraid";
				break;
			default:
				throw new IllegalArgumentException("Unknown ad type, cannot process!");
		}
		templateId = advertisement.getTemplateId();
		if(userID == null){
			this.userID = "";
		} else {
			this.userID = userID;
		}
		ordinal = advertisement.getAdConfig().getOrdinal();
	}

	/**
	 * Record the given action in the current report.
	 * @param action The action taken
	 * @param value The value, if any, of the action
	 * @param timestamp The time at which the action occurred, in milliseconds since epoch.
	 */
	public synchronized void recordAction(String action, String value, long timestamp){
		userActions.add(new UserAction(action, value, timestamp));
		clickedThrough.add(action);
		if(action.equals("download")){
			wasCTAClicked = true;
		}
	}

	/**
	 * Record the error that occurred while rendering the advertisement.
	 * @param description The description of the error
	 */
	public synchronized void recordError(String description){
		errors.add(description);
	}


	/**
	 * Records the updated progress of the advertisement. New values override the older values.
	 * @param viewed The amount of the advertisement that has been viewed, in milliseconds.
	 */
	public void recordProgress(int viewed){
		videoViewed = viewed;
	}

	/**
	 * Provides the report with the duration of the advertisement.
	 * @param duration The duration of the advertisement, in milliseconds.
	 */
	public void setVideoLength(long duration){
		videoLength = duration;
	}

	/**
	 * Provide the report with complete duration of ad experience. Video viewed + postroll.
	 * @param duration in milliseconds
	 */
	public void setAdDuration(int duration){
		adDuration = duration;
	}

	/**
	 * Accessor for the placement identifier this report is for.
	 * @return The placement identifier
	 */
	public String getPlacementId(){
		return placementId;
	}

	/**
	 * Accessor for the advertisement identifier this report is about.
	 * @return The advertisement identifier.
	 */
	public String getAdvertisementID(){
		return advertisementID;
	}

	/**
	 * If CTA action was ever performed on the Ad, it is used to return on
	 *  {@link com.vungle.warren.PlayAdCallback#onAdEnd(String, boolean, boolean)}
	 * @return
	 */
	public boolean isCTAClicked() {
		return wasCTAClicked;
	}

    /**
     * Get userId for reporting Incentivized ads
     * @return
     */
    public String getUserID() {
        return userID;
    }
	/**
	 * Generates the report body that needs to be sent to the server in order to count the ad as
	 * viewed and to generate revenue. It has the following basic structure:
	 * <code><pre>
	   "request": {
	  		 "placement_reference_id": "string",
	 		 "ad_token": "string",
			 "app_id": "string",
			 "incentivized": 0,
			 "adStartTime": 0,
			 "plays": [
				 {
					 "startTime": 0,
					 "userActions": [
						 {
							 "action": "string",
							 "value": "string",
							 "timestamp_millis": 0
	 					}
	 				 ]
				 }
			 ],
			 "clickedThrough": [
				 "string"
	 		 ],
			 "url": "string",
			 "adDuration": 0,
			 "ttDownload": 0,
			 "campaign": "string",
			 "adType": "string",
			 "errors": [
				 "string"
			 ],
			 "templateId": "string"
		 },
	 * </pre></code>
	 * @return
	 */
	public JsonObject toReportBody(){
		JsonObject request = new JsonObject();

		request.addProperty("placement_reference_id", placementId);
		request.addProperty("ad_token", adToken);
		request.addProperty("app_id", appId);
		request.addProperty("incentivized", incentivized ? 1 : 0);
		request.addProperty("adStartTime", adStartTime);

		//on v5 server it is better not to send url if url field does not exist
		//since it saves some processing time and memory since server creates pointers
		//on empty string
		if (!TextUtils.isEmpty(url)) {
			request.addProperty("url", url);
		}

		request.addProperty("adDuration", adDuration);
		request.addProperty("ttDownload", ttDownload);
		request.addProperty("campaign", campaign);
		request.addProperty("adType", adType);
		request.addProperty("templateId", templateId);


		/// Plays
		JsonArray plays = new JsonArray();
		JsonObject playsObject = new JsonObject();
		playsObject.addProperty("startTime", adStartTime);
		if(videoViewed > 0){
			playsObject.addProperty("videoViewed", videoViewed);
		}
		if(videoLength > 0) {
			playsObject.addProperty("videoLength", videoLength);
		}
		JsonArray userActionJson = new JsonArray();
		for(UserAction action : userActions){
			userActionJson.add(action.toJson());
		}
		playsObject.add("userActions", userActionJson);
		plays.add(playsObject);
		request.add("plays", plays);

		/// Errors
		JsonArray errorsJson = new JsonArray();
		for(String s : errors){
			errorsJson.add(s);
		}
		request.add("errors", errorsJson);

		/// Clicked Through
		JsonArray clicked = new JsonArray();
		for(String s : clickedThrough){
			clicked.add(s);
		}
		request.add("clickedThrough", clicked);

		/// Incentivized User ID
		if(incentivized && !TextUtils.isEmpty(userID)){
			request.addProperty("user", userID);
		}

		/// Ordinal data. Only include this if value is >0.
		if(ordinal > 0){
			request.addProperty("ordinal_view", ordinal);
		}

		return request;
	}


	@NonNull
	public String getId() {
		return placementId + "_" + adStartTime;
	}

    public long getAdStartTime() {
        return adStartTime;
    }

	public static class UserAction {

		private String action;
		private String value;
		private long timestamp;
		private String id;

		public UserAction(String action, String value, long timestamp){
			this.action = action;
			this.value = value;
			this.timestamp = timestamp;
			this.id = "" + id;
		}

		@Override
		public boolean equals(Object obj) {
			if(obj instanceof UserAction){
				UserAction wrap = (UserAction) obj;
				if(!wrap.action.equals(action)) return false;
				if(!wrap.value.equals(value)) return false;
				if(wrap.timestamp != timestamp) return false;
				return true;
			}
			return false;
		}

		public JsonObject toJson() {
			JsonObject ret = new JsonObject();
			ret.addProperty("action", action);

			//allow us to send json objects with empty values
			if (value != null && !value.isEmpty()) {
				ret.addProperty("value", value);
			}

			ret.addProperty("timestamp_millis", timestamp);

			return ret;
		}
	}

	@Override
	public boolean equals(Object obj) {
		if(obj instanceof Report){
			Report wrap = (Report) obj;
			if(!wrap.placementId.equals(placementId)) return false;
			if(!wrap.adToken.equals(adToken)) return false;
			if(!wrap.appId.equals(appId)) return false;
			if(wrap.incentivized != incentivized) return false;
			if(wrap.adStartTime != adStartTime) return false;
			if(!wrap.url.equals(url)) return false;
			if(wrap.videoLength != videoLength) return false;
			if(wrap.adDuration != adDuration) return false;
			if(wrap.ttDownload != ttDownload) return false;
			if(!wrap.campaign.equals(campaign)) return false;
			if(!wrap.adType.equals(adType)) return false;
			if(!wrap.templateId.equals(templateId)) return false;
			if(wrap.wasCTAClicked != wasCTAClicked) return false;
			if(!wrap.userID.equals(userID)) return false;

			if(wrap.clickedThrough.size() != clickedThrough.size()) return false;
			for(int x = 0; x < clickedThrough.size(); x++){
				if(!wrap.clickedThrough.get(x).equals(clickedThrough.get(x))){
					return false;
				}
			}

			if(wrap.errors.size() != errors.size()) return false;
			for(int x = 0; x < errors.size(); x++){
				if(!wrap.errors.get(x).equals(errors.get(x))){
					return false;
				}
			}

			if(wrap.userActions.size() != userActions.size()) return false;
			for(int x = 0; x < userActions.size(); x++){
				if(!wrap.userActions.get(x).equals(userActions.get(x))){
					return false;
				}
			}

			return true;
		}
		return false;
	}

	@Override
	public int hashCode() {
		return super.hashCode();
	}

	public @Status int getStatus() {
		return status;
	}

	public void setStatus(@Status int status) {
		this.status = status;
	}
}