package com.vungle.warren.ui;

import android.annotation.SuppressLint;
import android.app.Activity;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.res.Configuration;
import android.os.Bundle;
import android.support.annotation.Nullable;
import android.support.v4.content.LocalBroadcastManager;
import android.util.Log;
import android.util.Pair;
import android.view.Window;
import android.view.WindowManager;
import com.vungle.warren.AdvertisementPresentationFactory;
import com.vungle.warren.BuildConfig;
import com.vungle.warren.Vungle;
import com.vungle.warren.error.VungleException;
import com.vungle.warren.ui.contract.AdContract;
import com.vungle.warren.ui.contract.AdContract.AdvertisementBus;
import com.vungle.warren.ui.state.BundleOptionsState;
import com.vungle.warren.ui.state.OptionsState;
import com.vungle.warren.ui.view.FullAdWidget;

import java.util.concurrent.atomic.AtomicBoolean;

/**
 * The Activity that the Vungle SDK uses to render advertisement content. This single activity can
 * start vungle_local and vungle_mraid advertisement types.
 */
public class VungleActivity extends Activity {

    public final static String PLACEMENT_EXTRA = "placement";

    private final static String TAG = "VungleActivity";
    public static final String PRESENTER_STATE = "presenter_state";


    /**
     * Eventbus to provide the presenter. Subscribers of this event bus use the lifecycle events
     * to update their own state.
     */
    private static AdContract.AdvertisementPresenter.EventListener bus;


    /**
     * The advertisement presenter does the heavy lifting for playing an advertisement. It has the
     * business logic to determine where to find the advertisement assets, manages creating the
     * report, and tracking state of the advertisement metadata. It needs to be attached to the view
     * in order to manage the view state.
     */
    @Nullable
    private AdContract.AdvertisementPresenter presenter;


    /**
     * Receiver used to handle events from outside the Activity, such as when the publisher calls
     * <code>close()</code> on an advertisement.
     */
    private BroadcastReceiver broadcastReceiver;

    /**
     * Placement identifier for which ad is going to be played.
     */
    private String placementId;
    private AdvertisementPresentationFactory presenterFactory;
    private OptionsState state;

    private AtomicBoolean pendingStart = new AtomicBoolean(false);
    /**
     * Provide an event listener for any event that happens in any instance of VungleActivity.
     *
     * @param listener Listener that will receive callbacks whenever events are reported.
     */
    public static void setEventListener(AdContract.AdvertisementPresenter.EventListener listener) {
        bus = listener;
    }

    @SuppressLint("SetJavaScriptEnabled")
    @Override
    protected void onCreate(@Nullable Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        requestWindowFeature(Window.FEATURE_NO_TITLE);

        getWindow().setFlags(
                WindowManager.LayoutParams.FLAG_HARDWARE_ACCELERATED,
                WindowManager.LayoutParams.FLAG_HARDWARE_ACCELERATED);

        FullAdWidget fullAdWidget = null;
        placementId = getIntent().getStringExtra(PLACEMENT_EXTRA);

        try {
            fullAdWidget = new FullAdWidget(this, getWindow());
        } catch (InstantiationException e) {
            if (bus != null) {
                bus.onError(e, placementId);
            }
            finish();

            if (BuildConfig.DEBUG) {
                throw new RuntimeException(e);
            }
            return;
        }

        if (!Vungle.isInitialized() || bus == null) {
            // Recovered from crash, deep sleep or else - this should not happen in normal life
            // SDK is not in operable state, BUS should we reworked and should not be set as static listener
            // instead we might use LocalBroadcastReceiver and let Vungle class to listen it
            // and dispatch event to appropriate listeners
            finish();
            return;
        }

        /// Initialize

        try {
            presenterFactory = new AdvertisementPresentationFactory(
                    placementId,
                    savedInstanceState,
                    new OrientationDelegate() {
                        @Override
                        public void setOrientation(int orientation) {
                            setRequestedOrientation(orientation);
                        }
                    },
                    new CloseDelegate() {
                        @Override
                        public void close() {
                            finish();
                        }
                    }
            );
        } catch (InstantiationException e) {
            Log.e(TAG, "error on crreating presentations" + e.getLocalizedMessage());
            deliverError(VungleException.AD_UNABLE_TO_PLAY, placementId);
            finish();
            return;
        }


        state = savedInstanceState == null
                ? null
                : (OptionsState) savedInstanceState.getParcelable(PRESENTER_STATE);


        presenterFactory.getFullScreenPresentation(
                this,
                fullAdWidget,
                state,
                fullscreenCallback);


        setContentView(fullAdWidget, fullAdWidget.getLayoutParams());

        connectBroadcastReceiver();
    }

    @Override
    protected void onNewIntent(Intent intent) {
        super.onNewIntent(intent);
        String oldPlacement = getIntent().getStringExtra(PLACEMENT_EXTRA);
        String newPlacement = intent.getStringExtra(PLACEMENT_EXTRA);
        if (oldPlacement != null && newPlacement != null && !oldPlacement.equals(newPlacement)){
            Log.d(TAG, "Tried to play another placement " + newPlacement + " while" +
                    " playing " + oldPlacement);
            deliverError(VungleException.ALREADY_PLAYING_ANOTHER_AD, newPlacement);
        }
    }

    private void deliverError(int code, String placementId) {
        if (bus != null) {
            bus.onError(new VungleException(code), placementId);
        }
    }


    private void connectBroadcastReceiver() {
        broadcastReceiver = new BroadcastReceiver() {
            @Override
            public void onReceive(Context context, Intent intent) {
                String command = intent.getStringExtra(AdvertisementBus.COMMAND);
                switch (command) {
                    case AdvertisementBus.CLOSE_FLEX: {
                        String placementID = intent.getStringExtra(AdvertisementBus.PLACEMENT);
                        if (presenter != null) {
                            presenter.handleExit(placementID);
                        }
                        break;
                    }
                    case AdvertisementBus.STOP_ALL: {
                        finish();
                        break;
                    }
                    default:
                        throw new IllegalArgumentException("No such command " + command);
                }
            }
        };

        LocalBroadcastManager.getInstance(getApplicationContext()).registerReceiver(broadcastReceiver, new IntentFilter(AdvertisementBus.ACTION));
    }

    @Override
    public void onWindowFocusChanged(boolean hasFocus) {
        super.onWindowFocusChanged(hasFocus);
        if (hasFocus) {
            if (presenter != null) {
                presenter.start();
            } else  {
                pendingStart.set(true);
            }
        } else {
            if (presenter != null) {
                presenter.stop(isChangingConfigurations(), isFinishing());
            }
            pendingStart.set(false);
        }
    }

    @Override
    public void onConfigurationChanged(Configuration newConfig) {
        super.onConfigurationChanged(newConfig);

        // Checks the orientation of the screen
        if (newConfig.orientation == Configuration.ORIENTATION_LANDSCAPE) {
            Log.d(TAG, "landscape");
        } else if (newConfig.orientation == Configuration.ORIENTATION_PORTRAIT) {
            Log.d(TAG, "portrait");
        }
        if (presenter != null) {
            presenter.onViewConfigurationChanged();
        }
    }

    @SuppressLint("ResourceType")
    @Override
    public void onBackPressed() {
        /// Only allow the advertisement to finish when back is pressed if the user has configured it.
        if (presenter != null) {
            presenter.handleExit(null);
        }
    }


    @Override
    protected void onSaveInstanceState(Bundle outState) {
        super.onSaveInstanceState(outState);
        Log.d(TAG, "onSaveInstanceState");
        BundleOptionsState optionsState = new BundleOptionsState();
        if (presenter != null) {
            presenter.generateSaveState(optionsState);
            outState.putParcelable(PRESENTER_STATE, optionsState);
        }
        presenterFactory.saveState(outState);
        /// Save the video position, if this is a video ad
    }

    @Override
    protected void onRestoreInstanceState(Bundle savedInstanceState) {
        super.onRestoreInstanceState(savedInstanceState);
        Log.d(TAG, "onRestoreInstanceState(" + savedInstanceState + ")");

        if (savedInstanceState != null && presenter != null) {
            presenter.restoreFromSave((OptionsState) savedInstanceState.getParcelable(PRESENTER_STATE));
        }
    }


    @Override
    public void setRequestedOrientation(int requestedOrientation) {
        if (canRotate()) {
            super.setRequestedOrientation(requestedOrientation);
        }
    }

    protected boolean canRotate() {
        return true;
    }


    @Override
    protected void onDestroy() {
        LocalBroadcastManager.getInstance(getApplicationContext()).unregisterReceiver(broadcastReceiver);

        if (presenter != null) {
            presenter.detach(isChangingConfigurations());
        } else if (presenterFactory != null) {
            presenterFactory.destroy();
            presenterFactory = null;
            if (bus != null) {
                bus.onError(new VungleException(VungleException.OPERATION_CANCELED), placementId);
            }
        }
        super.onDestroy();
        System.gc();
    }

    private AdvertisementPresentationFactory.FullScreenCallback fullscreenCallback = new AdvertisementPresentationFactory.FullScreenCallback() {
        @Override
        public void onResult(@Nullable Pair<AdContract.AdView, AdContract.AdvertisementPresenter> result, @Nullable Exception error) {
            //advertisement no longer exists kill activity

            if (result == null || error != null) {
                deliverError(VungleException.AD_UNABLE_TO_PLAY, placementId);
                finish();
                return;
            }

            presenter = result.second;
            presenter.setEventListener(bus);
            
            AdContract.AdView adView = result.first;

            presenter.attach(adView, state);

            if (pendingStart.getAndSet(false)) {
                presenter.start();
            }
        }
    };

}