package com.vungle.warren.utility;

import android.support.annotation.NonNull;
import android.support.annotation.Nullable;
import android.util.Log;

import com.vungle.warren.BuildConfig;

import java.io.BufferedInputStream;
import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 * TODO: Add JavaDoc
 */

public class FileUtility {

    private static final String TAG = FileUtility.class.getSimpleName();

    public static void printDirectoryTree(File folder) {
        if (!BuildConfig.DEBUG) return;

        if (folder == null) {
            Log.d(TAG, "File is null ");
            return;
        }

        if (!folder.exists()) {
            Log.d(TAG, "File does not exist " + folder.getPath());
            return;
        }

        if (!folder.isDirectory()) {
            Log.d(TAG, "File is not a directory " + folder.getPath());
            return;
        }

        int indent = 0;
        StringBuilder sb = new StringBuilder();
        printDirectoryTree(folder, indent, sb);
        Log.d("Vungle", sb.toString());
    }

    private static void printDirectoryTree(File folder, int indent,
                                           StringBuilder sb) {

        if (folder == null) {
            return;
        }

        if (!folder.isDirectory()) {
            throw new IllegalArgumentException("folder is not a Directory");
        }

        sb.append(getIndentString(indent));
        sb.append("+--");
        sb.append(folder.getName());
        sb.append("/");
        sb.append("\n");

        File[] files = folder.listFiles();
        if (files == null) {
            return;
        }

        for (File file : folder.listFiles()) {
            if (file.isDirectory()) {
                printDirectoryTree(file, indent + 1, sb);
            } else {
                printFile(file, indent + 1, sb);
            }
        }
    }

    private static void printFile(File file, int indent, StringBuilder sb) {
        sb.append(getIndentString(indent));
        sb.append("+--");
        sb.append(file.getName());
        sb.append("\n");
    }

    private static String getIndentString(int indent) {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < indent; i++) {
            sb.append("|  ");
        }
        return sb.toString();
    }

    /**
     * Helper method to recursively delete a directory. Used to clean up assets by their identifying
     * folder names.
     *
     * @param f The file or directory to delete.
     * @throws IOException if deleting a file fails for any reason.
     */
    public static void delete(File f) throws IOException {
        if (f == null || !f.exists()) return;

        if (f.isDirectory()) {
            File[] files = f.listFiles();
            if (files == null)
                return;

            for (File c : files)
                delete(c);
        }
        if (!f.delete())
            throw new FileNotFoundException("Failed to delete file: " + f);
    }

    public static void closeQuietly(Closeable closeable) {
        try {
            if (closeable != null)
                closeable.close();
        } catch (IOException ignored) {

        }
    }

    /**
     * Helper method extract bytes from file
     * folder names.
     *
     * @param file The file or directory to delete.
     * @return byte[] which is empty in case file is null or file size is 0
     * @throws IOException if fails to extract bytes.
     */
    @NonNull
    public static byte[] extractBytes(@Nullable File file) throws IOException {
        if (file == null) {
            return new byte[0];
        }

        int size = (int) file.length();

        if (size > 0) {
            byte[] bytes = new byte[size];
            BufferedInputStream buffy = null;
            try {
                buffy = new BufferedInputStream(new FileInputStream(file));
                int readSize = buffy.read(bytes);
                if (readSize < size) {
                    /// We were not able to read the contents of the file for some reason.
                    throw new IOException("Failed to read all bytes in the file, object recreation will fail");
                }
                /// Recreate the object now. The byte array must be cast to an Object because of the
                /// ambiguity of passing an array to a varargs method.
                return bytes;
            } catch (FileNotFoundException ex) {
                ex.printStackTrace();
            } finally {
                FileUtility.closeQuietly(buffy);
            }
        }

        return new byte[0];
    }

    @NonNull
    public static Map<String, String> readMap(@NonNull String path) {
        File file = new File(path);

        if (file.exists()) {
            Object ser = readSerializable(file);

            if (ser instanceof HashMap) {
                return (HashMap<String, String>) ser;
            }
        }

        return new HashMap<>();
    }

    public static void writeMap(@NonNull String path, @NonNull HashMap<String, String> map) {
        File file = new File(path);

        if (file.exists())
            file.delete();

        if (map.isEmpty())
            return;

        writeSerializable(file, map);
    }

    @NonNull
    public static ArrayList<String> readAllLines(@NonNull String path) {
        File file = new File(path);

        if (file.exists()) {
            Object ser = readSerializable(file);

            if (ser instanceof ArrayList) {
                return (ArrayList<String>) ser;
            }
        }

        return new ArrayList<>();
    }

    public static void writeAllLines(@NonNull String path, @NonNull ArrayList<String> lines) {
        File file = new File(path);

        if (file.exists())
            file.delete();

        if (lines.isEmpty())
            return;
        
        writeSerializable(file, lines);
    }

    public static void writeSerializable(File file, Serializable serializable) {
        ObjectOutputStream out = null;
        try {
            out = new ObjectOutputStream(new FileOutputStream(file));
            out.writeObject(serializable);
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            closeQuietly(out);
        }
    }

    @Nullable
    public static Object readSerializable(File file) {
        ObjectInputStream in = null;

        try {
            in = new ObjectInputStream(new FileInputStream(file));
            return in.readObject();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        } finally {
            closeQuietly(in);
        }

        return null;
    }
}
