package com.vungle.warren.utility;

import android.Manifest;
import android.content.Context;
import android.net.ConnectivityManager;
import android.net.Network;
import android.net.NetworkCapabilities;
import android.net.NetworkInfo;
import android.net.NetworkRequest;
import android.os.Build;
import android.os.Handler;
import android.os.Looper;
import android.support.annotation.Nullable;
import android.support.annotation.RequiresApi;
import android.support.v4.content.PermissionChecker;
import android.util.Log;
import com.vungle.warren.NetworkProviderReceiver;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

public class NetworkProvider {

    private static final String TAG = NetworkProvider.class.getSimpleName();
    private static volatile NetworkProvider INSTANCE;
    private final Context ctx;
    @Nullable
    private final ConnectivityManager cm;
    private AtomicInteger currentNetwork = new AtomicInteger();
    private ConnectivityManager.NetworkCallback networkCallback;
    private Set<NetworkListener> listeners = Collections.synchronizedSet(new HashSet<NetworkListener>());
    static final int TYPE_NONE = -1;
    private boolean enabled = false;
    private Handler handler = new Handler(Looper.getMainLooper());
    static final long NETWORK_CHECK_DELAY = 30 * 1000;


    private NetworkProvider(Context context) {
        ctx = context.getApplicationContext();
        cm = (ConnectivityManager) ctx.getSystemService(Context.CONNECTIVITY_SERVICE);
        currentNetwork.set(getCurrentNetworkType());
        if (Build.VERSION.SDK_INT < Build.VERSION_CODES.LOLLIPOP) {
            enabled = NetworkProviderReceiver.isEnabledBroadcastReceiver(context);
        } else {
            //enabled despite to manifest on some devices
            NetworkProviderReceiver.enableBroadcastReceiver(context, false);
        }
    }

    public static NetworkProvider getInstance(Context context) {
        if (INSTANCE == null) {
            synchronized (NetworkProvider.class) {
                if (INSTANCE == null) {
                    INSTANCE = new NetworkProvider(context);
                }
            }
        }

        return INSTANCE;
    }

    private synchronized void setEnableNetworkListener(boolean enabled) {
        if (this.enabled == enabled)
            return;
        this.enabled = enabled;

        if (Build.VERSION.SDK_INT < Build.VERSION_CODES.LOLLIPOP) {
            if (NetworkProviderReceiver.hasReceiver(ctx)) {
                NetworkProviderReceiver.enableBroadcastReceiver(ctx, enabled);
            } else {
                if (enabled) {
                    handler.postDelayed(typeRunnable, NETWORK_CHECK_DELAY);
                } else {
                    handler.removeCallbacks(typeRunnable);
                }
            }
        } else if (cm != null) {
            try {
                if (enabled) {
                    NetworkRequest.Builder builder = new NetworkRequest.Builder();
                    builder.addCapability(NetworkCapabilities.NET_CAPABILITY_INTERNET);
                    cm.registerNetworkCallback(builder.build(), getNetworkCallback());
                } else {
                    cm.unregisterNetworkCallback(getNetworkCallback());
                }
            } catch (Exception ex) {
                Log.e(TAG, ex.getMessage());
            }
        }
    }

    @RequiresApi(api = Build.VERSION_CODES.LOLLIPOP)
    private ConnectivityManager.NetworkCallback getNetworkCallback() {
        return networkCallback != null ? networkCallback
                : (networkCallback = new ConnectivityManager.NetworkCallback() {

            @Override
            @RequiresApi(api = Build.VERSION_CODES.LOLLIPOP)
            public void onAvailable(Network network) {
                super.onAvailable(network);
                onNetworkChanged();
            }

            @Override
            @RequiresApi(api = Build.VERSION_CODES.LOLLIPOP)
            public void onLost(Network network) {
                super.onLost(network);
                onNetworkChanged();
            }
        });
    }

    private void postToListeners(final int currentNetwork) {
        handler.post(new Runnable() {
            @Override
            public void run() {
                for (NetworkListener listener : listeners) {
                    listener.onChanged(currentNetwork);
                }
            }
        });
    }

    public void onNetworkChanged() {
        getCurrentNetworkType();
    }

    public int getCurrentNetworkType() {
        if (cm == null || PermissionChecker.checkCallingOrSelfPermission(ctx,
                Manifest.permission.ACCESS_NETWORK_STATE) != PermissionChecker.PERMISSION_GRANTED) {

            currentNetwork.set(TYPE_NONE);
            return TYPE_NONE;
        }

        NetworkInfo activeNetwork = cm.getActiveNetworkInfo();

        int newNetwork = activeNetwork != null && activeNetwork.isConnectedOrConnecting()
                ? activeNetwork.getType()
                : TYPE_NONE;

        int old = currentNetwork.getAndSet(newNetwork);

        if (newNetwork != old) {
            Log.d(TAG, "on network changed: " + old + "->" + newNetwork);
            postToListeners(newNetwork);
        }

        setEnableNetworkListener(!listeners.isEmpty());
        return newNetwork;
    }

    public void addListener(NetworkListener networkListener) {
        listeners.add(networkListener);
        setEnableNetworkListener(true);
    }

    public void removeListener(NetworkListener networkListener) {
        listeners.remove(networkListener);
        setEnableNetworkListener(!listeners.isEmpty());
    }

    private Runnable typeRunnable = new Runnable() {
        @Override
        public void run() {
            if (!listeners.isEmpty()) {
                onNetworkChanged();
                handler.postDelayed(typeRunnable, NETWORK_CHECK_DELAY);
            }
        }
    };

    public interface NetworkListener {

        void onChanged(int type);

    }
}
