package com.vungle.warren.utility;

import android.os.Handler;
import android.os.Looper;
import android.support.annotation.NonNull;

import java.util.List;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;

public class SDKExecutors {

    private static final int SINGLE_CORE_POOL_SIZE = 1;

    private static final int IO_KEEP_ALIVE_TIME_SECONDS = 5;
    private static final int VUNGLE_KEEP_ALIVE_TIME_SECONDS = 10;

    private static final int JOBS_KEEP_ALIVE_TIME_SECONDS = 1;
    private static int NUMBER_OF_CORES = Runtime.getRuntime().availableProcessors();

    /**
     * Executor to execute IO file, db calls.
     */
    private static final ExecutorService IO_EXECUTOR;
    /**
     * Executor to execute Vungle API calls.
     */
    private static final ExecutorService VUNGLE_EXECUTOR;
    /**
     * Executor to execute tasks and jobs.
     */
    private static final ExecutorService JOB_EXECUTOR;
    /**
     * Main/UI thread executor implementation.
     */
    private static final ExecutorService UI_EXECUTOR;

    static {
        ExecutorService uiExecutor = new AbstractExecutorService() {

            private final Handler UIHandler = new Handler(Looper.getMainLooper());

            @Override
            public void shutdown() {
                //no-op
            }


            @Override
            public List<Runnable> shutdownNow() {
                return null;
            }

            @Override
            public boolean isShutdown() {
                return false;
            }

            @Override
            public boolean isTerminated() {
                return false;
            }

            @Override
            public boolean awaitTermination(long timeout, @NonNull TimeUnit unit) {
                return false;
            }

            @Override
            public void execute(@NonNull Runnable command) {
                UIHandler.post(command);
            }

        };

        ThreadPoolExecutor jobExecutor = new ThreadPoolExecutor(
                NUMBER_OF_CORES,
                NUMBER_OF_CORES,
                JOBS_KEEP_ALIVE_TIME_SECONDS,
                TimeUnit.SECONDS,
                new PriorityBlockingQueue<Runnable>(),
                new NamedThreadFactory("vng_jr")
        );
        jobExecutor.allowCoreThreadTimeOut(true);

        ThreadPoolExecutor ioExecutor = new ThreadPoolExecutor(
                SINGLE_CORE_POOL_SIZE,
                SINGLE_CORE_POOL_SIZE,
                IO_KEEP_ALIVE_TIME_SECONDS,
                TimeUnit.SECONDS,
                new LinkedBlockingQueue<Runnable>(),
                new NamedThreadFactory("vng_io")
        );
        ioExecutor.allowCoreThreadTimeOut(true);

        ThreadPoolExecutor vungleExecutor = new ThreadPoolExecutor(
                SINGLE_CORE_POOL_SIZE,
                SINGLE_CORE_POOL_SIZE,
                VUNGLE_KEEP_ALIVE_TIME_SECONDS,
                TimeUnit.SECONDS,
                new LinkedBlockingQueue<Runnable>(),
                new NamedThreadFactory("vng_main")
        );
        vungleExecutor.allowCoreThreadTimeOut(true);

        UI_EXECUTOR = uiExecutor;
        JOB_EXECUTOR = jobExecutor;
        IO_EXECUTOR = ioExecutor;
        VUNGLE_EXECUTOR = vungleExecutor;
    }

    public ExecutorService getUIExecutor() {
        return UI_EXECUTOR;
    }

    public ExecutorService getVungleExecutor() {
        return VUNGLE_EXECUTOR;
    }

    public ExecutorService getIOExecutor() {
        return IO_EXECUTOR;
    }

    public ExecutorService getJobExecutor() {
        return JOB_EXECUTOR;
    }

    private static class NamedThreadFactory implements ThreadFactory {
        private final String name;
        private ThreadFactory threadFactory = Executors.defaultThreadFactory();
        private AtomicInteger atomicInteger = new AtomicInteger(0);

        public NamedThreadFactory(String name) {
            this.name = name;
        }

        public Thread newThread(@NonNull Runnable r) {
            Thread t = threadFactory.newThread(r);
            t.setName(name + "-th-" + atomicInteger.incrementAndGet());
            return t;
        }
    }

}
