package com.vungle.warren.utility;

import android.util.Log;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

/**
 * Utility
 */
public class UnzipUtility {
    /**
     * Size of the buffer to read/write data
     */
    private static final int BUFFER_SIZE = 4096;
    private static final String TAG = UnzipUtility.class.getCanonicalName();

    /**
     * Extracts a zip file specified by the zipFilePath to a directory specified by
     * destDirectory (will be created if does not exists)
     *
     * @param zipFilePath
     * @param destDirectory
     * @throws IOException
     */
    public static List<File> unzip(String zipFilePath, String destDirectory) throws IOException {
        return unzip(zipFilePath, destDirectory, null);
    }


    public static List<File> unzip(String zipFilePath, String destDirectory, Filter filter) throws IOException {
        File destDir = new File(destDirectory);
        if (!destDir.exists()) {
            destDir.mkdir();
        }

        ZipInputStream zipIn = new ZipInputStream(new FileInputStream(zipFilePath));
        ZipEntry entry = zipIn.getNextEntry();
        List<File> extractedFiles = new ArrayList<>();

        // iterates over entries in the zip file
        while (entry != null) {
            String filePath = destDirectory + File.separator + entry.getName();

            if (filter == null || filter.matches(filePath)) {
                validateFilename(filePath, destDirectory);
                if (entry.isDirectory()) {
                    // if the entry is a directory, make the directory if it does not exist.
                    File dir = new File(filePath);
                    if (!dir.exists()) {
                        dir.mkdir();
                    }
                } else {
                    // if the entry is a file, extracts it
                    extractFile(zipIn, filePath);
                    extractedFiles.add(new File(filePath));
                }
            }

            zipIn.closeEntry();
            entry = zipIn.getNextEntry();
        }
        zipIn.close();

        return extractedFiles;
    }

    /**
     * Extracts a zip entry (file entry)
     *
     * @param zipIn
     * @param filePath
     * @throws IOException
     */
    private static void extractFile(ZipInputStream zipIn, String filePath) throws IOException {
        File dest = new File(filePath);

        if (dest.exists())
            dest.delete();

        BufferedOutputStream bos = null;
        FileOutputStream fos = null;
        try {
            bos = new BufferedOutputStream(fos = new FileOutputStream(filePath));
            byte[] bytesIn = new byte[BUFFER_SIZE];
            int read = 0;
            while ((read = zipIn.read(bytesIn)) != -1) {
                bos.write(bytesIn, 0, read);
            }
        } catch (FileNotFoundException fnf) {
            /// The file will not be found if the parent directory of filePath does not exist yet.

            if (!dest.getParentFile().exists()) {
                File folder = dest.getParentFile();
                while (!folder.mkdir()) {
                    folder = folder.getParentFile();
                }
                extractFile(zipIn, filePath); /// Try again to unzip
            }
        } finally {
            FileUtility.closeQuietly(bos);
            FileUtility.closeQuietly(fos);
        }
    }

    private static String validateFilename(String filename, String intendedDir) throws IOException {
        File f = new File(filename);
        String canonicalPath = f.getCanonicalPath();

        File iD = new File(intendedDir);
        String canonicalID = iD.getCanonicalPath();

        if (canonicalPath.startsWith(canonicalID)) {
            return canonicalPath;
        } else {
            final String msg = "File is outside extraction target directory.";
            Log.e(TAG, msg);
            throw new ZipSecurityException(msg);
        }
    }

    static class ZipSecurityException extends IOException {
        ZipSecurityException(String message) {
            super(message);
        }
    }

    public interface Filter {
        boolean matches(String extractPath);
    }
}
